package archeopterix.algorithm;

import archeopterix.algorithm.performance.HypervolumeCalculator;
import archeopterix.conf.Conf;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.Vector;

import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import archeopterix.evaluation.Attribute;
import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.reporting.HypervolumeDevelopment;
import archeopterix.solution.Architecture;
import archeopterix.solution.ConstraintHandling;
import archeopterix.solution.comparator.SolutionComparator;
import archeopterix.util.PropertyMap;
import java.util.Set;

public abstract class Algorithm {

    private static final Log log = LogFactory.getLog(Algorithm.class);
    public static final String PARAM_STOP_CONDITION = "StopCondition";
    public static final String PARAM_MAX_ITERATIONS = "MaxIterations";
    public static final String PARAM_MAX_EVALUATIONS = "MaxEvaluations";
    public static final String PARAM_INIT_POP_CONSTRAINT_HANDLING = "InitPopConstraintHandling";
    public static final String PARAM_CONSTRAINT_HANDLING = "ConstraintHandling";
    protected int i_ECUs;
    protected int i_Comps;
    protected static EmbeddedSystemProblem problem;
    protected PropertyMap params;
    protected PopulationGenerator popGenerator;
    protected int maxEvaluations;
    public int currIteration;
    protected List<Architecture> population;
    protected List<Architecture> nondominatedSolutions;
    protected Watch m_Watch;
    protected SolutionComparator comp; //For the sorting in HV calculation
    protected List<Attribute> lst_MaxAttribValues;
    //protected ConstraintHandling initPopConstraintHandling;
    protected ConstraintHandling constraintHandling;
    private static HypervolumeCalculator hvCalculator ;
    /** Holds zeroes, because nothing beats a quick memory copy for zeroing an array :). */
    protected int[] allZeroes;
    /** Keeps track of the current number of solutions in each successive front. */
    protected int[] frontSizes;


    protected Algorithm() {
        params = new PropertyMap();
        params.put(PARAM_STOP_CONDITION, StopCondition.AFTER_N_ITERATIONS);
        params.put(PARAM_MAX_EVALUATIONS, 100);
        params.put(PARAM_INIT_POP_CONSTRAINT_HANDLING, ConstraintHandling.IGNORE);
      //  params.put(PARAM_CONSTRAINT_HANDLING, ConstraintHandling.IGNORE);

        //FIXME Determine most appropriate List implementation for these structures
        population = new LinkedList<Architecture>();
        nondominatedSolutions = new LinkedList<Architecture>();
        lst_MaxAttribValues = new Vector<Attribute>();
        m_Watch = new Watch();
        comp = new SolutionComparator();
        hvCalculator = new HypervolumeCalculator();
    }

    /**
     * Initialises the algorithm; in the top-most {@code Algorithm} class does
     * <em>nothing</em> with {@code partialInitPop} or {@code popGenerator},
     * leaving these for subclasses.
     * @param problem EmbeddedSystemProblem that is to be solved.
     * @param params Algorithm parameters to use.
     * @param partialInitPop A collection of solutions that should partially    
     * fill the initial population
     * @param popGenerator The population generator to generate all remaining
     * solutions.
     */
    public void initialise(EmbeddedSystemProblem oproblem, PropertyMap newparams, Collection<Architecture> partialInitPop, 
        PopulationGenerator opopGenerator) {
        params.addAll(newparams);
        problem = oproblem;
        popGenerator = opopGenerator;
        i_Comps = problem.getComponentCount();
        i_ECUs = problem.getECUCount();
        population.clear();
        nondominatedSolutions.clear();
        lst_MaxAttribValues.clear();
        maxEvaluations = params.getInt(PARAM_MAX_EVALUATIONS);
        constraintHandling = (ConstraintHandling) params.get(PARAM_CONSTRAINT_HANDLING);
        currIteration = 0;
        m_Watch.start();
    }
    public void setCurrIteration(int x){
	currIteration = x;
    }
    public double getQuality(){
        return computeBalancedHypervolume(nondominatedSolutions);
    }
    public EmbeddedSystemProblem getProblem() {
        return problem;
    }
    public synchronized void updatePopulation (Collection<Architecture> newPopulation){
        population.clear();
        population.addAll(newPopulation);
        processNewPopulation();
    }

    public synchronized void injectSolutions(Collection<Architecture> newPopulation){
        population.addAll(newPopulation);
    }
    /** Changes the parameter settings of the algorithm during a run. */
    public synchronized void changeParameters(PropertyMap pms) {
        //May duplicate some code from init(), but likely to be special cases since
        // this method could be called while the algorithm is in the middle of a run
        // (but not inside oneIteration())
        maxEvaluations = pms.getInt(PARAM_MAX_EVALUATIONS);
        constraintHandling = (ConstraintHandling) pms.get(PARAM_CONSTRAINT_HANDLING);
    }

    protected void incrementIteration() {
        currIteration++;
    }

//As this appears to be counted by the Stats object
//	protected void incrementEvaluations(int byHowMany) { currEvals += byHowMany; }
    public boolean hasMoreFunctionEvaluations() {
        Stats stats = Stats.instance();
        return stats.getEvaluationCounter() < maxEvaluations;
    }

    /**
     * Post-processing on newly generated solutions.
     * FIXME Propagate this method to all algorithm's that require it.
     */
    protected synchronized void processNewPopulation() {
        for (Architecture arch : population){
            problem.evaluateAttributes(arch);
            arch.setFeasible(problem.validate(arch));
            arch.setConstraintViolation(problem);
        }
    }

    public abstract void oneIteration();

    public void finalise(Log log) {
    	m_Watch.stop();
        printCurrentNondominatedSolutions("Final");
        double hv = computeBalancedHypervolume(nondominatedSolutions);
        log.debug(hv);
        HypervolumeDevelopment.getInstance().addHypervolumeForRun(hv);
    }

    public List<Architecture> getPopulation() {
        return population;
    }

    public List<Architecture> getNondominatedSet() {
        return nondominatedSolutions;
    }

    public static List<Architecture> extractNondominatedSet(List<Architecture> lstIn) {
        List<Architecture> lstOut = new LinkedList<Architecture>();
        for (int i = 0; i < lstIn.size(); i++) {
            Architecture oArch = lstIn.get(i);
           // problem.evaluateAttributes(oArch);
            lstOut.add(oArch);
        }
        for (int i = 0; i < lstOut.size(); i++) {
            for (int j = 0; j < lstOut.size(); j++) {
                if (i == j) {
                    continue;
                }
                if (IsDominated(lstOut.get(j), lstOut.get(i))) {
                    //d_SumOfFitness -= lst_Solutions.get(j).d_Fitness;
                    lstOut.remove(j);
                    i = (i > j) ? (i - 1) : i;
                    j = (j > 0) ? (j - 1) : 0;
                }
            }
        }
        return lstOut;
    }

    //FIXME What if iIndex > lst_MaxAttribValues.size() + 1? Then the given oAtt will be added to the wrong position!
    //FIXME Index-based system will need to be replaced with something better (or with more guarantees of which index represents which objective)
    protected boolean updateMaxAttributeValues(Attribute oAtt, int iIndex) {
        if (lst_MaxAttribValues.size() > iIndex) {
            Attribute oMax = lst_MaxAttribValues.get(iIndex);
            //if ( oMax.GetOptimizerValue() < oAtt.GetOptimizerValue()){
            //FIXME Er, GetOptimizerValue() may be a negative number, while GetFaceValue() might not be, how does this guarantee that we store the maximum/extreme value?
            if (oMax.getOptimizerValue() < oAtt.getValue()) {
                lst_MaxAttribValues.set(iIndex, oAtt);
                return true;
            } else {
                return false;
            }
        } else {
            lst_MaxAttribValues.add(oAtt);
            return true;
        }

    }

    protected double calculateFitness(Architecture sol) {
        if (lst_MaxAttribValues.isEmpty()) //in case called before these have been updated
        {
            return 0.0;
        }
        int attribIndex = 0;
        double dFitness = 0.0;
        for (Attribute attribute : sol.getAttributeList()) {
            dFitness += attribute.getOptimizerValue() / lst_MaxAttribValues.get(attribIndex).getOptimizerValue();
            attribIndex++;
        }
        sol.d_Fitness = (dFitness) / sol.getAttributeList().size();
        return sol.d_Fitness;
    }

    /**
     * Returns {@code true} iff it is dominated by an existing solution,
     * {@code false} otherwise.
     */
    protected boolean checkDominance(Architecture oDep) {
        for (int j = 0; j < nondominatedSolutions.size(); j++) {
            if (IsDominated(oDep, nondominatedSolutions.get(j))) {
                return true;
            }
            if (IsDominated(nondominatedSolutions.get(j), oDep)) {
                nondominatedSolutions.remove(j);
                j--;//TODO::Critical Check
            }
        }
        if (!isContaining(oDep)) //Newly added to avoid duplicates in the pareto fronts
        {
            nondominatedSolutions.add(oDep);
        }
        return false;

        //aadlplugin.Logger.Logln("Adding "+ oDep.toString());
        //oAnt.UpdateTrack();
    }

    /* Check whether the architecture is already there in the solutions list */
    protected boolean isContaining(Architecture oArc) {
        for (int j = 0; j < nondominatedSolutions.size(); j++) {
            if (nondominatedSolutions.get(j).getSerializedString().equals(oArc.getSerializedString())) {
                return true;
            }
        }
        return false;
    }
    /*
    Returns whether oD1 is Dominated by oD2
     */

    protected static boolean IsDominated(Architecture oD1, Architecture oD2) {
        List<Attribute> oAttSetDep1 = oD1.getAttributeList();
        List<Attribute> oAttSetDep2 = oD2.getAttributeList();

        for (int i = 0; i < oAttSetDep1.size(); i++) {
            Attribute oAttDep1 = oAttSetDep1.get(i);
            Attribute oAttDep2 = oAttSetDep2.get(i);
            if (oAttDep1.getOptimizerValue() > oAttDep2.getOptimizerValue()) {
                return false;
            }
        }
        return true;
    }

    protected void printCurrentNondominatedSolutions(String sText) {

        printSolutions("Non-dom " + sText, nondominatedSolutions);
    }

    protected void printSolutions(String sText, List<Architecture> lst_sol) {
        log.debug(sText);
        for (int i = 0; i < lst_sol.size(); i++) {
            Architecture oArch = lst_sol.get(i);
		log.debug("Sol" + i + "," + oArch.toString());
            //  Logger.Logln("Sol" + i + "," + oDep.toString());
            HypervolumeDevelopment.getInstance().writeToResultFile(oArch.toString());
        }
		log.debug("Hypervolume:"+HypervolumeDevelopment.getInstance().getHypervolume(lst_sol));
        //HypervolumeDevelopment.getInstance().writeToResultFile(System.getProperty("line.separator"));
    }

    protected void UpdateSolutions(List<Architecture> lstPop) {
        for (int i = 0; i < lstPop.size(); i++) {
            checkDominance(lstPop.get(i));
        }
    }

    protected double GetDistributionAndSpread() {
        Collections.sort(nondominatedSolutions, comp);
        double[] afAtts1 = new double[2];
        double[] afAtts2 = new double[2];
        double x, y;
        double delta = 0.0;
        double average = 0.0, sum = 0.0, mean = 0.0;
        double[] di = new double[nondominatedSolutions.size() - 1];

        for (int i = 0; i < nondominatedSolutions.size() - 1; i++) {
            Architecture oDep1 = nondominatedSolutions.get(i);
            Architecture oDep2 = nondominatedSolutions.get(i + 1);
            List<Attribute> lstAttributes1 = oDep1.getAttributeList();
            List<Attribute> lstAttributes2 = oDep2.getAttributeList();

            for (int j = 0; j < 2; j++) {
                Attribute oAtt1 = lstAttributes1.get(j);
                Attribute oAtt2 = lstAttributes2.get(j);

                if (oAtt1.isMinimizing()) {
                    afAtts1[j] = (lst_MaxAttribValues.get(j).getValue() - oAtt1.getValue());
                } else {
                    afAtts1[j] = oAtt1.getValue();
                }

                if (oAtt2.isMinimizing()) {
                    afAtts2[j] = (lst_MaxAttribValues.get(j).getValue() - oAtt2.getValue());
                } else {
                    afAtts2[j] = oAtt2.getValue();
                }
            }
            x = afAtts1[0] - afAtts2[0];
            y = afAtts1[1] - afAtts2[1];
            di[i] = Math.sqrt(x * x + y * y);
            if (i != 0 && i != (nondominatedSolutions.size() - 2)) {
                sum = sum + di[i];
            }
        }
        for (int i = 1; i < nondominatedSolutions.size() - 2; i++) {
            mean = Math.abs(di[i] - average);
        }
        average = sum / nondominatedSolutions.size() - 3;
        delta = (di[0] + di[nondominatedSolutions.size() - 2] + mean) / (di[0] + di[nondominatedSolutions.size() - 2] + (nondominatedSolutions.size() - 1) * average);
        return delta;
    }

    public double computeHypervolume(List<Architecture> solutions){
        
        int iSolutions = solutions.size();
        int iAttributes = Conf.instance().getEvaluatorsCount();
	if (iAttributes == 1)
            return solutions.get(0).getAttributeValue(0);
        double[][] adFront = new double[iSolutions][iAttributes];
        for (int i = 0; i < solutions.size(); i++) {
            Architecture arch = solutions.get(i);
            for (int j = 0; j < iAttributes; j++) {
                adFront[i][j] = arch.getAttributeValue(j);
            }
        }
        return hvCalculator.calculateHypervolume(adFront, iSolutions, iAttributes);
    }
	
    public static double computeBalancedHypervolume(List<Architecture> solutions){
        //System.out.println();
        int iSolutions = solutions.size();
        int iAttributes = Conf.instance().getEvaluatorsCount();
		//there should only be one individual for a single objective.
		//do not spend time checking. just use the first individual.
		if (iAttributes == 1)
			return solutions.get(0).getAttributeValue(0);
        double[][] adFront = new double[iSolutions][iAttributes];
        for (int i = 0; i < solutions.size(); i++) {
            Architecture arch = solutions.get(i);
            List<Attribute> attributes = arch.getAttributeList();
            try{
                for (int j = 0; j < iAttributes; j++)
                    adFront[i][j] = attributes.get(j).convertToNormalisedForHV();
            }catch (Exception ex){
               log.info("Solution excluded from HV ", ex);
               continue;
            }
            
        }
        return hvCalculator.calculateHypervolume(adFront, iSolutions, iAttributes);
    }

    public double computeMinWeightedFitness(List<Architecture> solutions){
        //System.out.println();
        int iAttributes = Conf.instance().getEvaluatorsCount();
		//there should only be one individual for a single objective.
		//do not spend time checking. just use the first individual.
	if (iAttributes == 1)
            return solutions.get(0).getAttributeValue(0);
            double min = 10010000.0;
        for (int i = 0; i < solutions.size(); i++) {
            double attrSum = 0.0;
            Architecture arch = solutions.get(i);
            List<Attribute> attributes = arch.getAttributeList();
            for (int j = 0; j < iAttributes; j++){
                attrSum = attrSum+attributes.get(j).getValue();
            }
            if(min<attrSum)
                min=attrSum;	
        }
	min=min/((double)iAttributes);
        return min;
    }


    /**
     * Finds the nondominated front in {@code lstSourcePop \ otherFronts} and
     * places these solutions in {@code lstNondominated}, which is cleared
     * first.
     * @param lstSourcePop The source population from which to extract the
     * nondominated front; this remains unchanged.
     * @param lstNondominated The resultant nondominated front.
     * @param otherFronts A {@code Set} containing solutions that will not be
     * considered for inclusion in the generated nondominated front; this set
     * is not modified.
     */
    public static void extractNondominatedFront(List<Architecture> lstSourcePop,
            LinkedList<Architecture> lstNondominated, Set<Architecture> otherFronts,
            ConstraintHandling constraintHandling, final double feasibilityRatio){

        lstNondominated.clear(); 
        if (lstSourcePop.isEmpty()){return;}	
        int first = 0;
        Architecture firstInclusion;
        do{
            firstInclusion = lstSourcePop.get(first);
            first++;
        } while (otherFronts != null && otherFronts.contains(firstInclusion));
        lstNondominated.add(firstInclusion);
        final int popSize = lstSourcePop.size(); 

        for (int i = first; i < popSize; i++){
            Architecture sol = lstSourcePop.get(i);
            assert !lstNondominated.contains(sol);
            if (otherFronts!=null &&!otherFronts.contains(sol)){ //consider this solution as per usual dominance check
                lstNondominated.add(sol);
                boolean isNondominated = true;
                int j = 0;
                while (isNondominated && j < lstNondominated.size() - 1){ //temporarily-added solution is guaranteed to be at the end
                    Architecture otherSol = lstNondominated.get(j);
                    if (solIsDominated(otherSol, sol, constraintHandling, feasibilityRatio)){
                        lstNondominated.remove(j);
                    } else if (solIsDominated(sol, otherSol, constraintHandling, feasibilityRatio)){
                        isNondominated = false;
                        lstNondominated.removeLast();
                    } else{
                        j++;
                    }
                }
            }
        }
    }
        
    private static boolean solIsDominated(Architecture sol1, Architecture sol2, ConstraintHandling constraintHandling, final double feasibilityRatio){				
        switch (constraintHandling){
            case PENALTY:
                return sol1.isPenaltyDominated(problem, sol2, feasibilityRatio);
            case CONSTRAIN_DOMINANCE:
                return sol1.isConstrainDominated(sol2);
            default:
                return sol1.isDominated(sol2);
        }
    }

    public String nonDominatedToString(){
        String sRet = "";
        for( int i=0; i<nondominatedSolutions.size(); i++){
            Architecture oDep = nondominatedSolutions.get(i);
            sRet +=  oDep.toString()+ System.getProperty("line.separator");


        }
        sRet += System.getProperty("line.separator");
        return sRet;
    }

     public int getNumberEvaluations(){
        return (int) Stats.instance().getEvaluationCounter();
     }

     protected void initialiseZeroArray(int howLong)
     {
        if (allZeroes == null || allZeroes.length < howLong) {
            allZeroes = new int[howLong];
            frontSizes = new int[howLong];
        } else {
            System.arraycopy(allZeroes, 0, frontSizes, 0, howLong);
        }
     }
}
