/**
 * 
 */
package archeopterix.algorithm;

import java.util.Collection;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import archeopterix.conf.Conf;
import archeopterix.gui.ChartUpdateListener;
import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.reporting.HypervolumeDevelopment;
import archeopterix.solution.Architecture;
import archeopterix.util.PropertyMap;
import archeopterix.util.io.ProblemXMLParser;
import java.io.File;
import java.io.IOException;
import java.util.List;

/**
 * TODO Make this the subject of remote calls to swap populations, change parameter settings, etc.
 * TODO Add cancel function; also add to Algorithm class
 */
public class AlgorithmExecutor {
	private static Log log = LogFactory.getLog(AlgorithmExecutor.class);
	
	/** Local reference to single configuration. */
	private Conf conf;
	private Algorithm alg;
	private EmbeddedSystemProblem problem;
	
	private boolean outsideChangeHasOccurred = false; 
	private Collection<Architecture> newPopulation;
	private Collection<Architecture> newSolutions;
	private PropertyMap newParams;
	
	//FIXME Possibly this belongs here
	protected ChartUpdateListener guiListener;
	
	public AlgorithmExecutor() {
            this( Conf.instance() );
	}

	public AlgorithmExecutor(Conf conf) {
            this.conf = conf;
	}

	public Algorithm run() {
            Class<?> algClass = conf.getAlgorithmClass();
            Class<?> problemClass = conf.getProblemClass();
            try {
                if (alg == null || ! alg.getClass().equals(algClass))
                        alg = (Algorithm) algClass.newInstance();
            } catch (Exception e) {
                String msg = "Error instantiating algorithm " + algClass.getCanonicalName();
                if (log.isErrorEnabled())
                        log.error(msg, e);
                throw new RuntimeException(msg, e);
            }

            //Always reload problem (easier than trying to determine if its source file has changed (also, might not come from a file in later versions)
            if (!conf.getProblemVaries())
                problem = ProblemXMLParser.loadProblem( conf.getProblemFile(), problemClass );

            //TODO Create chart window? (probably not here)   ChartWindow cw = null;

            //FIXME Where would PopulationGenerator be specified??
            conf.resetRandomSeed();
            PropertyMap map = conf.getApplicableParameters(algClass);
           
            if (map.get(Algorithm.PARAM_STOP_CONDITION) == StopCondition.AFTER_N_EVALUATIONS)
                HypervolumeDevelopment.createInstance(map.getInt(Algorithm.PARAM_MAX_EVALUATIONS), conf.getRunRepeats(), 500);
            else
                HypervolumeDevelopment.createInstance(map.getInt(Algorithm.PARAM_MAX_ITERATIONS), conf.getRunRepeats(), 0);

              
           
            HypervolumeDevelopment.getInstance().writeToResultFile(conf.toString());

            //Try to extract population from a file.
            List<Architecture> initialPopulation = null;
            if (conf.getPopulationFileName() != null){
                try {
                    initialPopulation = DefaultPopulationGenerator.ExtractPopulationFromFile(problem, conf.getPopulationFileName(), conf.getSolutionRepresentation().getName());
                } catch (IOException e) {
                    String msg = "Error reading the population file " + conf.getPopulationFileName();
                    System.out.println(msg);
                }
            }
            for (int repeat = 0; repeat < conf.getRunRepeats(); repeat++) {
                //FIXME May need a cleaner mechanism for iterating over successive problem files
                if (conf.getProblemVaries())
                    problem = ProblemXMLParser.loadProblem( new File( conf.getProblemFile().getPath().replaceAll("\\.xml$", "_" + (repeat+1) + ".xml") ), problemClass );

                if (conf.getRunRepeats() > 1 && conf.isRunResultsSeparate())
                    Conf.instance().initResultsLog((repeat+1));

                alg.initialise(problem, conf.getApplicableParameters(algClass), initialPopulation, new DefaultPopulationGenerator());
                Stats.instance().reset();

                while (alg.hasMoreFunctionEvaluations()){
                    dealWithAnyOutsideChange();
                    alg.oneIteration();
                }
                alg.finalise(log);  
                conf.useNextRandomSeed();
            }
            HypervolumeDevelopment.getInstance().logHypervolumes();
            return alg;
	}
	
	protected synchronized void dealWithAnyOutsideChange() {
            if (outsideChangeHasOccurred) {
                if (newPopulation != null) {
                        alg.updatePopulation(newPopulation);
                        newPopulation = null;
                }
                if (newSolutions != null) {
                        alg.injectSolutions(newSolutions);
                        newSolutions = null;
                }
                //FIXME Should the new params be recorded in the Configuration, replacing any old values? In which case, should they also be specified by Algorithm class?
                if (newParams != null) {
                        alg.changeParameters(newParams);
                        newParams = null;
                }
                outsideChangeHasOccurred = false;
            }
	}
	
	public synchronized void changePopulation(Collection<Architecture> newPopulation) {
            this.newPopulation = newPopulation;
            outsideChangeHasOccurred = true;
	}
	
	public synchronized void injectSolutions(Collection<Architecture> newSolutions) {
            this.newSolutions = newSolutions;
            outsideChangeHasOccurred = true;
	}
	
	public synchronized void changeParameters(PropertyMap params) {
            this.newParams = params;
            outsideChangeHasOccurred = true;
	}

}
