/**
 * 
 */
package archeopterix.algorithm;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.List;
import java.util.Vector;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import archeopterix.conf.Conf;
import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.solution.Architecture;
import archeopterix.solution.ConstraintHandling;
import archeopterix.solution.OutOfOptionsException;
import archeopterix.solution.Repairable;
import archeopterix.solution.comparator.SolutionComparator;
import archeopterix.util.StackTraceToString;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Collections;
import java.util.Iterator;
import java.util.Random;

/**
 * Adapted from Algorithm's GenerateRandomPopulation, but with only the code
 * for {@code ECUBasedDeployments} at this stage. A better system for selecting
 * the solution representation (based on the problem?) needs to be devised.
 * @author JMontgomery
 *
 */
public class DefaultPopulationGenerator implements PopulationGenerator {

	private static final Log log = LogFactory.getLog(DefaultPopulationGenerator.class);
	protected Class<? extends Architecture> lastUsedArchRep;
	protected Constructor<? extends Architecture> archIntIntCons;
	protected Constructor<? extends Architecture> archProblemCons;

	@Override
	public List<Architecture> generatePopulation(final int size, final EmbeddedSystemProblem problem) {
		return generatePopulation(size, problem, ConstraintHandling.IGNORE);
	}

	@Override
	public List<Architecture> generatePopulation(final int size, final EmbeddedSystemProblem problem,
			ConstraintHandling constraintHandling)
	{
		return generatePopulation(size, problem, constraintHandling, Conf.instance().getSolutionRepresentation());
	}

	
	@Override
	public List<Architecture> generatePopulation(final int size, final EmbeddedSystemProblem problem,
			ConstraintHandling constraintHandling, final Class<?> archRep)
	{
		return generatePopulation(new Vector<Architecture>(size), size, problem, constraintHandling, archRep);
	}

	@Override
	public List<Architecture> generatePopulation(List<Architecture> target, final int toSize,
			final EmbeddedSystemProblem problem, ConstraintHandling constraintHandling)
	{
		return generatePopulation(target, toSize, problem, constraintHandling,
				Conf.instance().getSolutionRepresentation());
	}
	
	
	@SuppressWarnings("unchecked")
	private Constructor<? extends Architecture> findArchConstructor(Class<?> archRep, Class<?>... sig) {
		try {
			lastUsedArchRep = (Class<? extends Architecture>) archRep;
			return (Constructor<? extends Architecture>) archRep.getConstructor(sig);
		} catch (SecurityException se) {
			throw new RuntimeException(se);
		} catch (NoSuchMethodException nsme) {
			return null;
		}
	}
    
	@SuppressWarnings("unchecked")
	@Override
	public List<Architecture> generatePopulation(List<Architecture> target, final int toSize,
			final EmbeddedSystemProblem problem, ConstraintHandling constraintHandling, final Class<?> archRep)
	{
		if (archIntIntCons == null && archProblemCons == null || !lastUsedArchRep.equals(archRep)) {
			lastUsedArchRep = (Class<? extends Architecture>) archRep;
			this.archProblemCons = findArchConstructor(archRep, EmbeddedSystemProblem.class);
			this.archIntIntCons = findArchConstructor(archRep, int.class, int.class);
			if (archIntIntCons == null && archProblemCons == null)
				throw new IllegalArgumentException("Preffered representation (" + archRep.getCanonicalName() + ") does not have a constructor with the signature (int, int)");
		}
		
		int initialContentSize = target.size();
		
		//Check constraints on existing solutions in target
		for (Architecture existingSol : target)
			existingSol.setFeasible( problem.validate(existingSol) );
		
		//final int maxAttempts = 20 * Math.abs(toSize - initialContentSize);
                final int maxAttempts = 100000;
		int attempts = 0;
		while (target.size() < toSize && attempts < maxAttempts) {
                    attempts++;
                    Architecture sol = createSolution(problem, constraintHandling);
                    if (sol != null)
                        target.add( sol );				
		}

		/*int selectedI=0;
		while (target.size() > toSize && attempts < maxAttempts && selectedI<target.size()) {
                    attempts++;
                    Architecture sol = target.get(selectedI);
                    for (int j=selectedI+1; j<target.size();j++){
                        if (NSGA.IsDominated(sol, target.get(j)) ) {
                            target.remove( sol );
                            break;
                        }
                    }
                    selectedI++;
		}*/
		Random rX = new Random();
		while (target.size() > toSize) {
                    target.remove(rX.nextInt(target.size()));
		}
		if (attempts == maxAttempts && target.size() < toSize) //unusual event, so make runtime rather than normal exception
			throw new RuntimeException("Unable to generate sufficient new solutions for population: needed to generate "
					+ (toSize - initialContentSize) + " but after " + attempts + " attempts only produced " +
					(target.size() - initialContentSize));



		return target;
	}
	
	/// ALDA: Needs to be merged with the above. I need this for single objective problems, where there are no dominance relations.


        @Override
	public List<Architecture> resizePopulation(List<Architecture> target, final int toSize,
			final EmbeddedSystemProblem problem, ConstraintHandling constraintHandling)
	{
		if (log.isDebugEnabled())
			log.debug("Generating Population of size " + (toSize - target.size()) );
		int initialContentSize = target.size();
		
		//Check constraints on existing solutions in target
		for (Architecture existingSol : target)
			existingSol.setFeasible( problem.validate(existingSol) );
		
		//final int maxAttempts = 20 * Math.abs(toSize - initialContentSize);
         final int maxAttempts = 100000;
		int attempts = 0;
		while (target.size() < toSize && attempts < maxAttempts) {
			attempts++;
			Architecture sol = createSolution(problem, constraintHandling);
			if (sol != null)
				problem.evaluateAttributes(sol);
				target.add( sol );				
		}

		SolutionComparator comparator = new SolutionComparator();
		comparator.setObjectiveForComparison(0);
		Collections.sort(target, comparator);
		while (target.size() > toSize) {
			target.remove(0);
		}
		if (attempts == maxAttempts && target.size() < toSize) //unusual event, so make runtime rather than normal exception
			throw new RuntimeException("Unable to generate sufficient new solutions for population: needed to generate "
					+ (toSize - initialContentSize) + " but after " + attempts + " attempts only produced " +
					(target.size() - initialContentSize));
		return target;
	}

	protected Architecture createSolution(final EmbeddedSystemProblem problem, ConstraintHandling constraintHandling) {
		Architecture sol = null;
		try {
			//Note: By preference calls the constructor that takes the EmbeddedSystemProblem, rather than just the number of ECUs and components
			sol = archProblemCons != null ? archProblemCons.newInstance(problem) :
					archIntIntCons.newInstance(problem.getECUCount(), problem.getComponentCount());
			//Deal with constraints
			sol.setFeasible( problem.validate(sol) );
			if ( ! sol.isFeasible() ) {
				if ( constraintHandling == ConstraintHandling.REPAIR && sol instanceof Repairable ) {
					try {
						((Repairable)sol).attemptRepair(problem);
					} catch (OutOfOptionsException ex) {
						if (log.isErrorEnabled())
							log.error(ex);
					}
					sol.setFeasible( problem.validate(sol) );
					if (! sol.isFeasible())
						sol = null;
				} else if ( constraintHandling == ConstraintHandling.DISCARD ) {
					sol = null;
				}
				/* All others are, at this stage, handled by doing nothing and accepting the solution as is. */
			}
		} catch (InvocationTargetException ite) {
			if (log.isErrorEnabled()) {
				log.error(ite.getCause());
				log.error(StackTraceToString.getStackTrace(ite));
			}
		} catch (Exception ex) {
			if (log.isErrorEnabled())
				log.error(ex);
		}
		return sol;
	}

	//FIXME Does not belong here
	public static void dumpPopulationToFile(List<Architecture> lstPop, String sFileName){
		PrintWriter oWriter;
		try {
			FileWriter outFile = new FileWriter(sFileName);
			oWriter = new PrintWriter(outFile);
		} catch (IOException e) {
			System.out.println("LOG:Error Occured in File Open for Write:" + sFileName);
			return;
		}
		Iterator<Architecture> itr = lstPop.iterator();
		if (lstPop.size() > 0){
			Class<?> cls = lstPop.get(0).getClass();
			//Constructor<?> cns = cls.getConstructors()[2];
			oWriter.println(cls.getName());
			//oWriter.println(cls.getName() + "," + cns.toString());
		}
		while(itr.hasNext())
			oWriter.println("["+itr.next().getSerializedString()+"]");

		oWriter.close();
	}

	//FIXME Does not belong here
	public static List<Architecture> ExtractPopulationFromFile(EmbeddedSystemProblem oProblem, String sFileName, String repType) throws IOException{
		BufferedReader in = null;
		StringBuilder depString = new StringBuilder();
		String line = null;
		String className = null;
		List<Architecture> lstPop = new Vector<Architecture>();
		try {
			in = new BufferedReader( new FileReader( sFileName) );
			className = in.readLine();
			// if we have the wrong representation, give up
			if (className != null && !className.equals(repType))
				return lstPop;
			while ((line = in.readLine()) != null)
			{
				//"readLine() strips the line breaks, so we have to put them back in
				depString.append(line+ System.getProperty("line.separator"));
				if (line.endsWith("]"))
				{
					line = depString.substring(depString.indexOf("[")+1, depString.indexOf("]"));
					depString = new StringBuilder();
					Architecture architecture;
					try {
						//(int,int) Constructor is assumed for (ECUs,Comps)
						architecture = (Architecture) Class.forName(className).getConstructor(Integer.TYPE,Integer.TYPE).newInstance(oProblem.getECUCount(),oProblem.getComponentCount()) ;
						architecture.initializeFromString(line, oProblem);
						lstPop.add( architecture );
					} catch (InstantiationException ex) {
						if (log.isErrorEnabled()) log.error(ex);
					} catch (IllegalAccessException ex) {
						if (log.isErrorEnabled()) log.error(ex);
					} catch (ClassNotFoundException ex) {
						if (log.isErrorEnabled()) log.error(ex);
					} catch (Exception ex) {
						if (log.isErrorEnabled()) log.error(ex);
					}
				}
			}
		} catch (IOException e) {
			System.out.println("Error Occured in File Open for Read: " + sFileName + ". " + e.getMessage());
		} finally {
			if (in != null)
				in.close();
		}

		return lstPop;
	}


}
