package archeopterix.algorithm;

import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.LinkedList;
import java.util.Random;
import java.util.BitSet;
import java.util.Vector;


import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


import archeopterix.algorithm.hillclimbing.HillClimbing;
import archeopterix.algorithm.paco.CompSelectionOrder;
import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.util.PropertyMap;
import archeopterix.conf.Conf;
import archeopterix.problem.Component;
import archeopterix.problem.Host;
import archeopterix.problem.deployment.solution.ECUBasedDeployment;
import archeopterix.solution.Architecture;
import archeopterix.solution.ArchitectureComparator;
import archeopterix.reporting.HypervolumeDevelopment;
import archeopterix.solution.OutOfOptionsException;

public class UniformRandom extends Algorithm
{

	private static final Log log = LogFactory.getLog(UniformRandom.class);
	//private static final Log resultsLog = LogFactory.getLog(Conf.RESULTS_LOG);
	public static final String PAR_SOL_PER_CYCLE = "solutionsCycle";
	public static final String PAR_USE_CONSTRAINTS = "useConstraints";
	public static final String PAR_COMP_SEL_ORDER = "compSelOrder";
	// Algorithm parameters
	private int solPerCycle;
	private int feasibleCount;
	private int evalCount;
	private int timeout;
	private CompSelectionOrder compSelOrder;

	private boolean useConstraints;
	private HillClimbing hillclimber;
	// Solutions in the Non-Dominated Front
	List<Architecture> selectedNondominated;
//	LinkedList<Architecture> currNondominated;
	// for sorting the front
	ArchitectureComparator[] comparator;
	private List<Component> components;
	private List<Component> orderedComponents;
	private List<Host> ecus;
	/**
	 * For each component the total memory size required by it and any other
	 * components with which it must be colocated.
	 */
	//private int[] compWithGroupSizes;
	// UniformRandom variable
	private Random rnd;
	private boolean doHillclimbing;

	public UniformRandom()
	{
		super();

		evalCount = Conf.instance().getEvaluatorsCount();
		rnd = Conf.instance().random;
		timeout = 30;

		selectedNondominated = new LinkedList<Architecture>();
//		currNondominated = new LinkedList<Architecture>();

		comparator = new ArchitectureComparator[evalCount];
		for (int i = 0; i < evalCount; i++)
		{
			comparator[i] = new ArchitectureComparator(i);
		}
	}

	@Override
	public void initialise(EmbeddedSystemProblem problem, PropertyMap newParams,
			Collection<Architecture> partialInitPop, PopulationGenerator popGenerator)
	{
		super.initialise(problem, newParams, partialInitPop, popGenerator);

		solPerCycle = params.getInt(PAR_SOL_PER_CYCLE);

		compSelOrder = (CompSelectionOrder) params.get(PAR_COMP_SEL_ORDER);
		doHillclimbing = params.getBoolean(HillClimbing.PAR_HILLCLIMBING);

		useConstraints = params.getBoolean(PAR_USE_CONSTRAINTS);
		feasibleCount = 0;

		switch (compSelOrder)
		{
			case RANDOMISED:
				orderedComponents = new Vector<Component>(problem.getComponents()); // in case wish to actually rearrange them
			case DEFAULT:
				orderedComponents = problem.getComponents();
				break;
			case FEWEST_ASSIGNMENTS:
				orderedComponents = problem.getCompsSortedByPossibleAssignments();
				break;
			case FEWEST_ASSIGNMENTS_LARGEST_SIZE:
				orderedComponents = problem.getCompsSortedByPossibleAssignmentsAndSize();
				break;
		}

		ecus = problem.getECUs();
		if (doHillclimbing)
		{
			hillclimber = new HillClimbing(problem, params);
		}
	}

	@Override
	public void oneIteration()
	{
		if (log.isDebugEnabled())
		{
			log.debug(m_Watch.GetElapsed() + ": Itr: " + (currIteration + 1) + " Evals="
					+ Stats.instance().getEvaluationCounter());
		}

		population.clear();

		if (useConstraints)
		{
			createFeasiblePopulation();
		} else
		{
			createRandomPopulation();
		}


		findNondominatedFront((LinkedList) nondominatedSolutions, population);

		if (doHillclimbing)
		{
			hillclimbSolutions(population);
		}
		double hv =computeHypervolume(nondominatedSolutions)*10;
		HypervolumeDevelopment.getInstance().addHypervolumeForIteration(hv, getNumberEvaluations());
		if (!useConstraints)
		{
			HypervolumeDevelopment.getInstance().writeToResultFile("Number of feasibles: " + feasibleCount);
			HypervolumeDevelopment.getInstance().writeToResultFile("Ratio: " + ((double) feasibleCount / (double) solPerCycle));
			feasibleCount = 0;
		}
		incrementIteration();
	}

	private void createRandomPopulation()
	{
		ECUBasedDeployment solution = null;
		int k = 0;
		while (k < solPerCycle)
		{
			k++;
			solution = constructNewSolution();
			problem.evaluateAttributes(solution);
			solution.setFeasible(problem.validate(solution));
			if (solution.isFeasible())
			{
				feasibleCount++;
				population.add(solution);
				System.out.println("Number of feasibles " + feasibleCount + " number of attempts " + k);
			}

		}
	}

	private void createFeasiblePopulation()
	{
		ECUBasedDeployment solution = null;
		int attempts = 0;
		int k = 0;
		int incomplete = 0;

		while (k < solPerCycle)
		{
			try
			{
				solution = constructNewSolutionUsingConstraints();
				problem.evaluateAttributes(solution);
				solution.setFeasible(problem.validate(solution));
				if (!solution.isFeasible())
				{
					System.out.println("INFEASIBLE");
				} else
				{
//                        System.out.println("feasible");
					population.add(solution);
				}
				attempts = 0;
				k++;
			} catch (OutOfOptionsException oe)
			{
				incomplete++;
				if (log.isDebugEnabled())
				{
					log.debug(m_Watch.GetElapsed() + "Failed PACO attempt " + oe.getMessage());
				}
				attempts++;
				if (attempts > timeout)
				{
					if (log.isInfoEnabled())
					{
						log.info(m_Watch.GetElapsed() + "No solution after " + timeout + " attempts. Giving up.");
					}
					System.exit(0);
				}

			}
		}

	}

	@Override
	public void finalise(Log log)
	{
		super.finalise(log);
		//FIXME Probably not a task that need be done by the algorithm itself; finalise() method only created to have a placeholder for this code, so potentially could be removed
		Collections.sort(nondominatedSolutions, comp);
		printCurrentNondominatedSolutions("Final");
		double hv = computeHypervolume(nondominatedSolutions)*10;
		HypervolumeDevelopment.getInstance().addHypervolumeForRun(hv);

		/*    if (doHillclimbing)
		{
		int avgImprovingMoves = hillclimber.getAverageImprovingMoves();
		resultLog.info("Hillclimber made an average " + avgImprovingMoves + " improving moves.");
		}*/
	}

	public ECUBasedDeployment constructNewSolutionUsingConstraints() throws OutOfOptionsException
	{
		List<Host> possibleEcus = new LinkedList<Host>();
		ECUBasedDeployment solution = new ECUBasedDeployment(i_ECUs, i_Comps);

		Component comp;
		int ecuIndex = 0;

		BitSet compBits = new BitSet();
		compBits.set(0, i_Comps);

		int[] memConstraints = new int[i_ECUs];
		boolean memOk;

		while (!compBits.isEmpty())
		{
			// Select next component to assign using current assignment order
			comp = selectNextComponent(compBits);

			// creating possible Host list
			possibleEcus.clear();
			possibleEcus.addAll(comp.getCanUseHosts());

			// Remove all Host from possible Host List where MustNotColocate
			// components have been assigned already
			for (Component c : comp.getMustNotColocateWith())
			{
				if (!compBits.get(c.index))
				{
					possibleEcus.remove(ecus.get(solution.getHostIndex(c.index)));
				}
			}


			if (possibleEcus.isEmpty())
			{
				throw new OutOfOptionsException("No ECU options because of enemy assignments");
			}

			memOk = false;
			while (!memOk)
			{
				ecuIndex = rnd.nextInt(possibleEcus.size());


				//Check Memory constraint
				if (memConstraints[ecuIndex] + comp.getIntProperty(Component.MEMORY) > ecus.get(ecuIndex).getIntProperty(Host.MEMORY))
				{
					if (log.isDebugEnabled())
					{
						log.debug(m_Watch.GetElapsed() + ": "
								+ " H" + i_ECUs + "_C" + i_Comps + ": Memory constraint violation ...");
					}
					possibleEcus.remove(ecus.get(ecuIndex));
					if (possibleEcus.isEmpty())
					{
						if (log.isDebugEnabled())
						{
							log.debug(m_Watch.GetElapsed() + ": "
									+ " H" + i_ECUs + "_C" + i_Comps + ": Solution Creation Error! No possible ECU's");
						}
						throw new OutOfOptionsException("No ECU options for comp " + comp.index);
					}

				} else
				{
					memConstraints[ecuIndex] += comp.getIntProperty(Component.MEMORY);
					memOk = true;
				}
			}

			// Check if Host assignments not possible

			solution.assignComponent(comp, ecuIndex);
			if (log.isDebugEnabled())
			{
				log.debug("Assigned component " + comp.index + " to ECU " + ecuIndex);
			}

			compBits.clear(comp.index);


		}
		if (log.isDebugEnabled())
		{
			log.debug("\n");
		}
		return solution;
	}

	public ECUBasedDeployment constructNewSolution()
	{
		ECUBasedDeployment solution = new ECUBasedDeployment(i_ECUs, i_Comps);

		Component comp;
		int ecuIndex = 0;

		BitSet compBits = new BitSet();
		compBits.set(0, i_Comps);


		while (!compBits.isEmpty())
		{
			// Select next component to assign using current assignment order
			comp = selectNextComponent(compBits);

			ecuIndex = rnd.nextInt(i_ECUs);


			solution.assignComponent(comp, ecuIndex);
			if (log.isDebugEnabled())
			{
				log.debug("Assigned component " + comp.index + " to ECU " + ecuIndex);
			}

			compBits.clear(comp.index);

		}
		if (log.isDebugEnabled())
		{
			log.debug("\n");
		}
		return solution;
	}

//this can add to an existing list of nondominated solutions
	public void findNondominatedFront(LinkedList<Architecture> currNondominated,
			List<Architecture> newSolutions)
	{
		boolean isNondominated;
		Architecture solution, domSolution;
		int i = 0;
		int j = 0;
		if (newSolutions.isEmpty())
		{
			return;
		}

		if (currNondominated.isEmpty())
		{
			currNondominated.add(newSolutions.get(0));
			i = 1;
		}

		for (; i < newSolutions.size(); i++)
		{
			solution = newSolutions.get(i);
			currNondominated.add(solution);
			isNondominated = true;
			j = 0;

			while (isNondominated && j < currNondominated.size() - 1)
			{
				domSolution = currNondominated.get(j);
				if (domSolution.isDominated(solution))
				{
					currNondominated.remove(j);
				} else if (solution.isDominated(domSolution))
				{
					isNondominated = false;
					currNondominated.removeLast();
				} else
				{
					j++;
				}
			}
		}

	}

	private void hillclimbSolutions(List<Architecture> deployments)
	{
		hillclimber.doHillclimbing(deployments, nondominatedSolutions);

	}

	protected Component selectNextComponent(BitSet available)
	{
		switch (compSelOrder)
		{
			case DEFAULT:
				return orderedComponents.get(available.nextSetBit(0)); //available indices are aligned with those in components list
			case FEWEST_ASSIGNMENTS:				//|-components' positions are not the same as their indices
			case FEWEST_ASSIGNMENTS_LARGEST_SIZE:	//|
				for (int chosen = 0; chosen < i_Comps; chosen++)
				{
					if (available.get(orderedComponents.get(chosen).index))
					{
						return orderedComponents.get(chosen);
					}
				}
			case RANDOMISED:
				return orderedComponents.get(getRndIndex(available));
		}
		return null; //Not possible to reach this point unless available.isEmpty()
	}

	protected int getRndIndex(BitSet available)
	{
		int index = -1;
		int bitCount = available.cardinality();
		if (bitCount != 0)
		{
			for (int i = 0, target = rnd.nextInt(bitCount); i <= target; i++)
			{
				index = available.nextSetBit(index + 1);
			}
		}
		return index;
	}
}
