package archeopterix.algorithm.aco;

import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


import archeopterix.algorithm.Algorithm;
import archeopterix.algorithm.PopulationGenerator;
import archeopterix.conf.Conf;
import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.solution.Architecture;
import archeopterix.solution.ArchitectureComparator;
import archeopterix.util.PropertyMap;

public class ACO extends Algorithm {

    private static final Log log = LogFactory.getLog(ACO.class);
    public static final String PARAM_N_ANTS = "numberOfAntsPerIteration";
    public static final String PARAM_BATCH_SIZE = "batchSize"; //FIXME What is batch size?
    public static final String PARAM_PHEROMONE_DECAY = "pheromoneDecay";
    public static final String PARAM_TRANSITION_RULE = "transitionRule";
    protected PheromoneMap[] arr_PhMap;
    protected int i_PhMapsCount;
    protected int i_Ants;
    protected int i_BatchSize;
    protected int i_MaxAntLifeTime;
    protected double d_PhDecay;
    protected double d_QTh; // To Decide whether to use probability or the Best
    //FIXME Does this belong here?
    protected Comparator<Architecture>[] arr_Comparator;
    protected ACOOperators operators;

    public ACO() {
        super();
    }

    @Override
    public void initialise(EmbeddedSystemProblem problem, PropertyMap params, Collection<Architecture> partialInitPop, PopulationGenerator popGenerator) {
        super.initialise(problem, params, partialInitPop, popGenerator);
        operators = ACOOperatorsFactory.instance().getSuitableOperators(Conf.instance().getSolutionRepresentation());
        i_PhMapsCount = Conf.instance().getNEvaluators();
        arr_PhMap = new PheromoneMap[i_PhMapsCount];
        arr_Comparator = (Comparator<Architecture>[]) new Comparator<?>[i_PhMapsCount];

        d_PhDecay = params.getDouble(PARAM_PHEROMONE_DECAY);
        d_QTh = params.getDouble(PARAM_TRANSITION_RULE);
        i_Ants = params.getInt(PARAM_N_ANTS);
        i_BatchSize = params.getInt(PARAM_BATCH_SIZE);

        for (int i = 0; i < i_PhMapsCount; i++) {
            arr_PhMap[i] = operators.createPheromoneMap(problem);
            arr_Comparator[i] = new ArchitectureComparator(i);
        }

        ReinitLocalObjects();
    }

    @Override
    public synchronized void changeParameters(PropertyMap params) {
        super.changeParameters(params);
        d_PhDecay = params.getDouble(PARAM_PHEROMONE_DECAY);
        d_QTh = params.getDouble(PARAM_TRANSITION_RULE);
        i_Ants = params.getInt(PARAM_N_ANTS);
        //TODO Now deal with different number of ants
        i_BatchSize = params.getInt(PARAM_BATCH_SIZE);
    }

    protected void ReinitLocalObjects() {
        for (PheromoneMap pher : arr_PhMap) {
            pher.reset();
        }
    }

    public int getPheromoneMapsCount() {
        return i_PhMapsCount;
    }

    public PheromoneMap getPheromoneMap(int iAttribute) {
        return arr_PhMap[iAttribute];
    }

    public double getPheromoneValue(int iAttribute, int iRow, int iCol) {
        return arr_PhMap[iAttribute].GetPheromone(iRow, iCol);
    }

    public void setPheromeneValue(int iAttribute, int iRow, int iCol, double dPh) {
        arr_PhMap[iAttribute].UpdatePheromone(iRow, iCol, dPh);
    }

    public double evaporatePheromone(double dPhOld) {
        return (dPhOld * (1 - d_PhDecay)) + (d_PhDecay);
    }

    public double calculateNewPheromone(double dPmOld, int iQuantity) {
        return (d_PhDecay * dPmOld) + ((1 - d_PhDecay) * iQuantity);
    }

    public double getThPheormone() {
        return d_QTh;
    }

    public int getBatchSize() {
        return i_BatchSize;
    }

    @Override
    public void oneIteration() {
        if (log.isTraceEnabled()) {
            log.trace(m_Watch.GetElapsed() + ":Starting "
                    + " H" + i_ECUs + "_C" + i_Comps
                    + " Ants=" + i_Ants + " MaxLife=" + i_MaxAntLifeTime
                    + " PhDecay=" + d_PhDecay + " Qtho" + d_QTh);
        }
        int iCurBatch = 0;
        for (int iAnt = 0; iAnt < i_Ants; iAnt++) {
            Ant oAnt = operators.CreateAnt(this, iAnt);
            do {
                oAnt.Travel();
            } while (!problem.validate(oAnt.m_Memory));
            problem.evaluateAttributes(oAnt.m_Memory);
            population.add(oAnt.m_Memory);
            if (++iCurBatch >= i_BatchSize) {
                System.out.print("#");
                CallUpdateTrack();
                iCurBatch = 0;
            }
        }
        incrementIteration();
    }

    protected void CallUpdateTrack() {
        // find first best and second best
        for (int i = 0, c = Conf.instance().getNEvaluators(); i < c; i++) {
            Ant fBest, sBest;
            Collections.sort(population, arr_Comparator[i]);
            if (population.size() >= 1) {
                fBest = (Ant) (population.get(0).m_UserData);
                fBest.UpdateTrack(i, 10);
                //System.out.println("Updating Best of ATT"+ i + " " + fBest.toString());
            }
            if (population.size() >= 2) {
                sBest = (Ant) (population.get(1).m_UserData);
                sBest.UpdateTrack(i, 5);
                //System.out.println("Updating SecBest of ATT"+ i + " " + sBest.toString());
            }
        }
    }
}
