/**
 * 
 */
package archeopterix.algorithm.ea;

import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.solution.Architecture;

/**
 * Common utility for calculating and assigning crowding distance measures to
 * solutions. Initial version adapted from NSGA-II C source. Other crowding
 * measures can be added later.
 * 
 * @author James Montgomery
 *
 */
public abstract class CrowdingUtils {
	
	/**
	 * Comparator that enforces the natural order of non-increasing crowding distance
	 */
	public static final Comparator<Architecture> CROWDING_DISTANCE_COMPARATOR = new Comparator<Architecture>() {
		public int compare(Architecture o1, Architecture o2) {
			double c1 = o1.getCrowdingDistance();
			double c2 = o2.getCrowdingDistance();
			if ( c1 > c2 )
				return -1;
			else if ( c1 < c2 )
				return 1;
			return 0;
		}
	};
	
	/**
	 * Calculates and assigns the crowding distance for individuals in the
	 * given population list that represent some front between
	 * {@code frontStart} and {@code frontEnd} <em>inclusive</em>. 
	 * @param pop The population from which a subset of solutions, representing
	 * some ranked front, will be assigned crowding distances.
	 * @param frontStart The first index of the front in the population list.
	 * @param frontEnd The last, inclusive index of the front in the population
	 * list.
 	 * @param context The problem context within which to judge solutions.
	 */
	public static void assignCrowdingDistance(List<Architecture> pop,
			int frontStart, int frontEnd, EmbeddedSystemProblem context)
	{
		final int frontSize = frontEnd - frontStart + 1;
		ObjectiveDistanceComparator objDistComparator = new ObjectiveDistanceComparator(0);
		
		//Deal with trivial case of two or fewer solutions
		if (frontSize <= 2) {
			pop.get(frontStart).setCrowdingDistance(Double.POSITIVE_INFINITY);
			if (frontSize == 2)
				pop.get(frontEnd).setCrowdingDistance(Double.POSITIVE_INFINITY);
			return;
		}
		
		//FIXME Sorry, but this is f**ked up: the problem should be able to say how many objectives there are, I should not have to go to an arbitrary solution to find out
		final int nObjectives = pop.get(frontStart).getAttributeList().size();
		Architecture[][] frontByObjective = new Architecture[nObjectives][frontSize];
		
		//Fill first objective list manually, then use quick arraycopy for others
		//Also use this loop to zero the front's constituents' crowding distances
		for (int i = 0, fi = frontStart; fi <= frontEnd; i++, fi++) {
			frontByObjective[0][i] = pop.get(fi);
			frontByObjective[0][i].setCrowdingDistance(0);
		}
		for (int obj = 0; obj < nObjectives; obj++) {
			if (obj > 0)
				System.arraycopy(frontByObjective[0], 0, frontByObjective[obj], 0, frontSize);
			objDistComparator.setCurrObjective(obj);
			Arrays.sort(frontByObjective[obj], objDistComparator);
			//Extremities are always kept hence considered not to be crowded
			frontByObjective[obj][0].setCrowdingDistance(Double.POSITIVE_INFINITY);
			frontByObjective[obj][frontSize - 1].setCrowdingDistance(Double.POSITIVE_INFINITY);
		}

		final int lastI = frontSize - 1;
		for (int obj = 0; obj < nObjectives; obj++) {
			double objRange = frontByObjective[obj][lastI].getAttributeValue(obj) - frontByObjective[obj][0].getAttributeValue(obj);
			//Check if something to add to crowding distance; if both extremes are at same location then can't add anything
			if ( objRange != 0 ) {
				for (int i = 1; i < lastI; i++) {
					if ( frontByObjective[obj][i].getCrowdingDistance() != Double.POSITIVE_INFINITY ) { //may be an extreme in another objective
						frontByObjective[obj][i].incrementCrowdingDistance(
								( frontByObjective[obj][i+1].getAttributeValue(obj) - frontByObjective[obj][i-1].getAttributeValue(obj) ) / objRange
							);
					}
				}
			}
		}
		//Normalise crowding distance by number of objectives
		for (Architecture sol : frontByObjective[0]) //each objective list contains same set of solutions
			if (sol.getCrowdingDistance() != Double.POSITIVE_INFINITY)
				sol.setCrowdingDistance( sol.getCrowdingDistance() / nObjectives );
	}
	
	/**
	 * Calculates and assigns the crowding distance for individuals in the
	 * given list and then sorts the list in non-increasing order of the
	 * crowding distances. 
	 * @param front The single front to assign crowding distances and sort.
	 * @param context The problem context within which to judge solutions. 
	 */
	public static void assignCrowdingDistanceAndSort(List<Architecture> front, EmbeddedSystemProblem context) {
		assignCrowdingDistance(front, 0, front.size() - 1, context);
		Collections.sort( front, CROWDING_DISTANCE_COMPARATOR );
	}
	
	//--Nested classes---------------------------------------------------------
	
	public static class ObjectiveDistanceComparator implements Comparator<Architecture> {
		private int currObjective = 0;
		
		public ObjectiveDistanceComparator(int currObjective) {
			setCurrObjective(currObjective);
		}
		
		public int getCurrObjective() { return currObjective; }
		
		public void setCurrObjective(int currObjective) { this.currObjective = currObjective; }
		
		public int compare(Architecture o1, Architecture o2) {
			//TODO If there is a way to cache these values, but still have them associated with the solutions, this would be much faster
			double obj1 = o1.getAttributeValue(currObjective);
			double obj2 = o2.getAttributeValue(currObjective);
			if (obj1 < obj2)
				return -1;
			else if (obj1 > obj2)
				return 1;
			return 0;
		}
	}; 

}
