package archeopterix.algorithm.ea;

import archeopterix.algorithm.Algorithm;
import archeopterix.algorithm.PopulationGenerator;
import archeopterix.conf.Conf;
import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.solution.Architecture;
import archeopterix.util.PropertyMap;

import java.util.Collection;
import java.util.LinkedList;


/**
 * Superclass of all EAs.
 */
public abstract class EA extends Algorithm {

    public static final String PARAM_POP_SIZE = "popSize";
    public static final String PARAM_GENERATION_RATE = "generationRate";
    public static final String PARAM_MUTATION_RATE = "mutationRate";
    public static final String PARAM_CROSSOVER_RATE = "crossoverRate";
    public static final String PARAM_MUTATION_OP_ID = "mutationOpID";
    public static final String PARAM_CROSSOVER_OP_ID = "crossoverOpID";
    public static final String PARAM_MUTATION_POINTS = "mutationPoints";

    protected int populationSize;
    protected int childrenPerIteration;
    protected double generationRate;
    protected double mutationProbability;
    protected MutationOperator mutationOperator;
    protected CrossoverOperator crossoverOperator;
    protected double crossoverProbability;
    protected int mutationPoints;
    protected String populationFile;
    protected double d_SumOfFitness;
protected EAOperators operators;


    protected EA() {
        super();
        params.put(PARAM_POP_SIZE, 100);
        params.put(PARAM_GENERATION_RATE, 0.8);
        params.put(PARAM_MUTATION_RATE, 0.9);
        params.put(PARAM_CROSSOVER_RATE, 0.9);
        params.put(PARAM_MUTATION_OP_ID, 0);

        population = new LinkedList<Architecture>();
    }

    @Override
    public void initialise(EmbeddedSystemProblem newProblem, PropertyMap externalParams, Collection<Architecture> partialInitPop, 
        PopulationGenerator popGenerator) {
        super.initialise(newProblem, externalParams, partialInitPop, popGenerator);

        operators = EAOperatorsFactory.instance()
                                                .getSuitableOperators( Conf.instance().getSolutionRepresentation() );
        nondominatedSolutions.clear();
        m_Watch.Reset();

        d_SumOfFitness = 0.0;
        populationSize = params.getInt(PARAM_POP_SIZE);

        //FIXME This is a still an odd way of determining the number of children to produce
        generationRate = params.getDouble(PARAM_GENERATION_RATE);
        childrenPerIteration = (int)((double)populationSize * generationRate);
        mutationProbability = params.getDouble(PARAM_MUTATION_RATE);
        mutationOperator = (MutationOperator)params.get(PARAM_MUTATION_OP_ID);
        crossoverOperator = (CrossoverOperator)params.get(PARAM_CROSSOVER_OP_ID);
        crossoverProbability = params.getDouble(PARAM_CROSSOVER_RATE);
        mutationPoints = params.getInt(PARAM_MUTATION_POINTS);

        if (partialInitPop != null) population.addAll( partialInitPop );
        popGenerator.generatePopulation(population, populationSize, problem, constraintHandling);
        processNewPopulation();
        printCurrentNondominatedSolutions("Intial");
    }

    @Override
    public void updatePopulation (Collection<Architecture> newPopulation){
        population.clear();
        population.addAll(newPopulation);
        popGenerator.generatePopulation(population, populationSize, problem, constraintHandling);
        processNewPopulation();
    }

    @Override
    public synchronized void changeParameters(PropertyMap newParameters) {
        super.changeParameters(newParameters);
        populationSize = newParameters.getInt(PARAM_POP_SIZE);
        LinkedList<Architecture> oldPopulation = new LinkedList<Architecture>();
        oldPopulation.addAll(population);
        updatePopulation(oldPopulation);

        childrenPerIteration = (int)((double)populationSize * newParameters.getDouble(PARAM_GENERATION_RATE));
        mutationProbability = newParameters.getDouble(PARAM_MUTATION_RATE);
        mutationOperator = (MutationOperator)newParameters.get(PARAM_MUTATION_OP_ID);
        crossoverOperator = (CrossoverOperator)newParameters.get(PARAM_CROSSOVER_OP_ID);
        crossoverProbability = newParameters.getDouble(PARAM_CROSSOVER_RATE);
        mutationPoints = newParameters.getInt(PARAM_MUTATION_POINTS);
        generationRate= newParameters.getDouble(PARAM_GENERATION_RATE);
    }

}
