package archeopterix.algorithm.ea;

import archeopterix.algorithm.DefaultPopulationGenerator;
import java.util.BitSet;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import archeopterix.algorithm.PopulationGenerator;
import archeopterix.algorithm.Stats;
import archeopterix.conf.Conf;
import archeopterix.evaluation.Attribute;
import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.reporting.HypervolumeDevelopment;
import archeopterix.solution.Architecture;
import archeopterix.solution.ConstraintHandling;
import archeopterix.solution.OutOfOptionsException;
import archeopterix.solution.Repairable;
import archeopterix.solution.comparator.SolutionComparator;
import archeopterix.util.PropertyMap;
import java.util.Collections;

public class GA extends EA {

	private final static Log log = LogFactory.getLog(NSGA.class);
	public static final String PARAM_MATING_POOL_SIZE = "matingPoolSize";
	public static final String PARAM_MATING_TOURNAMENT_SIZE = "matingTournamentSize";
	private int matingPoolSize;
	private int matingTournamentSize;
	private List<Architecture> matingPool;

	List<Architecture> lstSelected = null;
	//some ad hoc run stats
	private int crossoverFailureCounter = 0;
	private int mutationFailureCounter = 0;
	private int mutationCounter = 0;
	private int crossoverCounter = 0;

	public GA(){
            super();
            matingPool = new LinkedList<Architecture>();
            lstSelected = new LinkedList<Architecture>();
	}

	@Override
	public void initialise(EmbeddedSystemProblem problem, PropertyMap externalParams, Collection<Architecture> partialInitPop, PopulationGenerator popGenerator)
	{
            super.initialise(problem, externalParams, partialInitPop, popGenerator);
            log.info(Conf.instance());

            matingPoolSize = params.getInt(PARAM_MATING_POOL_SIZE);
            matingTournamentSize = params.getInt(PARAM_MATING_TOURNAMENT_SIZE);

            initialiseZeroArray(populationSize);
            nondominatedSolutions.clear();
            m_Watch.Reset();
            lst_MaxAttribValues.clear();

            System.out.println("Size of adopted population " + (partialInitPop == null ? 0 : partialInitPop.size()));

            if (log.isDebugEnabled()) {
                log.debug(m_Watch.GetElapsed() + ":Starting "
                                    + " H" + i_ECUs + "_C" + i_Comps
                                    + " PopSize=" + populationSize + " MaxEvals=" + maxEvaluations);
            }

            UpdateSolutions(population);
            if (log.isDebugEnabled()){
                log.debug(m_Watch.GetElapsed() + ": INITIAL POPULATION");
                printSolutions("Solutions: ", population);
            }
            if (Conf.instance().getPopulationFileName() == null){
                DefaultPopulationGenerator.dumpPopulationToFile(population, "./out/POP_H" + i_ECUs + "_C" + i_Comps + ".txt");
            }
	}

	@Override
	public synchronized void changeParameters(PropertyMap pms) {
		super.changeParameters(pms);
		if(pms.getInt(PARAM_MATING_POOL_SIZE)>0){
                    matingPoolSize = pms.getInt(PARAM_MATING_POOL_SIZE);
		}
		else{
                    matingPoolSize = params.getInt(PARAM_MATING_POOL_SIZE);
		}
		if(pms.getInt(PARAM_MATING_TOURNAMENT_SIZE)>0){
                    matingTournamentSize = pms.getInt(PARAM_MATING_TOURNAMENT_SIZE);
		}
		else{
                    matingTournamentSize = params.getInt(PARAM_MATING_TOURNAMENT_SIZE);
		}
		if (allZeroes.length < populationSize){
                    initialiseZeroArray(populationSize);
		}
	}

	@Override
	public void updatePopulation(Collection<Architecture> newPopulation)
	{
		population.clear();
		population.addAll(newPopulation);
		popGenerator.resizePopulation(population, populationSize, problem, constraintHandling);
		processNewPopulation();
	}


        @Override
	public void oneIteration()
	{
		SolutionComparator comparator = new SolutionComparator();
		comparator.setObjectiveForComparison(0);
		Collections.sort(population, comparator);
		//createMatingPool();
		doStochasticUniversalSelection(); // This is for better diversity in single objective problems
		performReproduction();
		popGenerator.resizePopulation(population, populationSize, problem, constraintHandling);
		processNewPopulation();
		
		nondominatedSolutions.clear();
		extractNondominatedFront(population, (LinkedList) nondominatedSolutions, ConstraintHandling.REPAIR, 1.0);

		double hv = computeBalancedHypervolume(nondominatedSolutions) ;
		HypervolumeDevelopment.getInstance().addHypervolumeForIteration(hv, getNumberEvaluations());
		if (log.isDebugEnabled()){
                    log.debug(m_Watch.GetElapsed() + ": Itr: " + (currIteration + 1) + " Evals="
                            + Stats.instance().getEvaluationCounter());
                    printCurrentNondominatedSolutions("Iteration :" + (currIteration + 1));
                    printSolutions("Iteration :" + (currIteration + 1),population);
		}
		
		incrementIteration();
	}

	private boolean solIsDominated(Architecture sol1, Architecture sol2,
            ConstraintHandling constraintHandling, final double feasibilityRatio){
            switch (constraintHandling){
                case PENALTY:
                    return sol1.isPenaltyDominated(problem, sol2, feasibilityRatio);
                case CONSTRAIN_DOMINANCE:
                    return sol1.isConstrainDominated(sol2);
                default:
                    return sol1.isDominated(sol2);
            }
	}


	private void performReproduction()
	{
		int iMatingPoolSize = matingPool.size();
		Architecture[] children ={null, null};
		int childrenCounter = 0;
		Random rnd = Conf.instance().random;
		//int itt=0;
		while (childrenCounter < childrenPerIteration)
		{
			Architecture child = null;
			if (mutationOperator==MutationOperator.UNIFORM_MUTATION){
				int iMutantIndex = rnd.nextInt(iMatingPoolSize);
				Architecture mutant = matingPool.get(iMutantIndex);
				child = operators.uniformMutation(mutant, problem, mutationProbability);
				processChild(child);
					childrenCounter++;
			}
			else if(rnd.nextDouble() < mutationProbability)
			{
				//Do Mutation
				int iMutantIndex = rnd.nextInt(iMatingPoolSize);
				Architecture mutant = matingPool.get(iMutantIndex);
				try
				{
					mutationCounter++;
					if (mutationOperator == MutationOperator.SINGLE_POINT_MUTATION){
                                            child = operators.singlePointMutation(mutant, problem);
					} else if (mutationOperator == MutationOperator.SWAP_MUTATATION){
                                            child = operators.swapMutation(mutant, problem);
					}
					
					else
					{
                                            child = operators.swapMutation(child, problem);
					}

					child.setFeasible(problem.validate(child));


					if (!child.isFeasible())
					{
                                            if (constraintHandling == ConstraintHandling.REPAIR){
                                                ((Repairable) child).attemptRepair(problem);
                                                child.setFeasible(true);
                                            } else//for other types of constraint handling we have to establish degree of infeasibilty
                                            {
                                                child.setConstraintViolation(problem);
                                            }
					}
					processChild(child);
					childrenCounter++;
				} catch (OutOfOptionsException ex)
				{
					mutationFailureCounter++;
				}

			}

			if (crossoverOperator==CrossoverOperator.UNIFORM_CROSSOVER){
                            Architecture parent1 = matingPool.get(rnd.nextInt(iMatingPoolSize));
                            Architecture parent2 = matingPool.get(rnd.nextInt(iMatingPoolSize));
                            children = operators.uniformCrossover(parent1, parent2, problem, crossoverProbability);
                            processChild(children[0]);
                            childrenCounter++;
                            processChild(children[1]);
                            childrenCounter++;
			}
			
			else if(rnd.nextDouble() < crossoverProbability)
			{
				Architecture parent1 = matingPool.get(rnd.nextInt(iMatingPoolSize));
				Architecture parent2 = matingPool.get(rnd.nextInt(iMatingPoolSize));
				try
				{
					crossoverCounter++;
					children = operators.singlePointCrossover(parent1, parent2, problem);
					for (int i = 0; i < children.length; i++)
					{
						children[i].setFeasible(problem.validate(children[i]));
						if (!children[i].isFeasible())
						{
							if (constraintHandling == ConstraintHandling.REPAIR)
							{
								//the bad thing is if this fails for one child, both are discarded
								((Repairable) children[i]).attemptRepair(problem);
								children[i].setFeasible(true);
							} else
							{
                                                            children[i].setConstraintViolation(problem);
							}
						}
					}

					processChild(children[0]);
					childrenCounter++;


					processChild(children[1]);
					childrenCounter++;
				} catch (OutOfOptionsException ex)
				{
					crossoverFailureCounter++;
					continue;
				}
			}
		}
	}


	private void doStochasticUniversalSelection(){
	   double min = population.get(0).d_Fitness;
	   double max = population.get(population.size()-1).d_Fitness;

	   double interval = (max - min)/(matingPoolSize+1);
	   //System.out.println("Max "+max+" min "+min+" interval "+interval);
	   double currentLocation = population.get(0).d_Fitness +  Conf.instance().random.nextDouble()*interval;
	   int i = 0;
	   int j = 0;
	   while ( i < matingPoolSize)
	   {
		  
               //System.out.println("Current location "+currentLocation);
               j = 0;
               while (j < population.size()-1)
               {
                   Architecture sol1 = population.get(j);
                   Architecture sol2 = population.get(j+1);

                   if (currentLocation>sol1.d_Fitness && currentLocation < sol2.d_Fitness)
                   {
                       if (sol1.d_Fitness - currentLocation > currentLocation - sol2.d_Fitness)
                            matingPool.add(sol1);
                       else
                            matingPool.add(sol2);
                       break;
                   }
                   else
                       j++;
               }

               i++;
               currentLocation += interval;
	   }
       }
        private void createMatingPool()
	{
		Random rnd = Conf.instance().random;
		matingPool.clear();

		lstSelected.clear();
		final int mergedPopSize = population.size();
		BitSet selectedIndices = new BitSet(mergedPopSize);

		//Will cache these as solutions added to tournament list
		int minRank, firstMinPos;
		int[] ranks = new int[matingTournamentSize];
		double[] crowdingDistances = new double[matingTournamentSize];

		for (int poolSize = 0; poolSize < matingPoolSize; poolSize++)
		{
			//Select elements for the tournament
			selectedIndices.clear();
			lstSelected.clear();
			minRank = populationSize;
			firstMinPos = 0;
			for (int i = 0; i < matingTournamentSize; i++)
			{
				int selectedI;
				do
				{
					selectedI = rnd.nextInt(mergedPopSize);
				} while (selectedIndices.get(selectedI));

				selectedIndices.set(selectedI);
				Architecture candidate = population.get(selectedI);
				lstSelected.add(candidate);
				ranks[i] = candidate.getRank();
				crowdingDistances[i] = candidate.getCrowdingDistance();
				if (ranks[i] < minRank)
				{
					minRank = ranks[i];
					firstMinPos = i;
				}
			}
			//Pick one that survives tournament using crowded comparison operator
			int winnerIndex = firstMinPos;
			for (int i = firstMinPos + 1; i < matingTournamentSize; i++)
			{
				if (ranks[i] == minRank && crowdingDistances[i] > crowdingDistances[winnerIndex])
				{
					winnerIndex = i;
				}
			}
			matingPool.add(lstSelected.get(winnerIndex));
		}

	}

	private void processChild(Architecture child){
            problem.evaluateAttributes(child);
            if (constraintHandling == ConstraintHandling.REPAIR && !child.isFeasible()){
                System.out.println("INVALID !");
            }
            population.add(child);
	}

	/**
	 * Finds the nondominated front in {@code lstSourcePop \ otherFronts} and
	 * places these solutions in {@code lstNondominated}, which is cleared
	 * first.
	 * @param lstSourcePop The source population from which to extract the
	 * nondominated front; this remains unchanged.
	 * @param lstNondominated The resultant nondominated front.
	 * considered for inclusion in the generated nondominated front; this set
	 * is not modified.
	 */
	public void extractNondominatedFront(List<Architecture> lstSourcePop,
            LinkedList<Architecture> lstNondominated,
            ConstraintHandling constraintHandling, final double feasibilityRatio){
            lstNondominated.clear();
            if (lstSourcePop.isEmpty()){
                return;
            }

            int first = 0;
            Architecture firstInclusion = lstSourcePop.get(first);
            lstNondominated.add(firstInclusion);
            final int popSize = lstSourcePop.size(); //saves method call overhead
            for (int i = first; i < popSize; i++){
                Architecture sol = lstSourcePop.get(i);
                //Deb's pseudocode includes a condition on the loop that should never occur, but just in case...
                assert !lstNondominated.contains(sol);
                lstNondominated.add(sol);
                boolean isNondominated = true;
                int j = 0;
                while (isNondominated && j < lstNondominated.size() - 1)
                { //temporarily-added solution is guaranteed to be at the end
                    Architecture otherSol = lstNondominated.get(j);
                    if (solIsDominated(otherSol, sol, constraintHandling, feasibilityRatio)){
                        lstNondominated.remove(j);
                    } else if (solIsDominated(sol, otherSol, constraintHandling, feasibilityRatio)){
                        isNondominated = false;
                        lstNondominated.removeLast();
                    } else{
                        j++;
                    }
                }
            }
	}


	//FIXME Should be stored and updated as new solutions enter population and old ones leave
	public int getFeasibleCount(){
            int feasible = 0;
            for (Architecture sol : population){
                if (sol.isFeasible()){
                    feasible++;
                }
            }
            return feasible;
	}


	private double getFeasibilityRatio(){
            return (double) getFeasibleCount() / population.size();
	}

	@Override
	protected synchronized void processNewPopulation(){
            int infeasiblesCounter = 0;
            for (Architecture sol : population)
            {
                problem.evaluateAttributes(sol);
                int attribIndex = 0;
                for (Attribute attribute : sol.getAttributeList()){
                    updateMaxAttributeValues(attribute, attribIndex++);
                }
                sol.setFeasible(problem.validate(sol));
                if (!sol.isFeasible()){
                    sol.setConstraintViolation(problem);
                    infeasiblesCounter++;
                }
            }
            System.out.println("Still infeasible " + infeasiblesCounter);
    }
}
