package archeopterix.algorithm.ea;

import archeopterix.algorithm.DefaultPopulationGenerator;
import java.util.BitSet;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import archeopterix.algorithm.PopulationGenerator;
import archeopterix.algorithm.Stats;
import archeopterix.conf.Conf;
import archeopterix.evaluation.Attribute;
import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.reporting.HypervolumeDevelopment;
import archeopterix.solution.Architecture;
import archeopterix.solution.ConstraintHandling;
import archeopterix.solution.OutOfOptionsException;
import archeopterix.solution.Repairable;
import archeopterix.util.PropertyMap;

public class NSGA extends EA {

    private final static Log log = LogFactory.getLog(NSGA.class);
    public static final String PARAM_MATING_POOL_SIZE = "matingPoolSize";
    public static final String PARAM_MATING_TOURNAMENT_SIZE = "matingTournamentSize";
    protected int matingPoolSize;
    private int matingTournamentSize;
    protected List<Architecture> matingPool;
    private LinkedList<Architecture> currFront = null;
    private List<Architecture> sortedNondominated = null;
    List<Architecture> lstSelected = null;
    //some ad hoc run stats
    protected int crossoverFailureCounter = 0;
    protected int mutationFailureCounter = 0;
    protected int mutationCounter = 0;
    protected int crossoverCounter = 0;

    protected Random rnd = Conf.instance().random;
    public NSGA()
    {
        super();
        matingPool = new LinkedList<Architecture>();
        currFront = new LinkedList<Architecture>();
        sortedNondominated = new LinkedList<Architecture>();
        lstSelected = new LinkedList<Architecture>();
    }

    @Override
    public void initialise(EmbeddedSystemProblem problem, PropertyMap externalParams, Collection<Architecture> partialInitPop, PopulationGenerator popGenerator)
    {
        super.initialise(problem, externalParams, partialInitPop, popGenerator);
        log.info(Conf.instance());
       
        matingPoolSize =(int)(params.getDouble(PARAM_MATING_POOL_SIZE) * populationSize);
        matingTournamentSize = params.getInt(PARAM_MATING_TOURNAMENT_SIZE);
        initialiseZeroArray(populationSize);
        nondominatedSolutions.clear();
        m_Watch.Reset();
        lst_MaxAttribValues.clear();
        System.out.println("Size of adopted population " + (partialInitPop == null ? 0 : partialInitPop.size()));
        fastNonDominatedSort(); //no solutions will be removed, but they will have ranks and crowding distances assigned and appear in blocks of successive rank in the population.
        UpdateSolutions(population);

        if (log.isDebugEnabled()){
            log.debug(m_Watch.GetElapsed() + ": INITIAL POPULATION");
            printSolutions("Solutions: ", population);
        }
        if (Conf.instance().getPopulationFileName() == null){
            DefaultPopulationGenerator.dumpPopulationToFile(population, "./POP_H" + i_ECUs + "_C" + i_Comps + ".txt");
        }
    }

    @Override
    public synchronized void changeParameters(PropertyMap newParameters) {
        super.changeParameters(newParameters);
        if(newParameters.getDouble(PARAM_MATING_POOL_SIZE)>0){
            matingPoolSize = (int)(newParameters.getDouble(PARAM_MATING_POOL_SIZE)*populationSize);
        }
        else{
            matingPoolSize =(int) (params.getDouble(PARAM_MATING_POOL_SIZE)*populationSize);
        }
        if(newParameters.getInt(PARAM_MATING_TOURNAMENT_SIZE)>0){
            matingTournamentSize = newParameters.getInt(PARAM_MATING_TOURNAMENT_SIZE);
        }
        else{
            matingTournamentSize = params.getInt(PARAM_MATING_TOURNAMENT_SIZE);
        }
        if (allZeroes.length < populationSize){
                initialiseZeroArray(populationSize);
        }
    }

    @Override
    public void updatePopulation(Collection<Architecture> newPopulation)
    {
        population.clear();
        population.addAll(newPopulation);
        popGenerator.generatePopulation(population, populationSize, problem, constraintHandling);
        processNewPopulation();
    }

    @Override
    public void oneIteration(){
        createMatingPool();
        evolve();
        fastNonDominatedSort();

        double hv = computeHypervolume(nondominatedSolutions) ;
        HypervolumeDevelopment.getInstance().addHypervolumeForIteration(hv, getNumberEvaluations());
        if (log.isDebugEnabled()){
            log.debug(m_Watch.GetElapsed() + ":Itr: " + (currIteration + 1) + ":Evals:"+ Stats.instance().getEvaluationCounter() + ":Hypervolume:"+ hv);
           // System.out.println(m_Watch.GetElapsed() + ":Itr: " + (currIteration + 1) + ":Evals:"+ Stats.instance().l_CntEvalus + ":Hypervolume:"+ hv);
           // printCurrentNondominatedSolutions("Iteration :" + (currIteration + 1));
        }
        incrementIteration();
    }


    public void evolve()
    {
        int iMatingPoolSize = matingPool.size();
        Architecture[] children ={null, null};
        int childrenCounter = 0;
        while (childrenCounter < childrenPerIteration)
        {
            children[0] = matingPool.get(rnd.nextInt(iMatingPoolSize));
            children[1]= matingPool.get(rnd.nextInt(iMatingPoolSize));
            if(rnd.nextDouble()<crossoverProbability){
                try{
                    crossoverCounter++;
                    children = operators.singlePointCrossover(children[0], children[1], problem);
                    for (int i = 0; i < children.length; i++){
                        children[i].setFeasible(problem.validate(children[i]));
                        if (!children[i].isFeasible()){
                            if (constraintHandling == ConstraintHandling.REPAIR){
                                ((Repairable) children[i]).attemptRepair(problem);
                                 children[i].setFeasible(true);
                            } else{
                                children[i].setConstraintViolation(problem);
                            }
                        }
                    }
                    processChild(children[0]);
                    childrenCounter++;
                    processChild(children[1]);
                    childrenCounter++; 

                } catch (OutOfOptionsException ex){
                    crossoverFailureCounter++;
                    continue;
                }
            }
            if (rnd.nextDouble() < mutationProbability){
                try{
                    mutationCounter++;
                    if (mutationOperator == MutationOperator.SINGLE_POINT_MUTATION){
                        children[0] = operators.singlePointMutation(children[0], problem);
                        children[1] = operators.singlePointMutation(children[1], problem);
                    } else if (mutationOperator == MutationOperator.SWAP_MUTATATION){
                        children[0] = operators.swapMutation(children[0], problem);
                        children[1] = operators.swapMutation(children[1], problem);
                    }
                    else if (mutationOperator==MutationOperator.UNIFORM_MUTATION){
                        children[0] = operators.uniformMutation(children[0], problem, mutationProbability);
                        children[1] = operators.uniformMutation(children[1], problem, mutationProbability);
                    }
                    children[0].setFeasible(problem.validate(children[0]));
                    children[1].setFeasible(problem.validate(children[1]));
                    if (!children[0].isFeasible()){
                        if (constraintHandling == ConstraintHandling.REPAIR){
                            ((Repairable) children[0]).attemptRepair(problem);
                            children[0].setFeasible(true);
                        } else{
                            children[0].setConstraintViolation(problem);
                        }
                    }
                    if (!children[1].isFeasible()){
                        if (constraintHandling == ConstraintHandling.REPAIR){
                            ((Repairable) children[1]).attemptRepair(problem);
                            children[1].setFeasible(true);
                        } else{
                            children[1].setConstraintViolation(problem);
                        }
                    }
                    processChild(children[0]);
                    childrenCounter++;
                    processChild(children[1]);
                    childrenCounter++; 
                } catch (OutOfOptionsException ex){
                    System.err.println(ex.getMessage());
                    mutationFailureCounter++;
                }   
            }   
        }
    }


    private void createMatingPool(){
        matingPool.clear();
        lstSelected.clear();
        final int mergedPopSize = population.size();
        BitSet selectedIndices = new BitSet(mergedPopSize);

        //Will cache these as solutions added to tournament list
        int minRank, firstMinPos;
        int[] ranks = new int[matingTournamentSize];
        double[] crowdingDistances = new double[matingTournamentSize];

        for (int poolSize = 0; poolSize < matingPoolSize; poolSize++){
            selectedIndices.clear();
            lstSelected.clear();
            minRank = populationSize;
            firstMinPos = 0;
            for (int i = 0; i < matingTournamentSize; i++){
                int selectedI;
                do{
                    selectedI = rnd.nextInt(mergedPopSize);
                } while (selectedIndices.get(selectedI));
                selectedIndices.set(selectedI);
                Architecture candidate = population.get(selectedI);
                lstSelected.add(candidate);
                ranks[i] = candidate.getRank();
                crowdingDistances[i] = candidate.getCrowdingDistance();
                if (ranks[i] < minRank){
                        minRank = ranks[i];
                        firstMinPos = i;
                }
            }
            //Pick one that survives tournament using crowded comparison operator
            int winnerIndex = firstMinPos;
            for (int i = firstMinPos + 1; i < matingTournamentSize; i++){
                if (ranks[i] == minRank && crowdingDistances[i] > crowdingDistances[winnerIndex]){
                    winnerIndex = i;
                }
            }
            matingPool.add(lstSelected.get(winnerIndex));
        }
    }

    protected void processChild(Architecture child){
        try{
            problem.evaluateAttributes(child);
        }
        catch(Exception ex){
            System.out.println(ex.toString());
        }
        population.add(child);
    }

    public void fastNonDominatedSort(){
        sortedNondominated.clear();
        Set<Architecture> allFronts = new HashSet<Architecture>();
        final double feasibilityRatio = (double) getFeasibleCount() / population.size();

        System.arraycopy(allZeroes, 0, frontSizes, 0, frontSizes.length);
        int rank = 0;
        while (sortedNondominated.size() < populationSize){
            rank++; 
            currFront.clear();
            extractNondominatedFront(population, currFront, allFronts, constraintHandling, feasibilityRatio);
            int currFrontSize = currFront.size();
            frontSizes[rank - 1] = currFrontSize;
            for (Architecture sol : currFront){ sol.setRank(rank); }
            //  if current front will fit into population, add it, otherwise use crowding measure to select from this, final front to be kept.
            if (sortedNondominated.size() + currFront.size() <= populationSize){
                sortedNondominated.addAll(currFront);
                allFronts.addAll(currFront);
            } else{//NSGA-II's crowding fill
                currFrontSize= populationSize - sortedNondominated.size();
                for (int i = 0; i < currFrontSize; i++){
                    sortedNondominated.add(currFront.get(i));
                }
            }
        }
        population.clear();
        population.addAll(sortedNondominated);
        nondominatedSolutions.clear();
        try{
            nondominatedSolutions.addAll(population.subList(0, frontSizes[0]));
        }catch(Exception ex){
            nondominatedSolutions=extractNondominatedSet(population);
        }
        
    }
    public int getFeasibleCount(){
        int feasible = 0;
        for (Architecture sol : population){
            if (sol.isFeasible()){
                feasible++;
            }
        }
        return feasible;
    }

    @Override
    protected synchronized void processNewPopulation(){
        int infeasiblesCounter = 0;
        for (Architecture sol : population){
            problem.evaluateAttributes(sol);
            int attribIndex = 0;
            for (Attribute attribute : sol.getAttributeList()){
               updateMaxAttributeValues(attribute, attribIndex++);
            }
            sol.setFeasible(problem.validate(sol));
            if (!sol.isFeasible()){
                sol.setConstraintViolation(problem);
                infeasiblesCounter++;
            }
        } 

    }	
}
