/**
 * 
 */
package archeopterix.algorithm.ea;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;
import java.util.Vector;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


import archeopterix.algorithm.Algorithm;
import archeopterix.algorithm.PopulationGenerator;
import archeopterix.conf.Conf;
import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.reporting.HypervolumeDevelopment;
import archeopterix.solution.Architecture;
import archeopterix.solution.ConstraintHandling;
import archeopterix.util.PropertyMap;

/**
 * Not an NSGA itself but manages two or more NSGAs and swaps certain solutions
 * between them at regular intervals.
 * 
 * @author James Montgomery
 *
 */
public class ParallelNSGA extends Algorithm {
	private static final Log log = LogFactory.getLog(ParallelNSGA.class);
	
	public static final String PARAM_SWAP_SOLS_AFTER_ITERATIONS = "itersBetweenSolSwaps";
	public static final String PARAM_ALG_CONSTRAINT_HANDLING = "algConstraintHandling";
	public static final String PARAM_POP_SWAP_REGIME = "popSwapRegime";
//	public static final String PARAM_NUM_CYCLES = "cycles";
//	public static final String PARAM_REUSE_NONDOMINATED_ONLY = "reuseNondominatedSolsOnly";
//	public static final String PARAM_ALGORITHMS = "algorithms";
	
	protected int swapPeriod;
	protected PopSwapRegime swapRegime;
//	protected int nCycles;
//	protected boolean reuseNondominatedSolsOnly;
	protected List<Class<?>> algClasses;
	protected List<NSGA> algs;
	/**
	 * Only needed to support changing of algorithms' parameters at runtime.
	 */
	protected List<PropertyMap> algParams;
//	protected int currAlgIndex;
//	protected Algorithm currAlg;
//	protected int currCycle;
	
	private List<ConstraintHandling> constraintHandlingUsed;
	private List<NSGARunner> runners;
	private ExecutorService threadPool;
	private Object lock = new Object();
	private int numRunning;
	
	public ParallelNSGA() {
		super();

		//Set defaults
		params.put(PARAM_SWAP_SOLS_AFTER_ITERATIONS, 50);
		params.put(PARAM_ALG_CONSTRAINT_HANDLING + 1,  ConstraintHandling.REPAIR);
		params.put(PARAM_ALG_CONSTRAINT_HANDLING + 2,  ConstraintHandling.CONSTRAIN_DOMINANCE);
		params.put(PARAM_POP_SWAP_REGIME, PopSwapRegime.ALL_FEASIBLE);
//		params.put(PARAM_NUM_CYCLES, 2);
//		params.put(PARAM_REUSE_NONDOMINATED_ONLY, true);
//		params.put(PARAM_ALGORITHMS, Collections.EMPTY_LIST);
		
		constraintHandlingUsed = new Vector<ConstraintHandling>(2);
		runners = new Vector<NSGARunner>();
		threadPool = Executors.newCachedThreadPool();
	}
	
        @Override
	public void initialise(EmbeddedSystemProblem problem, PropertyMap externalParams, Collection<Architecture> partialInitPop, PopulationGenerator popGenerator) {
		super.initialise(problem, externalParams, partialInitPop, popGenerator);
		
		log.info(Conf.instance());
		
		swapPeriod = params.getInt(PARAM_SWAP_SOLS_AFTER_ITERATIONS);
		swapRegime = (PopSwapRegime) params.get(PARAM_POP_SWAP_REGIME);
			
//		nCycles = params.getInt(PARAM_NUM_CYCLES);
//		reuseNondominatedSolsOnly = params.getBoolean(PARAM_REUSE_NONDOMINATED_ONLY);
//		algClasses = params.getClassList(PARAM_ALGORITHMS);
		//FIXME Hack to reuse some of copied Hybrid code
		algClasses = Arrays.asList( new Class<?>[]{NSGA.class, NSGA.class} );

		constraintHandlingUsed.clear(); 
		for (int algID = 1; params.get(PARAM_ALG_CONSTRAINT_HANDLING + algID) != null; algID++)
			constraintHandlingUsed.add( (ConstraintHandling) params.get(PARAM_ALG_CONSTRAINT_HANDLING + algID) );

		//Note two guarded sections here that assume that the algorithms will be reused, so avoid creating them or the NSGARunners again
		if (algs == null)
			instantiateAlgorithms();

		PropertyMap nsgaParams = algParams.get(0); //both using same parameters, so will manually alter the constraint handling

		for (int i = 0; i < algs.size(); i++) {
			NSGA alg = algs.get(i);
			nsgaParams.put(NSGA.PARAM_CONSTRAINT_HANDLING, constraintHandlingUsed.get(i));
			alg.initialise(problem, nsgaParams, null, popGenerator);
			if (runners.size() < algs.size())
				runners.add( new NSGARunner(alg) );
		}

		//FIXME Probably need a neater way to choose an appropriate number of "iterations" for this algorithm
		//Not needed as have redefined hasMoreIterations() to base it on subordinate NSGA's
//		int nsgaNIters = nsgaParams.getInt(Algorithm.PARAM_MAX_ITERATIONS);
//		int parallelIters = nsgaNIters / swapPeriod;
//		maxIterations = nsgaNIters % swapPeriod == 0 ? parallelIters : parallelIters + 1;
	}
	
	public synchronized void changePopulation(Collection<Architecture> newPopulation) {
		throw new UnsupportedOperationException();
//		currAlg.changePopulation(newPopulation);
	}
	
        @Override
	public synchronized void injectSolutions(Collection<Architecture> newSolutions) {
		throw new UnsupportedOperationException();
//		currAlg.injectSolutions(newSolutions);
	}
	
        @Override
	public synchronized void changeParameters(PropertyMap params) {
		throw new UnsupportedOperationException("Not implemented yet");
//		super.changeParameters(params);
//		
//		//TODO If changing swapEachIteration to true, some algorithms might not yet have been init()ialised.
//		swapEachIteration = params.getBoolean(PARAM_SWAP_SOLS_AFTER_ITERATIONS);
//		nCycles = params.getInt(PARAM_NUM_CYCLES);
//		reuseNondominatedSolsOnly = params.getBoolean(PARAM_REUSE_NONDOMINATED_ONLY);
//		
//		//Cannot change list of algorithms at runtime (not yet at least)
//
//		//FIXME Warning: If two different, controlled algorithms have parameters with the same name, this will use the same value for both of them (if in the new parameters collection)
//		for (int i = 0, c = algs.size(); i < c; i++) {
//			algParams.get(i).addAll(params);
//			algs.get(i).changeParameters(params);
//		}
	}
	
	protected void instantiateAlgorithms() {
		Conf conf = Conf.instance();
		if (algs == null) {
			algs = new Vector<NSGA>();
			algParams = new Vector<PropertyMap>();
		}
		for (Class<?> algClass : algClasses) {
			try {
				algs.add( (NSGA) algClass.newInstance() );
				algParams.add( conf.getApplicableParameters(algClass) );
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		}
	}
	
	public boolean hasMoreIterations() {
		//More general, just in case they have differing number of iterations; end when first one is done
		for (Algorithm alg : algs)
			if (! alg.hasMoreFunctionEvaluations())
				return false;
		return true;
	}
	
	@Override
	public void oneIteration() {
		assert numRunning == 0;

		for (NSGARunner runner : runners)
			startAlg(runner, swapPeriod);
		
		synchronized(this) {
			while (numRunning > 0) {
				try {
					wait();
				} catch (InterruptedException ie) {
					if (log.isWarnEnabled())
						log.warn("Interrupted while waiting for processing threads to finish executing");
				}
			}
		}
		assert numRunning == 0;
		
		//For interest and possibly to influence how long each algorithm is run (perhaps one could have more iterations but take the same time)
		
			String runnerTimesMsg = "";
			for (int i = 0; i < runners.size(); i++)
				runnerTimesMsg += "\t...with " + constraintHandlingUsed.get(i) + ": " + runners.get(i).getElapsedTime() + "ms";
			log.info("Time for " + swapPeriod + " iterations of NSGA" + runnerTimesMsg);
		

		int minFeasible = determineMinFeasibleCount();
		log.info("After " + algs.get(0).currIteration + " iterations minimum number of feasible solutions is " + minFeasible);

		if (hasMoreIterations()) {
			swapPopulations(minFeasible);
			//Some duplication of above activity here, but done infrequently and is not symmetrical as above
			mergeAlgorithmFronts();
		}
		
		//Commented out to match behaviour of NSGA, which currently has disabled recording of HV over time
		//FIXME Should be asking the problem, not the configuration
//		if (Conf.instance().getEvaluatorsCount() > 1) {
//			//FIXME Should eventually be changed to use new hypervolume calculator that works with more than two objectives, accessible via getHyperVolume (sic)
//			double hv = HypervolumeDevelopment.getInstance().getHypervolume(nondominatedSolutions);
//			HypervolumeDevelopment.getInstance().addHypervolumeForIteration(hv, currIteration);
//		}
		
	}
	
        @Override
	public void finalise(Log log) {
		super.finalise(log);
		mergeAlgorithmFronts();
		printCurrentNondominatedSolutions("Final");
	    double hv = HypervolumeDevelopment.getInstance().getHypervolume(nondominatedSolutions);
        HypervolumeDevelopment.getInstance().addHypervolumeForRun(hv);
        
		for (Algorithm alg : algs)
			alg.finalise(log);
	}

	private int determineMinFeasibleCount() {
		int minFeasible = Integer.MAX_VALUE;
		for (NSGA alg : algs)
			minFeasible = Math.min(minFeasible, alg.getFeasibleCount());
		return minFeasible;
	}
	
	private void swapPopulations(final int minFeasible) {
		switch (swapRegime) {
		case ALL_FEASIBLE: swapFirstNSolutions(minFeasible); break;
		case RANDOMLY_REALLOCATE: randomlyReallocateIndividuals(); break;
		}
	}
	
	//FIXME Currently requires that exactly two algorithms be used
	private void swapFirstNSolutions(final int n) {
		if (n > 0) {
			List<Architecture> pop1 = algs.get(0).getPopulation();
			List<Architecture> pop2 = algs.get(1).getPopulation();
			List<Architecture> toPop1 = new ArrayList<Architecture>(n);
			List<Architecture> toPop2 = new ArrayList<Architecture>(n);

			toPop1.addAll( pop2.subList(0, n) );
			toPop1.addAll( pop1.subList(n, pop1.size()) );
			toPop2.addAll( pop1.subList(0, n) );
			toPop2.addAll( pop2.subList(n, pop2.size()) );

			//TODO Do they need to do any housekeeping or just let them run on the new solutions as they are?
			pop1.clear();
			pop1.addAll(toPop1);
			pop2.clear();
			pop2.addAll(toPop2);
		}
	}
	
	private void randomlyReallocateIndividuals() {
		Random rand = Conf.instance().random;
		List<Architecture> pop1 = algs.get(0).getPopulation();
		List<Architecture> pop2 = algs.get(1).getPopulation();
		final int pop1TargetSize = pop1.size(), pop2TargetSize = pop2.size();
		
		List<Architecture> mergedPop = new ArrayList<Architecture>(pop1TargetSize + pop2TargetSize);
		mergedPop.addAll(pop1);
		mergedPop.addAll(pop2);
		pop1.clear();
		pop2.clear();
		
		int i = 0;
		while (pop1.size() < pop1TargetSize && pop2.size() < pop2TargetSize) {
			//Pick either pop1 or pop2
			(rand.nextDouble() < 0.5 ? pop1 : pop2).add( mergedPop.get(i++) );
		}
		//Fill in remaining gaps in less full population
		while (pop1.size() < pop1TargetSize)
			pop1.add( mergedPop.get(i++) );
		while (pop2.size() < pop2TargetSize)
			pop2.add( mergedPop.get(i++) );
		
		for (NSGA alg : algs)
			alg.fastNonDominatedSort();
	}
	
	private void mergeAlgorithmFronts() {
		population.clear();
		for (Algorithm alg : algs)
			extractFirstFront(alg.getPopulation(), population);
		//This next step is not entirely safe, since the implementation of
		//extractNondominatedFront is not stateless (it uses some persistent
		//structures in NSGA) but it should be OK as algs have stopped running.
		algs.get(0).extractNondominatedFront(population,
				(LinkedList<Architecture>) nondominatedSolutions,
				new HashSet<Architecture>(), ConstraintHandling.IGNORE, 1.0);
	}
	
	private void extractFirstFront(List<Architecture> srcPop, List<Architecture> destPop) {
		Iterator<Architecture> solIter = srcPop.iterator();
		Architecture aSol = solIter.next();
		while (aSol != null && aSol.isFeasible() && aSol.getRank() == 1) {
			destPop.add(aSol);
			aSol = solIter.hasNext() ? solIter.next() : null;
		}
	}
	
	public void startAlg(NSGARunner runner, int iterationsToRun) {
		synchronized ( lock ) {
			runner.iterationsToRun = iterationsToRun;
			numRunning++;
		}
		threadPool.execute( runner );
	}
	
	protected class NSGARunner implements Runnable {
		int iterationsToRun; 
		Algorithm algToRun;
		long start, end;
		
		public NSGARunner(Algorithm algToRun) {
			this.algToRun = algToRun;
		}
		
		public long getElapsedTime() { return end - start; }

		public void run() {
			start = System.currentTimeMillis();
			for (int i = 0; i < iterationsToRun && algToRun.hasMoreFunctionEvaluations(); i++)
				algToRun.oneIteration();
			
			synchronized (lock) {
				end = System.currentTimeMillis();
				numRunning--;
				synchronized (ParallelNSGA.this) {
					ParallelNSGA.this.notify();
				}
			}
		}
		
	}

}

