package archeopterix.algorithm.hillclimbing;

import archeopterix.algorithm.Algorithm;
import java.util.List;
import java.util.LinkedList;
import java.util.Random;
import java.util.BitSet;


import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.util.PropertyMap;
import archeopterix.conf.Conf;
import archeopterix.problem.Component;
import archeopterix.problem.Host;
import archeopterix.solution.Architecture;
import archeopterix.problem.deployment.solution.ECUBasedDeployment;
import archeopterix.solution.OutOfOptionsException;

public class HillClimbing extends Algorithm
{

	private static final Log log = LogFactory.getLog(HillClimbing.class);
	public static final String HILLCLIMBING_TYPE = "hillclimber";
	public static final String SOLS_TO_HILLCLIMB = "howManyToHillclimb";
	public static final String TIMEOUT = "timeout";
	public static final String PAR_HILLCLIMBING = "useHillclimbing";
	//  public static final String PAR_TIMEOUT= "timeout";
	// Algorithm parameters
	private int timeout = 30; // if we can't do a feasible move in that time, give up
	private int evalCount;
	private int noImprovingMoves;
	private int noUnsuccessfulMoves;
	private int hcEvaluations;
	double successRatio;
	LinkedList<Architecture> currNondominated;
	//protected ECUBasedDeployment startsolution, solution = null;
	// for sorting the front
	public final HillClimbingType hcType;
	// Random variable
	private Random rnd;
	private EmbeddedSystemProblem problem;
	private int howManyToHillclimb;

	public HillClimbing(EmbeddedSystemProblem problem, PropertyMap newParams)
	{
		this.problem = problem;
		evalCount = 0;
		rnd = Conf.instance().random;

		noImprovingMoves = 0;
		noUnsuccessfulMoves = 0;
		hcEvaluations = 0;
		hcType = (HillClimbingType) newParams.get(HILLCLIMBING_TYPE);
		howManyToHillclimb = newParams.getInt(SOLS_TO_HILLCLIMB);
		timeout = newParams.getInt(TIMEOUT);
		successRatio = 0;
	}

	public void doHillclimbing(List<Architecture> solsToHillclimb, List<Architecture> compareToList)
	{
		ECUBasedDeployment deployment = null;

		for (int i = 0; i < solsToHillclimb.size(); i++)
		{
			int attempts = 0;

			//BitSet alreadyTried = new BitSet(problem.getComponentCount());

			while (attempts < timeout)
			//trying as many times as we have components should be enough
			{
				try
				{
					deployment = doOneMove((ECUBasedDeployment) solsToHillclimb.get(i));
					if (hasBeenAddedToNondominated(deployment, compareToList))
					{
						attempts = 0;
						noImprovingMoves++;
					} else
					{
						//eventually, the hill climber will terminate because further improvement is not possible
						attempts++;
						noUnsuccessfulMoves++;
					}

				} catch (Exception e)
				{
					if (log.isDebugEnabled())
					{
						log.debug("Hill Climber can't find possible move.");
					}
					break;
				}
			}


		}
	}

	public ECUBasedDeployment doOneMove(ECUBasedDeployment previousSolution) throws Exception
	{
		boolean foundNewSol = false;
		ECUBasedDeployment solution = null;
		int attempts = 0;

		while (!foundNewSol)
		{
			switch (hcType)
			{
				case SOPHISTICATED:
					solution = newSolutionThroughFollowOnMoves(previousSolution);
					break;
				case SIMPLE: //do not attempt to shift anything out of the way
					solution = newSolutionThroughSimpleReassignment(previousSolution);
					break;
				case SWAP: //swap two components if possible
					solution = newSolutionThroughSwapMove(previousSolution);
					break;
			}

			if (solution == null)
			{
				attempts++;
			} else
			{
				foundNewSol = true;
				problem.evaluateAttributes(solution);
				hcEvaluations++;
				/*solution.setFeasible(problem.validate(solution));

				if (!solution.isFeasible())
				{
				System.out.println("INFEASIBLE");
				foundNewSol = false;
				attempts++;
				}*/

			}
			if (attempts > 2)
			{
				
				throw new Exception();
			}
		}

//        updateNondominatedFront(nondominatedSolutions, solution);
		return solution;

	}

	/*@Override
	public void finalise()
	{
	super.finalise();
	//FIXME Probably not a task that need be done by the algorithm itself; finalise() method only created to have a placeholder for this code, so potentially could be removed
	printCurrentNondominatedSolutions("Final");
	}*/
	public ECUBasedDeployment newSolutionThroughFollowOnMoves(ECUBasedDeployment previousSolution)
	{
// Select a component randomly for relocation
		ECUBasedDeployment solution = (ECUBasedDeployment) previousSolution.clone();
		Component comp = null;
		Host baseEcu = null;
		//find the component through picking a random Host and then choosing a random component on this Host.
		//not all ECUs have assigned components, so repeat.
		while (comp == null)
		{
			baseEcu = problem.getECU(rnd.nextInt(problem.getECUCount()));
			List<Component> compList = solution.getComponentListForHost(baseEcu.index, problem);
			if (!compList.isEmpty()){
                            comp = compList.get(rnd.nextInt(compList.size()));
			}
		}


		// get the list of possible ECUs as well as the
		// base Host of this component (using the solution)
		// and a randomly selected target Host for relocation
		Host targetEcu = null;
		List<Host> possibleEcus = comp.getCanUseHosts();
		do
		{
			targetEcu = possibleEcus.get(rnd.nextInt(comp.getCanUseHostListLength()));
		} while (targetEcu == baseEcu);

		int sizeCompsOnTarget = 0;

		List<Component> compsOnBase = solution.getComponentListForHost(baseEcu.index, problem);


		List<Component> compsOnTarget = solution.getComponentListForHost(targetEcu.index, problem);


		if (log.isDebugEnabled())
		{
			log.debug("\nBase ECU: " + baseEcu.index);
			for (Component c : compsOnBase)
			{
				log.debug("Component on base: " + c.index);
			}

			log.debug("\nTarget ECU: " + targetEcu.index);
			for (Component c : compsOnTarget)
			{
				log.debug("Component on target: " + c.index + "\n");
			}
		}
		// try to shift conflicting components somewhere else

		for (Component c : compsOnTarget)
		{
			sizeCompsOnTarget += c.getIntProperty(Component.MEMORY);
		}
		List<Component> remainingOnTarget = new LinkedList<Component>();

		for (Component c : compsOnTarget)
		{
			if (!comp.canColocateWith(c))
			{
				try
				{
					Host newECU = findNewHomeFor(solution, c, targetEcu);
					log.debug("Moved enemy: " + c.index + " to ECU " + newECU.index);
					sizeCompsOnTarget -= c.getIntProperty(Component.MEMORY);
					solution.getComponentListForHost(newECU.index, problem).add(c);
				} catch (OutOfOptionsException ex)
				{
					return null;
				}
			} else
			{
				remainingOnTarget.add(c);
			}
		}

		solution.assignComponentList(targetEcu.index, remainingOnTarget);

		compsOnBase.remove(comp);

		sizeCompsOnTarget += comp.getIntProperty(Component.MEMORY);

		// try to move the target component to the target Host
		// if this is not possible due to memory constraints try to move components
		// from the target Host to other ECUs
		while (true)
		{
			if (sizeCompsOnTarget > targetEcu.getIntProperty(Host.MEMORY))
			{

				if (remainingOnTarget.isEmpty())//this means the initial compBaseToTarget by itself is too big for the target
				{
					//if (log.isDebugEnabled())
					log.debug("Comp to shift is too big for target");
					return null;
				} else
				{
					//try all options
					BitSet doneBefore = new BitSet(remainingOnTarget.size());
					int next = 0;

					if (doneBefore.cardinality() == doneBefore.size())
					{
						return null;
					}
					do
					{
						next = rnd.nextInt(remainingOnTarget.size());
					} while (doneBefore.get(next));
					Component compToMove = remainingOnTarget.get(next);

					Host newPlacement;
					try
					{
						newPlacement = findNewHomeFor(solution, compToMove, targetEcu);
						log.debug("Moved component: " + compToMove.index + " to ECU " + newPlacement.index);
					} catch (OutOfOptionsException ex)
					{
						return null;
					}
					remainingOnTarget.remove(compToMove);
					sizeCompsOnTarget -= compToMove.getIntProperty(Component.MEMORY);
					solution.getComponentListForHost(newPlacement.index, problem).add(compToMove);
				}

			} else
			{
				remainingOnTarget.add(comp);
				if (log.isDebugEnabled())
				{
					log.debug("After checks, we have: " + baseEcu.index);
					log.info("Base ECU: " + baseEcu.index);
					for (Component c : compsOnBase)
					{
						log.info("Component on base: " + c.index);
					}

					log.info("\nTarget ECU: " + targetEcu.index);
					for (Component c : remainingOnTarget)
					{
						log.info("Component on target: " + c.index);
					}
				}
				return solution;
			}

		}
	}

	/*
	 * Choose one Host as origin to move components away from. Choose random target Host.
	 * If not enough memory on target, move random component from target to random other Host.
	 */
	public ECUBasedDeployment newSolutionThroughSimpleReassignment(ECUBasedDeployment previousSolution)
	{
// Select a component randomly for relocation
		ECUBasedDeployment solution = (ECUBasedDeployment) previousSolution.clone();
		Component comp = null;
		Host baseEcu = null;
		//find the component through picking a random Host and then choosing a random component on this Host.
		//not all ECUs have assigned components, so repeat.
		while (comp == null)
		{
			baseEcu = problem.getECU(rnd.nextInt(problem.getECUCount()));
			List<Component> compList = solution.getComponentListForHost(baseEcu.index, problem);
			if (compList.size() != 0)
			{
				comp = compList.get(rnd.nextInt(compList.size()));
			}
		}

		List<Component> compsOnBase = solution.getComponentListForHost(baseEcu.index, problem);

		//boolean foundNewHost = false;
		List<Host> possibleEcus = comp.getCanUseHosts();
		BitSet targetsConsidered = new BitSet(comp.getCanUseHostListLength());
		targetsConsidered.set(baseEcu.index);

		while (true)
		{
			if (targetsConsidered.cardinality() == comp.getCanUseHostListLength())
			{
				return null; //have tried all possible
			}
			Host targetEcu = null;
			do
			{
				targetEcu = possibleEcus.get(rnd.nextInt(comp.getCanUseHostListLength()));
			} while (targetsConsidered.get(targetEcu.index));

			targetsConsidered.set(targetEcu.index);

			int sizeCompsOnTarget = 0;

			List<Component> compsOnTarget = solution.getComponentListForHost(targetEcu.index, problem);

			if (log.isDebugEnabled())
			{
				log.debug("\nBase ECU: " + baseEcu.index);

				for (Component c : compsOnBase)
				{
					log.debug("Component on base: " + c.index);
				}

				log.debug("\nTarget ECU: " + targetEcu.index);
				for (Component c : compsOnTarget)
				{
					log.debug("Component on target: " + c.index + "\n");
				}
			}
			// try to shift conflicting components somewhere else

			boolean impossibleAssignment = false;
			for (Component c : compsOnTarget)
			{
				if (!comp.canColocateWith(c))
				{
					impossibleAssignment = true;
				}
				sizeCompsOnTarget += c.getIntProperty(Component.MEMORY);
			}

			if (impossibleAssignment)
			{
				continue;
			}

			sizeCompsOnTarget += comp.getIntProperty(Component.MEMORY);
			if (sizeCompsOnTarget > targetEcu.getIntProperty(Host.MEMORY))
			{
				continue;
			}

			compsOnBase.remove(comp);

			compsOnTarget.add(comp);
			if (log.isDebugEnabled())
			{
				log.debug("After checks, we have: " + baseEcu.index);
				log.info("Base ECU: " + baseEcu.index);
				for (Component c : compsOnBase)
				{
					log.info("Component on base: " + c.index);
				}

				log.info("\nTarget ECU: " + targetEcu.index);
				for (Component c : compsOnTarget)
				{
					log.info("Component on target: " + c.index);
				}
			}
			return solution;
		}
	}

	private Host findNewHomeFor(ECUBasedDeployment deployment, Component comp, Host origin) throws OutOfOptionsException
	{
		BitSet bs = new BitSet(problem.getECUCount());
		bs.set(origin.index);
		int i = 0;
		int sizeCompsOnTarget = 0;
		boolean canUseThis = true;
		while (true)
		{
			if (bs.cardinality() >= problem.getECUCount())
			{
				throw new OutOfOptionsException();
			}
			do
			{
				i = rnd.nextInt(problem.getECUCount());
			} while (bs.get(i));
			bs.set(i);
			canUseThis = true;
			sizeCompsOnTarget = 0;
			List<Component> alreadyAssignedToDest = deployment.getComponentListForHost(i, problem);
			for (Component c : alreadyAssignedToDest)
			{
				if (!comp.canColocateWith(c))//this is a dead end
				{
					canUseThis = false;
					break;
				}
				sizeCompsOnTarget += c.getIntProperty(Component.MEMORY);
			}
			if (!canUseThis)
			{
				continue;
			}
			if (sizeCompsOnTarget + comp.getIntProperty(Component.MEMORY) <= problem.getECU(i).getIntProperty(Host.MEMORY))
			{
				return problem.getECU(i);
			}
		}
	}

	/*
	 * This version moved components from target ECUs only to the origin, if there
	 * was not enough memory on the target

	public ECUBasedDeployment findNewSolution(ECUBasedDeployment previousSolution)
	{
	// Select a component randomly for relocation
	ECUBasedDeployment solution = (ECUBasedDeployment)previousSolution.clone();
	Component compBaseToTarget = null;
	Host baseEcu = null;
	//find the component through picking a random Host and then choosing a random component on this Host.
	//not all ECUs have assigned components, so repeat.
	while (compBaseToTarget == null)
	{
	baseEcu =problem.getECU(rnd.nextInt(problem.getECUCount()));
	List<Component> compList = solution.getComponentListForHost(baseEcu.index, problem);
	if (compList.size() != 0)
	compBaseToTarget = compList.get(rnd.nextInt(compList.size()));
	}


	// get the list of possible Host's as well as the
	// base Host of this component (using the solution)
	// and a randomly selected target Host for relocation
	Host targetEcu = null;
	List<ECU> possibleEcus = compBaseToTarget.getCanUseHosts();
	do{
	targetEcu= possibleEcus.get(rnd.nextInt(compBaseToTarget.getCanUseHostListLength()));
	}while (targetEcu == baseEcu);


	// copy both Host's lists for manipulation
	int sizeCompsOnTarget = 0;
	//int sizeCompsMustMoveFromTarget = 0;
	// List<Component> compsOnBase = new LinkedList<Component>();
	// compsOnBase.addAll(solution.getComponentListForHost(baseEcu.index, problem));
	List<Component> compsOnBase = solution.getComponentListForHost(baseEcu.index, problem);
	int sizeCompsOnBase = 0;
	for(Component c:compsOnBase)
	sizeCompsOnBase+= c.getIntProperty(Component.MEMORY);

	//  List<Component> compsOnTarget = new LinkedList<Component>();
	//  compsOnTarget.addAll(solution.getComponentListForHost(targetEcu.index, problem));
	List<Component> compsOnTarget = solution.getComponentListForHost(targetEcu.index, problem);

	if (log.isDebugEnabled())
	{
	log.debug("\nBase Host: "+baseEcu.index);
	for(Component c: compsOnBase)
	{
	log.debug("Component on base: "+c.index);
	}

	log.debug("\nTarget Host: "+targetEcu.index);
	for(Component c: compsOnTarget)
	{
	log.debug("Component on target: "+c.index);
	}
	}
	// try to shift conflicting components back from the target Host to the base
	// this can fail when the shifting component conflicts with a resident of the base
	for(Component c: compsOnTarget)
	{
	if(!compBaseToTarget.canColocateWith(c))
	{
	for(Component b:compsOnBase)
	{
	if(!c.canColocateWith(b))//if compBaseToTarget can't live with c,
	return null;        //but c can't live with someone on the base ecu, we are stuffed
	else
	{
	compsOnBase.add(c);
	sizeCompsOnBase+= c.getIntProperty(Component.MEMORY);
	}
	}

	}
	else
	sizeCompsOnTarget+= c.getIntProperty(Component.MEMORY);
	}

	compsOnBase.remove(compBaseToTarget);
	// positive colocation has been removed from the deployment problem

	sizeCompsOnTarget += compBaseToTarget.getIntProperty(Component.MEMORY);
	// try to move the target component (and its group) to the target Host
	// if this is not possible due to memory constraints try to move components
	// from the target Host to the base Host
	boolean compMoved = false;
	while(!compMoved )
	{
	if(sizeCompsOnTarget > targetEcu.getIntProperty(Host.MEMORY))
	{
	if (compsOnTarget.isEmpty())//this means the initial compBaseToTarget by itself is too big for the target
	{
	if (log.isDebugEnabled())
	log.debug("Comp to shift is too big for target");
	return null;
	}
	else
	{
	Component compMoveToBase = compsOnTarget.get(rnd.nextInt(compsOnTarget.size()));
	if(compMoveToBase.getIntProperty(Component.MEMORY)+sizeCompsOnBase <= baseEcu.getIntProperty(Host.MEMORY))
	{
	for(Component b:compsOnBase)
	if(!compMoveToBase.canColocateWith(b))
	return null; //too difficult. giving up

	compsOnBase.add(compMoveToBase);
	compsOnTarget.remove(compMoveToBase);
	sizeCompsOnBase+=compMoveToBase.getIntProperty(Component.MEMORY);
	sizeCompsOnTarget-=compMoveToBase.getIntProperty(Component.MEMORY);
	}
	else
	return null; //too difficult; base has no room for comps moved out of the way
	}

	}
	else// if(!compsOnTarget.isEmpty())//how could it be empty if the size of the resident components is bigger than the memory size?
	{
	compMoved=true;
	//do this last
	compsOnTarget.add(compBaseToTarget);
	}
	}

	if(!compMoved)
	return null;

	if (log.isDebugEnabled())
	{
	log.debug("\n After checks, we have: "+baseEcu.index);
	log.debug("Base Host: "+baseEcu.index);
	for(Component c: compsOnBase)
	{
	log.debug("Component on base: "+c.index);
	}

	log.debug("\nTarget Host: "+targetEcu.index);
	for(Component c: compsOnTarget)
	{
	log.debug("Component on target: "+c.index);
	}
	}
	//  solution.assignComponentList(baseEcu.index, compsOnBase);
	// solution.assignComponentList(targetEcu.index, compsOnTarget);
	return solution;
	}
	 */
	/*
	 * SWAP - the equivalent of swap mutation. Attempts to swap two components subject
	 * to compatibility and memory availability.
	 */
	public ECUBasedDeployment newSolutionThroughSwapMove(ECUBasedDeployment previousSolution)
	{
		// Select a component randomly for relocation
		ECUBasedDeployment solution = (ECUBasedDeployment) previousSolution.clone();
		Component compBaseToTarget = null;
		Host baseEcu = null;
		List<Component> compsOnBase = null;
		List<Component> compsOnTarget = null;

		//find the component through picking a random Host and then choosing a random component on this Host.
		//not all ECUs have assigned components, so repeat.
		while (compBaseToTarget == null)
		{
			baseEcu = problem.getECU(rnd.nextInt(problem.getECUCount()));
			compsOnBase = solution.getComponentListForHost(baseEcu.index, problem);
			if (compsOnBase.size() != 0)
			{
				compBaseToTarget = compsOnBase.remove(rnd.nextInt(compsOnBase.size()));
			}

		}


		// get the list of possible Host's as well as the
		// base Host of this component (using the solution)
		// and a randomly selected target Host for relocation
		Host targetEcu = null;
		List<Host> possibleEcus = compBaseToTarget.getCanUseHosts();
		do
		{
			targetEcu = possibleEcus.get(rnd.nextInt(compBaseToTarget.getCanUseHostListLength()));
			compsOnTarget = solution.getComponentListForHost(targetEcu.index, problem);
		} while ((targetEcu == baseEcu) || (compsOnTarget.size() == 0));

		Component compTargetToBase = compsOnTarget.remove(rnd.nextInt(compsOnTarget.size()));

		int sizeCompsOnTarget = 0;
		int sizeCompsOnBase = 0;

		for (Component c : compsOnBase)
		{
			if (!compTargetToBase.canColocateWith(c))
			{
				return null;
			}
			sizeCompsOnBase += c.getIntProperty(Component.MEMORY);
		}

		for (Component c : compsOnTarget)
		{
			if (!compBaseToTarget.canColocateWith(c))
			{
				return null;
			}
			sizeCompsOnTarget += c.getIntProperty(Component.MEMORY);
		}

		sizeCompsOnBase += compTargetToBase.getIntProperty(Component.MEMORY);

		if (sizeCompsOnBase > baseEcu.getIntProperty(Host.MEMORY))
		{
			return null;
		}

		sizeCompsOnTarget += compBaseToTarget.getIntProperty(Component.MEMORY);

		if (sizeCompsOnTarget > targetEcu.getIntProperty(Host.MEMORY))
		{
			return null;
		}

		compsOnBase.add(compTargetToBase);
		compsOnTarget.add(compBaseToTarget);

		return solution;
	}

	public void updateNondominatedFront(List<Architecture> currNondominated, Architecture newSolution)
	{
            currNondominated.add(newSolution);
            if (currNondominated.size() == 1){
                return;
            }

            boolean isNondominated = true;
            int j = 0;
            while (isNondominated && j < currNondominated.size() - 1){
                Architecture domSolution = currNondominated.get(j);
                if (domSolution.isDominated(newSolution)){
                        currNondominated.remove(j);
                } else if (newSolution.isDominated(domSolution)){
                    isNondominated = false;
                    currNondominated.remove(currNondominated.size() - 1);
                } else{
                    j++;
                }
            }
	}

	public double getAverageImprovingMoves(){
            return (double) noImprovingMoves / ((double) noImprovingMoves + (double) noUnsuccessfulMoves);
	}

	public int howManySolsToHillclimb(){
            return howManyToHillclimb;
	}

	public int getNumberOfUsedEvaluations(){
            return hcEvaluations;
	}

	public void addHillClimbingStats(int improvements, int unsuccessful){
            this.noImprovingMoves += improvements;
            this.noUnsuccessfulMoves += unsuccessful;
	}

	protected boolean hasBeenAddedToNondominated(Architecture deployment, List<Architecture> nondominatedSolutions) {
            for (int j = 0; j < nondominatedSolutions.size(); j++) {
                if (deployment.isDominated(nondominatedSolutions.get(j)))
                    return false;
            
            if (nondominatedSolutions.get(j).isDominated(deployment)) {
                nondominatedSolutions.remove(j);
                j--;//TODO::Critical Check
            }
        }
        if (!nondominatedSolutions.contains(deployment)){ //Newly added to avoid duplicates in the pareto fronts
            nondominatedSolutions.add(deployment);
        }
        return true;


    }

    @Override
    public void oneIteration() {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }
}
