/**
 * 
 */
package archeopterix.algorithm.hybrid;

import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Vector;

import archeopterix.algorithm.Algorithm;
import archeopterix.algorithm.PopulationGenerator;
import archeopterix.conf.Conf;
import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.solution.Architecture;
import archeopterix.util.PropertyMap;

/**
 * @author James Montgomery
 *
 */
public class Hybrid extends Algorithm {

	public static final String PARAM_SWAP_EACH_ITERATION = "swapEachIteration";
	public static final String PARAM_NUM_CYCLES = "cycles";
	public static final String PARAM_REUSE_NONDOMINATED_ONLY = "reuseNondominatedSolsOnly";
	public static final String PARAM_ALGORITHMS = "algorithms";
	
	protected boolean swapEachIteration;
	protected int nCycles;
	protected boolean reuseNondominatedSolsOnly;
	protected List<Class<?>> algClasses;
	protected List<Algorithm> algs;
	/**
	 * Only needed to support changing of algorithms' parameters at runtime.
	 */
	protected List<PropertyMap> algParams;
	protected int currAlgIndex;
	protected Algorithm currAlg;
	protected int currCycle;
	
	public Hybrid() {
		super();

		//Set defaults
		params.put(PARAM_SWAP_EACH_ITERATION, false);
		params.put(PARAM_NUM_CYCLES, 2);
		params.put(PARAM_REUSE_NONDOMINATED_ONLY, true);
		params.put(PARAM_ALGORITHMS, Collections.EMPTY_LIST);
	}
	
        @Override
	public void initialise(EmbeddedSystemProblem problem, PropertyMap externalParams, Collection<Architecture> partialInitPop, PopulationGenerator popGenerator) {
		super.initialise(problem, externalParams, partialInitPop, popGenerator);
		
		swapEachIteration = params.getBoolean(PARAM_SWAP_EACH_ITERATION);
		nCycles = params.getInt(PARAM_NUM_CYCLES);
		reuseNondominatedSolsOnly = params.getBoolean(PARAM_REUSE_NONDOMINATED_ONLY);
		algClasses = params.getClassList(PARAM_ALGORITHMS);

		instantiateAlgorithms();

		currAlgIndex = 0;
		currAlg = algs.get(currAlgIndex);
		//FIXME Decide what to do about the popGenerator
		if (swapEachIteration) {
			int i = 0;
			for (Algorithm alg : algs) {//all must be ready to be swapped at each iteration
				alg.initialise(problem, algParams.get(i++), null, popGenerator);
			}
		} else { //only first should be ready, each successive alg will be initialised when encountered
			currAlg.initialise(problem, algParams.get(currAlgIndex), null, popGenerator);
		}
		
		currCycle = 0;
			
	}

        @Override
	public synchronized void changeParameters(PropertyMap params) {
		super.changeParameters(params);
		
		//TODO If changing swapEachIteration to true, some algorithms might not yet have been init()ialised.
		swapEachIteration = params.getBoolean(PARAM_SWAP_EACH_ITERATION);
		nCycles = params.getInt(PARAM_NUM_CYCLES);
		reuseNondominatedSolsOnly = params.getBoolean(PARAM_REUSE_NONDOMINATED_ONLY);
		
		//Cannot change list of algorithms at runtime (not yet at least)

		//FIXME Warning: If two different, controlled algorithms have parameters with the same name, this will use the same value for both of them (if in the new parameters collection)
		for (int i = 0, c = algs.size(); i < c; i++) {
			algParams.get(i).addAll(params);
			algs.get(i).changeParameters(params);
		}
			
	}
	
	protected void instantiateAlgorithms() {
		Conf conf = Conf.instance();
		if (algs == null) {
			algs = new Vector<Algorithm>();
			algParams = new Vector<PropertyMap>();
		}
		for (Class<?> algClass : algClasses) {
			try {
				algs.add( (Algorithm) algClass.newInstance() );
				algParams.add( conf.getApplicableParameters(algClass) );
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		}
	}
	
	@Override
	public void oneIteration() {
		currAlg.oneIteration();
		
		//TODO if curr algorithm now finished (an iteration or its run) get population and give to next algorithm
		if (swapEachIteration || ! currAlg.hasMoreFunctionEvaluations()) {
			List<Architecture> pop = reuseNondominatedSolsOnly ? currAlg.getNondominatedSet() : currAlg.getPopulation();
			currAlg = algs.get( ++currAlgIndex );
			if (! swapEachIteration) {
				//FIXME But make it also take the previous population as well
				currAlg.initialise(problem, algParams.get(currAlgIndex), null, popGenerator);
			}
			
			this.population = currAlg.getPopulation();
			this.nondominatedSolutions = currAlg.getNondominatedSet();
		}
		
		//No need to incrementIteration() as each iteration is of a controlled
		//  algorithm and hybrid algorithm will repeat until nCycles are completed
	}

	

}
