package archeopterix.algorithm.paco;

import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.LinkedList;
import java.util.Random;
import java.util.BitSet;
import java.util.Vector;


import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


import archeopterix.algorithm.Algorithm;
import archeopterix.algorithm.PopulationGenerator;
import archeopterix.algorithm.Stats;
import archeopterix.algorithm.hillclimbing.HillClimbing;
import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.util.PropertyMap;
import archeopterix.conf.Conf;
import archeopterix.problem.Component;
import archeopterix.problem.Host;
import archeopterix.problem.deployment.solution.ECUBasedDeployment;
import archeopterix.solution.Architecture;
import archeopterix.solution.ArchitectureComparator;
import archeopterix.reporting.HypervolumeDevelopment;
import archeopterix.solution.OutOfOptionsException;
import java.util.HashSet;
import java.util.Set;

public class PACO extends Algorithm {

    private static final Log log = LogFactory.getLog(PACO.class);
    private static final Log resultsLog = LogFactory.getLog(Conf.RESULTS_LOG);
    public static final String PAR_SOL_PER_CYCLE = "solutionsCycle";
    public static final String PAR_PHERO_ADDITION = "pheroAddition";
    public static final String PAR_TRANS_RULE = "transitionRule";
    public static final String PAR_BUILD_PHER_FROM_K = "kSolutionsUpdatePher";
    public static final String PAR_COMP_SEL_ORDER = "compSelOrder";
    // Algorithm parameters
    private int solPerCycle;
    private double pheroAddition, transitionRule;
    private int kSolsToBuildPher;
    private int evalCount;
    private int timeout;
    private CompSelectionOrder compSelOrder;
    private boolean doHillclimbing;
    private HillClimbing hillclimber;
    // Pheromone Matrix
    private PheromoneMatrix pheroMatrix;
    // Solutions in the Non-Dominated Front
    List<Architecture> selectedNondominated;
//	LinkedList<Architecture> currNondominated;
    // for sorting the front
    ArchitectureComparator[] comparator;
    private List<Component> components;
    private List<Component> orderedComponents;
    private List<Host> ecus;

	LinkedList<Architecture> currFront;
	List<Architecture> sortedNondominated;
    /**
     * For each component the total memory size required by it and any other
     * components with which it must be colocated.
     */
    //private int[] compWithGroupSizes;
    // Random variable
    private Random rnd;
    OutOfOptionsException ooex;
    public PACO() {
        super();

        evalCount = Conf.instance().getEvaluatorsCount();
        rnd = Conf.instance().random;
        timeout = 30;

        selectedNondominated = new LinkedList<Architecture>();
//		currNondominated = new LinkedList<Architecture>();

        comparator = new ArchitectureComparator[evalCount];
        for (int i = 0; i < evalCount; i++) {
            comparator[i] = new ArchitectureComparator(i);
        }

		currFront = new LinkedList<Architecture>();
		sortedNondominated = new LinkedList<Architecture>();
    }

    @Override
    public void initialise(EmbeddedSystemProblem problem, PropertyMap newParams,
            Collection<Architecture> partialInitPop, PopulationGenerator popGenerator) {
        super.initialise(problem, newParams, partialInitPop, popGenerator);

		ooex = new OutOfOptionsException("No ECU options because of enemy assignments");
        solPerCycle = params.getInt(PAR_SOL_PER_CYCLE);
        pheroAddition = params.getDouble(PAR_PHERO_ADDITION);
        transitionRule = params.getDouble(PAR_TRANS_RULE);
        kSolsToBuildPher = params.getInt(PAR_BUILD_PHER_FROM_K);
        compSelOrder = (CompSelectionOrder) params.get(PAR_COMP_SEL_ORDER);
        doHillclimbing = params.getBoolean(HillClimbing.PAR_HILLCLIMBING);

        switch (compSelOrder) {
            case RANDOMISED:
                orderedComponents = new Vector<Component>(problem.getComponents()); // in case wish to actually rearrange them
            case DEFAULT:
                orderedComponents = problem.getComponents();
                break;
            case FEWEST_ASSIGNMENTS:
                orderedComponents = problem.getCompsSortedByPossibleAssignments();
                break;
            case FEWEST_ASSIGNMENTS_LARGEST_SIZE:
                orderedComponents = problem.getCompsSortedByPossibleAssignmentsAndSize();
                break;
        }

		initialiseZeroArray(solPerCycle);

        ecus = problem.getECUs();
        pheroMatrix = new PheromoneMatrix(i_Comps, i_ECUs, pheroAddition);

        components = problem.getComponents();
      
        if (doHillclimbing) {
            hillclimber = new HillClimbing(problem, params);
        }
		printCurrentNondominatedSolutions("Intial");
    }

    @Override
    public void oneIteration() {
        ECUBasedDeployment solution = null;
        int attempts = 0;

        if (log.isDebugEnabled()) {
            log.debug(m_Watch.GetElapsed() + ": Itr: " + (currIteration + 1) + " Evals="
                    + Stats.instance().getEvaluationCounter());
        }

        // Initialize Pheromone Matrix
        pheroMatrix.init();
        population.clear();

        if (!nondominatedSolutions.isEmpty()) {
            // choose (k-1)/2 solutions above and below a randomly chosen solution
            findPartialFront(nondominatedSolutions, selectedNondominated);
            assert selectedNondominated.size() > 0 && selectedNondominated.size() <= kSolsToBuildPher;

            // Pheromone Matrix Update
            pheroMatrix.updatePheromones(selectedNondominated);
        }

        // Construct new solutions (using solPerCycle)
        // Use random-proportional rule
        int k = 0;
        int incomplete = 0;

        while (k < solPerCycle) {
            try {
                solution = constructNewSolution();
                problem.evaluateAttributes(solution);
             /*   solution.setFeasible(problem.validate(solution));
                if (!solution.isFeasible()) {
                    System.out.println("INFEASIBLE");
                } else {
//                        System.out.println("feasible");
                    population.add(solution);
                }*/
				population.add(solution);
                attempts = 0;
                k++;
            } catch (OutOfOptionsException oe) {
                incomplete++;
                if (log.isDebugEnabled()) {
                    log.debug(m_Watch.GetElapsed() + "Failed PACO attempt " + oe.getMessage());
                }
                attempts++;
                if (attempts > timeout) {
                    if (log.isInfoEnabled()) {
                        log.info(m_Watch.GetElapsed() + "No solution after " + timeout + " attempts. Giving up.");
                    }
                    System.exit(0);
                }

            }
        }


		//fastNonDominatedSort();
       findNondominatedFront((LinkedList) nondominatedSolutions, population);

		 if (doHillclimbing)
		 {
            doLocalSearch(nondominatedSolutions);
        }

	   if (log.isDebugEnabled()) {
			log.debug(m_Watch.GetElapsed() + ": Itr: " + (currIteration + 1) + " Evals=" + Stats.instance().getEvaluationCounter());
			printCurrentNondominatedSolutions("Iteration :" + (currIteration + 1));
		}
	    double hv = computeBalancedHypervolume(nondominatedSolutions);
		HypervolumeDevelopment.getInstance().addHypervolumeForIteration(hv, getNumberEvaluations());
        incrementIteration();
    }

    @Override
    public void finalise(Log log) {
        super.finalise(log);
        //FIXME Probably not a task that need be done by the algorithm itself; finalise() method only created to have a placeholder for this code, so potentially could be removed
        Collections.sort(nondominatedSolutions, comp);
        printCurrentNondominatedSolutions("Final");
		//double hv = computeHypervolume(nondominatedSolutions)*10;
        //HypervolumeDevelopment.getInstance().addHypervolumeForRun(hv);
        if (doHillclimbing)
        {
            double avgImprovingMoves = hillclimber.getAverageImprovingMoves();
            log.info("Hillclimber made an average " + avgImprovingMoves + " improving moves.");
            log.info("Hillclimber used " + (double) hillclimber.getNumberOfUsedEvaluations()/(double)Stats.instance().getEvaluationCounter()+ " of the evaluations.");
        }
    }

    public ECUBasedDeployment constructNewSolution() throws OutOfOptionsException {
        List<Host> possibleEcus = new LinkedList<Host>();
        ECUBasedDeployment solution = new ECUBasedDeployment(i_ECUs, i_Comps);


        Component comp;
        int ecuIndex = 0;

        BitSet compBits = new BitSet();
        compBits.set(0, i_Comps);

        int[] memConstraints = new int[i_ECUs];
        boolean memOk;

        while (!compBits.isEmpty()) {
            // Select next component to assign using current assignment order
            comp = selectNextComponent(compBits);

            // creating possible Host list
            possibleEcus.clear();
            possibleEcus.addAll(comp.getCanUseHosts());

            // Remove all Host from possible Host List where MustNotColocate
            // components have been assigned already
            for (Component c : comp.getMustNotColocateWith()) {
                if (!compBits.get(c.index)) {
                    possibleEcus.remove(ecus.get(solution.getHostIndex(c.index)));
                }
            }


            if (possibleEcus.isEmpty()) {
                throw ooex;
            }


            memOk = false;
            while (!memOk) {
                // Proportional Rule: roulette Select or Max Select
                if (rnd.nextDouble() >= transitionRule) {
                    ecuIndex = pheroMatrix.rouletteSelect(comp.index, possibleEcus);
                } else {
                    ecuIndex = pheroMatrix.maxSelect(comp.index, possibleEcus);
                }


                //Check Memory constraint
                if (memConstraints[ecuIndex] + comp.getIntProperty(Component.MEMORY) > ecus.get(ecuIndex).getIntProperty(Host.MEMORY)) {
                    if (log.isTraceEnabled()) {
                        log.trace(m_Watch.GetElapsed() + ": " 
                                + " H" + i_ECUs + "_C" + i_Comps + ": Memory constraint violation ...");
                    }
                    possibleEcus.remove(ecus.get(ecuIndex));
                    if (possibleEcus.isEmpty()) {
                        if (log.isTraceEnabled()) {
                            log.trace(m_Watch.GetElapsed() + ": "
                                    + " H" + i_ECUs + "_C" + i_Comps + ": Solution Creation Error! No possible ECU's");
                        }

                        throw ooex;
                    }

                } else {
                    memConstraints[ecuIndex] += comp.getIntProperty(Component.MEMORY);
                    memOk = true;
                }
            }

            // Check if Host assignments not possible

            solution.assignComponent(comp, ecuIndex);
            if (log.isTraceEnabled()) {
                log.trace("Assigned component " + comp.index + " to ECU " + ecuIndex);
            }

            compBits.clear(comp.index);

            //Deprecated - groups no longer used.
            // Set all MustColocate components for the selected Host
          /*  if(comp.mustColocate())
            {
            for(Component c: comp.getMustColocateWith())
            if(compBits.get(c.index))
            {
            solution.set(c.index, ecuIndex);
            compBits.clear(c.index);
            }
            }*/
        }
        if (log.isTraceEnabled()) {
            log.trace("\n");
        }
        return solution;
    }

    protected Component selectNextComponent(BitSet available) {
        switch (compSelOrder) {
            case DEFAULT:
                return orderedComponents.get(available.nextSetBit(0)); //available indices are aligned with those in components list
            case FEWEST_ASSIGNMENTS:				//|-components' positions are not the same as their indices
            case FEWEST_ASSIGNMENTS_LARGEST_SIZE:	//|
                for (int chosen = 0; chosen < i_Comps; chosen++) {
                    if (available.get(orderedComponents.get(chosen).index)) {
                        return orderedComponents.get(chosen);
                    }
                }
            case RANDOMISED:
                return orderedComponents.get(getRndIndex(available));
        }
        return null; //Not possible to reach this point unless available.isEmpty()
    }

    protected int getRndIndex(BitSet available) {
        int index = -1;
        int bitCount = available.cardinality();
        if (bitCount != 0) {
            for (int i = 0, target = rnd.nextInt(bitCount); i <= target; i++) {
                index = available.nextSetBit(index + 1);
            }
        }
        return index;
    }

    public void findPartialFront(List<Architecture> nondominatedSolutions,
            List<Architecture> selectedNondominated) {
        Collections.sort(nondominatedSolutions, comparator[rnd.nextInt(evalCount)]);

        final int solutionsCount = nondominatedSolutions.size();
        final int k = Math.min(solutionsCount, kSolsToBuildPher);
        final int range = (k - 1) / 2;
        int intRnd = rnd.nextInt(solutionsCount);
        int iLower = Math.max(0, intRnd - range);
        int iUpper = Math.min(solutionsCount, intRnd + range + 1);

        selectedNondominated.clear();
        selectedNondominated.addAll(nondominatedSolutions.subList(iLower, iUpper));
    }

//this can add to an existing list of nondominated solutions
    public void findNondominatedFront(LinkedList<Architecture> currNondominated,
            List<Architecture> newSolutions) {
        boolean isNondominated;
        Architecture solution, domSolution;
        int i = 0;
        int j = 0;
        if (newSolutions.isEmpty()) {
            return;
        }

        if (currNondominated.isEmpty()) {
            currNondominated.add(newSolutions.get(0));
            i = 1;
        }

        for (; i < newSolutions.size(); i++) {
            solution = newSolutions.get(i);
            currNondominated.add(solution);
            isNondominated = true;
            j = 0;

            while (isNondominated && j < currNondominated.size() - 1) {
                domSolution = currNondominated.get(j);
                if (domSolution.isDominated(solution)) {
                    currNondominated.remove(j);
                } else if (solution.isDominated(domSolution)) {
                    isNondominated = false;
                    currNondominated.removeLast();
                } else {
                    j++;
                }
            }
        }

    }
/*
    private void hillclimbSolutions(List<Architecture> deployments) {
        ECUBasedDeployment deployment;

		if (hillclimber.howManySolsToHillclimb() < deployments.size())
		{
			BitSet doneBefore = new BitSet(deployments.size());
			List<Architecture> newList = new LinkedList<Architecture>();
			int solToHillclimb =0;
			while (true)
			{
				if (doneBefore.cardinality() > hillclimber.howManySolsToHillclimb())
				{
					return;
				}
				do{
					solToHillclimb = rnd.nextInt(deployments.size());
				}while (doneBefore.get(solToHillclimb));
				doneBefore.set(solToHillclimb);
				newList.add(deployments.remove(solToHillclimb));
				//hillclimber.doHillclimbing(newList);
			}
		}

        for (int i = 0; i < deployments.size(); i++) {
            try {
                deployment = (ECUBasedDeployment) deployments.get(i);
                if (log.isDebugEnabled()) {
                    log.debug("Deployment before hill climbing " + deployment);
                }
                deployment = hillclimber.hillClimbSolution(deployment);
                if (log.isDebugEnabled()) {
                    log.debug("Deployment after hill climbing " + deployment);
                }
            } catch (Exception ex) {
                if (log.isDebugEnabled()) {
                    log.debug("No improvement possible using solution " + i);
                }
            }
        }
       
    }*/

	private void fastNonDominatedSort() {
		//the nondominated set is not in the population; if not added, they are lost
		population.addAll(nondominatedSolutions);
    	currFront.clear();
    	sortedNondominated.clear();
    	//lstInfeasibles = extractInfeasibles();

    	//int reducedPopSize = Math.min(i_PopulationSize, lst_LocalPopulation.size());
    	Set<Architecture> allFronts = new HashSet<Architecture>();

    	System.arraycopy(allZeroes, 0, frontSizes, 0, frontSizes.length);
    	int rank = 0;
    	while (sortedNondominated.size() < solPerCycle) {
    		rank++; //this will be 1 on first read
    		currFront.clear();
    		extractNondominatedFront(population, currFront, allFronts);

    		int currFrontSize = currFront.size();

    		for (Architecture sol : currFront) {
    			sol.setRank(rank);
    		}

    		//Currently if current front will fit into population, add it, otherwise use
    		//  crowding measure to select from this, final front to be kept.
    		//  Note: If only the 1st ranked front fits, currently that is all that will
    		//  be kept; we must make additional arrangements if we would like to
    		//  ensure > 1 front at all times.
    		if (sortedNondominated.size() + currFront.size() <= solPerCycle) {
    			sortedNondominated.addAll(currFront);
    			allFronts.addAll(currFront);


//				CrowdingUtils.assignCrowdingDistance(lstSortedNondominated,
//					lstSortedNondominated.size() - currFrontSize, lstSortedNondominated.size() - 1, m_Context);
    		} else {
    			//NSGA-II's crowding fill
//				CrowdingUtils.assignCrowdingDistanceAndSort(lstCurrFront, m_Context);
    			currFrontSize = solPerCycle - sortedNondominated.size(); //will only have this many in this front eventually
    			for (int i = 0; i < currFrontSize; i++) {
    				sortedNondominated.add(currFront.get(i));
    			}

    			//Below is older version that randomly selects individuals from final front
//				currFrontSize = i_PopulationSize - lstSortedNondominated.size(); //will only have this many in this front eventually
//				while ( lstSortedNondominated.size() < i_PopulationSize ) {
//					Architecture sol = lstCurrFront.get( rnd.nextInt(currFrontSize) );
//					if ( ! allFronts.contains(sol) ) {
//						allFronts.add(sol);
//						lstSortedNondominated.add(sol);
//					}
//				}
    		}
    		frontSizes[rank - 1] = currFrontSize;
    	}

    	/*         if (lstSortedNondominated.size() < i_PopulationSize)
            {
            rank++;
            iaFrontSizes[rank-1] = i_PopulationSize - lstSortedNondominated.size();
            for (int i = 0; i < iaFrontSizes[rank-1]; i++)
            {
            lstInfeasibles.get(i).setRank(rank);
            lstSortedNondominated.add(lstInfeasibles.get(i));
            }

            }*/

    	
    	nondominatedSolutions.clear();
    	nondominatedSolutions.addAll(sortedNondominated.subList(0, frontSizes[0]));


    }


    /**
     * Finds the nondominated front in {@code lstSourcePop \ otherFronts} and
     * places these solutions in {@code lstNondominated}, which is cleared
     * first.
     * @param lstSourcePop The source population from which to extract the
     * nondominated front; this remains unchanged.
     * @param lstNondominated The resultant nondominated front.
     * @param otherFronts A {@code Set} containing solutions that will not be
     * considered for inclusion in the generated nondominated front; this set
     * is not modified.
     */
    public void extractNondominatedFront(List<Architecture> lstSourcePop,
    		LinkedList<Architecture> lstNondominated, Set<Architecture> otherFronts)
    {
    	lstNondominated.clear();

    	if (lstSourcePop.isEmpty())
    		return;

    	//Everything here could be done with iterators, but this results in many Iterator objects being created, so making careful use of indices instead
    	int first = 0;
    	Architecture firstInclusion;
    	do {
    		firstInclusion = lstSourcePop.get(first);
    		first++;
    	} while (otherFronts != null && otherFronts.contains(firstInclusion));
    	lstNondominated.add(firstInclusion);

    	final int popSize = lstSourcePop.size(); //saves method call overhead
    	for (int i = first; i < popSize; i++)
		{
    		Architecture sol = lstSourcePop.get(i);
    		//Deb's pseudocode includes a condition on the loop that should never occur, but just in case...
    		assert !lstNondominated.contains(sol);

    		if (otherFronts == null || !otherFronts.contains(sol))
			{ //consider this solution as per usual dominance check
    			lstNondominated.add(sol);

    			boolean isNondominated = true;
    			int j = 0;
    			while (isNondominated && j < lstNondominated.size() - 1) { //temporarily-added solution is guaranteed to be at the end
    				Architecture otherSol = lstNondominated.get(j);
    				if ( otherSol.isDominated(sol) ) {
    					lstNondominated.remove(j);
    				} else if ( sol.isDominated(otherSol) ) {
    					isNondominated = false;
    					lstNondominated.removeLast();
    				} else {
    					j++;
    				}
    			}
    		
			}
    
		}
	}

	private void doLocalSearch(List<Architecture> nondominatedSolutions)
	{
		currFront.addAll(nondominatedSolutions);
		boolean added = false;
		int hcSuccessCounter = 0;
		int repeats;
		int attempts = currFront.size();
		for (Architecture architecture : currFront)
		{
			ECUBasedDeployment deployment = null;
			repeats = -1;

			switch(hillclimber.hcType){
				case SWAP:
					while (deployment == null)
					{
						deployment = hillclimber.newSolutionThroughSwapMove((ECUBasedDeployment)architecture);
						repeats++;
					}
					break;
				case SIMPLE:
					while (deployment == null)
					{
						deployment = hillclimber.newSolutionThroughSimpleReassignment((ECUBasedDeployment)architecture);
						repeats++;
					}
					break;
				case SOPHISTICATED:
					while (deployment == null)
					{
						deployment = hillclimber.newSolutionThroughFollowOnMoves((ECUBasedDeployment)architecture);
						repeats++;
					}
					break;

			}
			problem.evaluateAttributes(deployment);
			added = !checkDominance(deployment);
			if (added)
				hcSuccessCounter++;
			attempts += repeats;
			if (!hasMoreFunctionEvaluations())
				return;
		}

		hillclimber.addHillClimbingStats(hcSuccessCounter, (attempts-hcSuccessCounter));
	}
}
