package archeopterix.algorithm.performance;

import archeopterix.evaluation.Attribute;
import archeopterix.solution.Architecture;
import java.util.List;

/*===========================================================================*
 * This class implements the binary epsilon additive indicator as proposed in
 * E. Zitzler, E. Thiele, L. Laummanns, M., Fonseca, C., and Grunert da Fonseca.
 * V (2003): Performance Assesment of Multiobjective Optimizers: An Analysis and
 * Review. The code is the a Java version of the orginal metric implementation 
 * by Eckart Zitzler.
 * Reference: 
 *           E. Zitzler, L. Thiele, M. Laummanns, C.M. Fonseca, and Grunert da Fonseca.
 *           Performance Assesment of Multiobjective Optimizars: An Analysis and
 *           Review 
 *           IEEE Transactions on Evolutionary Computation, vol. 3, no. 4, 
 *           pp. 257-271, 2003.
 */

public class SumEpsilonIndicator {

	/* stores the number of objectives */
	int dim_;
	/* obj_[i]=0 means objective i is to be minimized. This code always suposse
	 * the maximisation of all the objectives
	 */
	int[] obj_;     /* obj_[i] = 0 means objective i is to be minimized */
	/* method_ = 0 means apply additive epsilon and method_ = 1 means multiplicative
	 * epsilon. This code always apply additive epsilon
	 */

	int method_;

	/**
	 * Returns the epsilon indicator.
	 * @param b. True Pareto front
	 * @param a. Solution front
	 * @return the value of the epsilon indicator
	 */
	public double calculateEpsilonValue(List<Architecture> lstFront1, List<Architecture> lstFront2){
		int iF1Size = lstFront1.size();
		int iDim = 0;
		if (iF1Size>0){
			iDim = lstFront1.get(0).getAttributeList().size();
		}else{
			return 0.0; //Front 1 is empty
		}
		double[][] adFront1 = new double[iF1Size][iDim];
		for (int iSol=0; iSol<lstFront1.size(); iSol++){
			List<Attribute> lstAttributes = lstFront1.get(iSol).getAttributeList();
			for(int iAtt=0; iAtt< lstAttributes.size(); iAtt++){
				adFront1[iSol][iAtt] = lstAttributes.get(iAtt).getOptimizerValue();
			}
		}


		int iF2Size = lstFront2.size();
		double[][] adFront2 = new double[iF2Size][iDim];
		for (int iSol=0; iSol<lstFront2.size(); iSol++){
			List<Attribute> lstAttributes = lstFront2.get(iSol).getAttributeList();
			for(int iAtt=0; iAtt< lstAttributes.size(); iAtt++){
				adFront2[iSol][iAtt] = lstAttributes.get(iAtt).getValue();
			}
		}

		return epsilon(adFront1, adFront2, iDim);
	}


	private double epsilon(double[][] b, double[][] a, int dim) {
		int i, j, k;
		double eps, eps_j = 0.0, eps_k = 0.0, eps_temp;
		dim_ = dim;
		set_params();

		if (method_ == 0) {
			eps = Double.MIN_VALUE;
		} else {
			eps = 0;
		}

		for (i = 0; i < a.length; i++) {
			for (j = 0; j < b.length; j++) {
				for (k = 0; k < dim_; k++) {
					switch (method_) {
						case 0:
							if (obj_[k] == 0) {
								eps_temp = b[j][k] - a[i][k];
							} //eps_temp = b[j * dim_ + k] - a[i * dim_ + k];
							else {
								eps_temp = a[i][k] - b[j][k];
							}
							//eps_temp = a[i * dim_ + k] - b[j * dim_ + k];
							break;
						default:
							if ((a[i][k] < 0 && b[j][k] > 0)
									|| (a[i][k] > 0 && b[j][k] < 0)
									|| (a[i][k] == 0 || b[j][k] == 0)) {
								//if ( (a[i * dim_ + k] < 0 && b[j * dim_ + k] > 0) ||
								//     (a[i * dim_ + k] > 0 && b[j * dim_ + k] < 0) ||
								//     (a[i * dim_ + k] == 0 || b[j * dim_ + k] == 0)) {
								System.err.println("error in data file");
								System.exit(0);
							}
							if (obj_[k] == 0) {
								eps_temp = b[j][k] / a[i][k];
							} //eps_temp = b[j * dim_ + k] / a[i * dim_ + k];
							else {
								eps_temp = a[i][k] / b[j][k];
							}
							//eps_temp = a[i * dim_ + k] / b[j * dim_ + k];
							break;
					}
					if (k == 0) {
						eps_k = eps_temp;
					} else if (eps_k < eps_temp) {
						eps_k = eps_temp;
					}
				}
				if (j == 0) {
					eps_j = eps_k;
				} else if (eps_j > eps_k) {
					eps_j = eps_k;
				}
			}
			
			eps = eps + eps_j;
			
		}
		return eps;
	} // epsilon

	/**
	 * Established the parameters by default
	 */
	void set_params() {
		int i;
		obj_ = new int[dim_];
		for (i = 0; i < dim_; i++) {
			obj_[i] = 0;
		}
		method_ = 0;
	} // set_params
} // Epsilon

