/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package archeopterix.algorithm.sa;

import archeopterix.algorithm.Algorithm;
import archeopterix.algorithm.DefaultPopulationGenerator;
import archeopterix.algorithm.PopulationGenerator;
import archeopterix.conf.Conf;
import archeopterix.evaluation.Attribute;
import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.solution.Architecture;
import archeopterix.solution.ConstraintHandling;
import archeopterix.solution.OutOfOptionsException;
import archeopterix.util.PropertyMap;
import java.io.IOException;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;
import java.util.logging.Level;
import java.util.logging.Logger;


import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/** Multiobjective Simulated Annealing Algorithm
 *
 * @author imeedeniya
 */
public class MOSA extends Algorithm {

    public static final String PAR_SAMPLING_TH ="samplingThreshold";
    public static final String PAR_ADDITIONAL_SAMPLES ="additionalSamples";
    public static final String PAR_LENGH_OF_EPOCH ="lengthOfEpoch";
    public static final String PAR_NO_OF_EPOCHS ="noOfEpoches";
    public static final String PAR_ACTIVE_PERIOD ="activePeriod";

    private static final Log log = LogFactory.getLog(MOSA.class);
  
  
    //Algorithm parameters
    protected int samplingThreshold = 30;
    protected int additionalSamples = 100;
    protected int epochLength = 100;
    protected int epochs = 10;
    private int currentEpoch = 0;
    protected double activePeriod = 0.75;
    protected double beta;
    protected double delta;
    

    private     double        currentTemperature = 0.0;
    protected   Architecture  currentSolution;
    protected   SAOperators   operators;
    private     Random rnd;
    public MOSA() {
        super();
        rnd = Conf.instance().random;
        samplingThreshold = 30;
        additionalSamples = 100;
        epochLength = 100;
        epochs = 10;

        activePeriod = 0.75;
        currentSolution = null;
    }

    @Override
    public void initialise(EmbeddedSystemProblem problem, PropertyMap newParams, Collection<Architecture> partialInitPop,
            PopulationGenerator oPopGenerator) {
        super.initialise(problem, newParams, partialInitPop, oPopGenerator);
        operators = SAOperatorsFactory.instance().getSuitableOperators(Conf.instance().getSolutionRepresentation());
        samplingThreshold = params.getInt(PAR_SAMPLING_TH);
        additionalSamples = params.getInt(PAR_ADDITIONAL_SAMPLES);
        epochLength = params.getInt(PAR_LENGH_OF_EPOCH);
        epochs = params.getInt(PAR_NO_OF_EPOCHS);
        activePeriod = params.getDouble(PAR_ACTIVE_PERIOD);

        currentEpoch = 0;
        burnIn();
    }

    @Override
    public synchronized void changeParameters(PropertyMap newParameters) {
        if(newParameters.getInt(PAR_SAMPLING_TH)>0){
            samplingThreshold = newParameters.getInt(PAR_SAMPLING_TH);
        }
        else{
            samplingThreshold = params.getInt(PAR_SAMPLING_TH);
        }
        if(newParameters.getInt(PAR_ADDITIONAL_SAMPLES)>0){
            additionalSamples = newParameters.getInt(PAR_ADDITIONAL_SAMPLES);
        }
        else{
            additionalSamples = params.getInt(PAR_ADDITIONAL_SAMPLES);
        }
        if (newParameters.getInt(PAR_LENGH_OF_EPOCH)>0){
            epochLength = newParameters.getInt(PAR_LENGH_OF_EPOCH);
        }
        else{
            epochLength = params.getInt(PAR_LENGH_OF_EPOCH);
        }
    }
    @Override
    public void oneIteration() {
        if (currIteration > epochLength) {
            printCurrentNondominatedSolutions("Epoch "+ currentEpoch + " Temp="+ currentTemperature);
            currentTemperature = currentTemperature * beta;
            currIteration = 0;
            currentEpoch++;
        }
        try {
            Architecture oNewArch = operators.mutationMove(currentSolution, problem);
            double[] adFitness = {evaluateFitness(currentSolution), evaluateFitness(oNewArch) };
            double dDelta = Math.abs(adFitness[0]- adFitness[1]);
            //Conditionally accept the new solution as current solution
            if (rnd.nextDouble() < Math.exp(-1*dDelta/currentTemperature)){
                currentSolution = oNewArch;
            }
            checkDominance(oNewArch); //Check dominance and add to Archive if nondominated
        } catch (OutOfOptionsException ex) {
            if (log.isInfoEnabled())
            log.debug(m_Watch.GetElapsed()+"No Mutation move for the solution " + ex.toString());
        }
        incrementIteration();
    }

    /**
     * BurnIn process. Create one solution at a time (m_Current).
     * Also computes currentTemperature, beta
     */
    protected void burnIn() {
		try {
			///Fill the archive with 2 non-dominated solutions
			//population = popGenerator.generatePopulation(100, problem,ConstraintHandling.IGNORE);
			//DefaultPopulationGenerator.dumpPopulationToFile(population, "pop.txt");
			population = DefaultPopulationGenerator.ExtractPopulationFromFile(problem, "pop.txt", "au.edu.swin.archeopterix.problem.deployment.solution.ECUBasedDeployment");
		} catch (IOException ex) {
			Logger.getLogger(MOSA.class.getName()).log(Level.SEVERE, null, ex);
		}
		System.exit(0);
        maintainArchiveSize(2);
        Architecture oX = popGenerator.generatePopulation(1, problem,ConstraintHandling.REPAIR).get(0);
        problem.evaluateAttributes(oX);
        currentSolution = oX;
        checkDominance(currentSolution); //Check and add to the archive
 
        int iNoOfNew = 50;
        double dSumFitnessDiff = 0.0;
        for (int i=0; i<iNoOfNew; i++){ // 50 runs
            Architecture oX2 = null;
            while ( oX2 == null ) {
                try {
                     oX2 = operators.mutationMove(currentSolution, problem);
                } catch (OutOfOptionsException ex) {
                    log.debug(m_Watch.GetElapsed()+"No Mutation move for the solution " + ex.toString());
                    continue;
                }
                dSumFitnessDiff = Math.abs(evaluateFitness(oX2));
                checkDominance(oX2);
            }
        }
        currentTemperature = dSumFitnessDiff / ((double) iNoOfNew * Math.log(2.0)); //Average fitness diff / ln(2)
        beta = -1* Math.pow(0.00001* currentTemperature, 1.0/((double)epochs*activePeriod));
        // k*p root of (0.00001* t0)
    }

    /**
     * Computes the domination based fintess for the current solution(x) and
     * a given architecture(oX2), and returs fitness diff (\delta)
     * @param oX2 Candidate Architecture X'
     * @return fitness difference( f(x)-f(x')
     */
    protected double evaluateFitness(Architecture oX2){ // X'
        List<Architecture> lstComparisonSet = new LinkedList<Architecture>();
        lstComparisonSet.addAll(nondominatedSolutions);
        if ( lstComparisonSet.size() < samplingThreshold){
            lstComparisonSet.addAll(sampleAdditionalPoints());
        }
        problem.evaluateAttributes(oX2);
        problem.evaluateAttributes(currentSolution);
        int iNX = 0; //Number of solutions that domiate X(m_current).
        for (Architecture oSol : lstComparisonSet){
            if (currentSolution.isDominated(oSol))
                iNX++;
        }
        if (currentSolution.isDominated(oX2))
            iNX++;
        double dX = (double)iNX/(double)(lstComparisonSet.size() + 1);

        int iNX2 = 0; //Number of solutions that dominate X2 (Candidate)
        for (Architecture oSol : lstComparisonSet){
            if (oX2.isDominated(oSol))
                iNX2++;
        }
        if (oX2.isDominated(currentSolution))
            iNX2++;
        double dX2 = (double)iNX2/(double)(lstComparisonSet.size() + 1);
        
        return (dX-dX2);
    }

    /**
     *
     * @return list of new samples, as dummy architectures. These architecutres
     * will not (should not) be used other than dominance comparision.
     * (The whole architecture objects are used because it is necessary to consider
     * minimizing/maximizing aspects of the attributes in dominance check)
     */
    private List<Architecture> sampleAdditionalPoints(){
        if (nondominatedSolutions.size() < 2){
            maintainArchiveSize(2);
        }
        List<Architecture> lstNewSamples = new LinkedList<Architecture>();
        Architecture oSeed = nondominatedSolutions.get(0); //An architecture that is used for
        int iNObjectives = Conf.instance().getNEvaluators();
        double[] adMaxValues = new double[iNObjectives];
        double[] adMinValues = new double[iNObjectives];
        for(int iAtt=0; iAtt<iNObjectives; iAtt++){ //Initialize array elements
            adMaxValues[iAtt] = Double.MIN_VALUE;
            adMinValues[iAtt] = Double.MAX_VALUE;
        }
        //Get min/max ranges for the quality attributes.
        for(Architecture oArch: nondominatedSolutions){
            List<Attribute> lstAttributes = oArch.getAttributeList();
            for(int iAtt=0; iAtt<iNObjectives; iAtt++){
                adMaxValues[iAtt] = Math.max(lstAttributes.get(iAtt).getOptimizerValue(), adMaxValues[iAtt]);
                adMinValues[iAtt] = Math.min(lstAttributes.get(iAtt).getOptimizerValue(), adMinValues[iAtt]);
            }
        }

        while (lstNewSamples.size() < additionalSamples ){
            //Creating a new dummy architecture to hold the new data point
            Architecture oDummy = oSeed.clone();
            oDummy.setAttributeList( oSeed.getAttributeList());
            for(int iAtt=0; iAtt<iNObjectives; iAtt++){
                //Get a random sample from each range
                double dObjVal = adMinValues[iAtt] + rnd.nextDouble() * (adMaxValues[iAtt] - adMinValues[iAtt]);
                oDummy.getAttributeList().get(iAtt).setOptimizerValue(dObjVal);
            }

            //First check for strict dominance criteria.
            boolean bIsDominated =false;
            for(Architecture oArch: nondominatedSolutions){
                if ( oDummy.isDominated(oArch)){
                    lstNewSamples.add(oDummy);
                    bIsDominated = true;
                    break;
                }
            }
            // If above fails, check for weak dominances.
            if (!bIsDominated){
                int iRndAtt = rnd.nextInt(iNObjectives);
                bIsDominated =false;
                for(Architecture oArch: nondominatedSolutions){
                    double dRandAtt = oArch.getAttributeValue(iRndAtt);
                    oDummy.getAttributeList().get(iRndAtt).setOptimizerValue(dRandAtt);
                    if ( oDummy.isDominated(oArch)){
                        lstNewSamples.add(oDummy);
                        bIsDominated = true;
                        break;
                    }
                }
           }
        }
        return lstNewSamples;
    }

    /**
     * If the archive(nondominatedSolutions) size is less than iSize,
     * try to find random solutions.
     * @param iSize
     */
    protected void maintainArchiveSize(int iSize){
        log.debug(m_Watch.GetElapsed()+"Attempt to maintain archive size ="+ iSize  );
        //Fill the archive with iSize non-dominated solutions
        while (nondominatedSolutions.size() < iSize ){
            Architecture oNew = popGenerator.generatePopulation(1, problem,ConstraintHandling.REPAIR).get(0);
            problem.evaluateAttributes(oNew);
            checkDominance(oNew);
        }
    }
}
