package archeopterix.evaluation;

import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.solution.Architecture;
import java.io.Serializable;

/**
 * The value of a single objective/attribute (as defined by its associated {@code
 * ArchitectureEvaluator}) for some {@code Architecture}.
 */
public class Attribute implements Serializable{
	private ArchitectureEvaluator evaluator;
	private double value;
	private boolean isMinimising;
	private Exception highEx;
	private Exception lowEx;

	public Attribute(ArchitectureEvaluator evaluator) {
            this.evaluator = evaluator;
            value = 0.0F;
            this.isMinimising = this.evaluator.isMinimisationObjective();
            highEx = new Exception("Solution exceeded range for objective value");
            lowEx = new Exception("Solution below objective value range");
	}
        public String getName(){
            return evaluator.getName();
        }
	/**
	 * Creates a new {@code Attribute} that is a duplicate of {@code parent}.
	 * Given the current implementation, making {@code Attribute}
	 * {@link Cloneable} would achieve the same effect, more quickly, than
	 * using this constructor.
	 */
	public Attribute(Attribute parent) {
		evaluator = parent.evaluator;
		value = parent.value;
		isMinimising = parent.isMinimising;
	}
	
	public boolean isMinimizing() { return isMinimising; }

	public double getOptimizerValue() { return isMinimising ? (evaluator.getMaxValue()- value) : value; }

        //Added to support additional sampling points in simulated annealing
        public void setOptimizerValue(double dVal) { value = isMinimising ? (evaluator.getMaxValue()- dVal) : dVal; }
	
	public double getNormalisedOptimiserValue(EmbeddedSystemProblem oProblem, Architecture oDep) {
		return getOptimizerValue()/evaluator.getMaxValue();
	}

	public double getNormalisedOptimiserValue() {
		return getOptimizerValue()/evaluator.getMaxValue();
	}
	public double getValue()
        {
		return value;
        }

    public double getMaxValue()
	{
		return evaluator.getMaxValue();
	}

	public double getMinValue()
	{
		return evaluator.getMinValue();
	}

	
	public double evaluate(EmbeddedSystemProblem oProblem, Architecture oDep) {
		value  = (double)evaluator.evaluate(oProblem, oDep);
		return getOptimizerValue();
	}

	public double convertToNormalisedForHV() throws Exception
	{
		if (value > evaluator.getMaxValue())
			throw highEx;

		if (value < evaluator.getMinValue())
			throw lowEx;

		if (isMinimising)
		{
			return  (evaluator.getMaxValue()- value)/(evaluator.getMaxValue() - evaluator.getMinValue());
		}
		return (value - evaluator.getMinValue())/(evaluator.getMaxValue() - evaluator.getMinValue());
		
	}


	@Override
	public String toString() { return  evaluator.getName() + " = " + value; }

}
