package archeopterix.evaluation;

import archeopterix.conf.Conf;
import java.util.List;

import archeopterix.problem.Component;
import archeopterix.problem.Host;
import archeopterix.problem.InteractionProperties;
import archeopterix.problem.NetworkProperties;
import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.problem.SchedulingStratergy;
import archeopterix.problem.componentselection.solution.ComponentSelectionSolution;
import archeopterix.problem.deployment.solution.DeploymentSolution;
import archeopterix.problem.redundancy.solution.RedundancyAlloc;
import archeopterix.solution.Architecture;
import archeopterix.util.PropertyMap;
import archeopterix.util.matrix.Matrix;
import archeopterix.util.probability.MonteCarlo;
import archeopterix.util.probability.MyStatistics;
import java.util.Iterator;

public class ResponseTimeEvaluator extends ArchitectureEvaluator {

    protected Matrix mat_Q0;
    protected Matrix mat_Pij;
    protected Matrix mat_Vi;//Expected No of visits for Comps
    protected EmbeddedSystemProblem m_Problem = null;
    protected boolean b_SeviceRespTime = false; //Service Reliability
    protected int i_ServiceID = 0;

    public ResponseTimeEvaluator() {
        super("Resp.Time", true);
    }

    public ResponseTimeEvaluator(String sName) {
        super(sName, false);

    }

    public ResponseTimeEvaluator(Integer iServiceID) {
        super("Resp.TimeSrv" + iServiceID, false);
        b_SeviceRespTime = true;
        i_ServiceID = iServiceID;
    }

    public ResponseTimeEvaluator(String sName, Integer iServiceID) {
        super(sName + "Srv" + iServiceID, false);
        b_SeviceRespTime = true;
        i_ServiceID = iServiceID;
    }

    @Override
    public double evaluate(EmbeddedSystemProblem oProblem, Architecture oArch) {
        if (m_Problem == null) {
            m_Problem = oProblem;
            ComputeExpectedVisitsMatrix();
        }

        if (oArch instanceof RedundancyAlloc) {
            return computeResponseTime((RedundancyAlloc) oArch);
        } else if (oArch instanceof DeploymentSolution) {
            return evaluateExpVistBased(oProblem, (DeploymentSolution) oArch);
        } else if (oArch instanceof ComponentSelectionSolution){
            ComponentSelectionSolution oCS = (ComponentSelectionSolution) oArch;
            if (subjectUncertainty) {
                double dVal = 0.0;
                MyStatistics oStat = getStats(oArch);
                uncertaintyHandler.ResetRuns();
                while (uncertaintyHandler.hasMoreRuns(oStat)) {
                    dVal = computeResponseTime(oCS);
                    oStat.update(dVal, true);
                }
                return uncertaintyHandler.getEstimate(oStat);
            } else {
                return computeResponseTime( oCS);
            }
        }else {
            return 0.0;
        }
    }

    protected double computeResponseTime(ComponentSelectionSolution oCS) {
        double dRet = 0.0;
        for (int iCompi = 0; iCompi < m_Problem.getComponentCount(); iCompi++) {
            Component oComp = m_Problem.getComponent(iCompi);
            int iOption = oCS.getCurrentSelection(iCompi);
            double dVi = mat_Vi.get(iCompi, 0);
            double dSojournTime;//  = getExecutionTime(iCompi, iHi, oDep) * dVi; //Execution Time * Expected Visits
            if (subjectUncertainty){
                dSojournTime = MonteCarlo.instance().Sample(oComp.getPDFFromListProperty(Component.OPTION_SOJOURN_TIME_DT, iOption));
            } else {
                dSojournTime = oComp.getDoubleFromListProperty(Component.OPTION_SOJOURN_TIME, iOption);
            }
            dRet += dSojournTime;
            for (int iCompj = 0; iCompj < m_Problem.getComponentCount(); iCompj++) {
                if (m_Problem.areInteracting(iCompi, iCompj)) {
                    dRet += m_Problem.getInteractionProperties(iCompi, iCompj).getDouble(InteractionProperties.TX_TIME) / (3.6E5);;
                }
            }
        }
        return dRet;
    }

    
    protected double computeResponseTime(RedundancyAlloc oRA) {
        List<Component> comps = m_Problem.getComponents();
        double dSumResponseTime = 0.0;
        for (int iCompi = 0, c = comps.size(); iCompi < c; iCompi++) {
            double dRT = 0.0;
            double dSojournTime = comps.get(iCompi).getDoubleProperty(Component.SOJOURN_TIME);

            //Calculating TX overhead due to RA
            for (int iCompj = 0; iCompj < c; iCompj++) {
                if (m_Problem.areInteracting(iCompi, iCompj)) {
                    int iRAj = oRA.getRALevel(iCompj);
                    double dDeltaTxTime = m_Problem.getInteractionProperties(iCompi, iCompj).getDouble(InteractionProperties.DELTA_TX_TIME);
                    dSojournTime += (iRAj * dDeltaTxTime);
                }
            }

            //Calculating RX overhead due to RA
            for (int iCompj = 0; iCompj < c; iCompj++) {
                if (m_Problem.areInteracting(iCompj, iCompi)) {
                    int iRAj = oRA.getRALevel(iCompj);
                    double dDeltaTxTime = m_Problem.getInteractionProperties(iCompj, iCompi).getDouble(InteractionProperties.DELTA_RX_TIME);
                    dSojournTime += (iRAj * dDeltaTxTime);
                }
            }

            //Response time calculation for component i
            double dVi = mat_Vi.get(iCompi, 0);
            dRT = dSojournTime * dVi;
            dSumResponseTime += dRT;

        }
        return dSumResponseTime;
    }

    /**
     * Expected number of visit based response time computation presente by Vibhu Saujanya Sharma a, Kishor S. Trivedi (Sharma07T)
     * @param oProblem
     * @param oDep
     * @return
     */
    protected double evaluateExpVistBased(EmbeddedSystemProblem oProblem, DeploymentSolution oDep) {
        double dRet = 0.0;
        for (int iCompi = 0; iCompi < oProblem.getComponentCount(); iCompi++) {
            int iHi = oDep.getHostIndex(iCompi);
            double dVi = mat_Vi.get(iCompi, 0);
            double dResTime = getExecutionTime(iCompi, iHi, oDep) * dVi; //Execution Time * Expected Visits
            dRet += dResTime;
            for (int iCompj = 0; iCompj < oProblem.getComponentCount(); iCompj++) {
                if (oProblem.areInteracting(iCompi, iCompj)) {
                    int iHj = oDep.getHostIndex(iCompj);
                    double dCLij = b_SeviceRespTime ? oProblem.getInteractionProperties(iCompi, iCompj).getDoubleFromList(InteractionProperties.COMM_LOAD_KB, i_ServiceID) : oProblem.getInteractionProperties(iCompi, iCompj).getDouble(InteractionProperties.COMM_LOAD_KB);
                    PropertyMap oNW = oProblem.getNetworkProperties(iHi, iHj);
                    if (oNW != null) {//Network exists
                        int iDRij = oNW.getInt(NetworkProperties.DATA_RATE_KBPS);
                        double dVij = mat_Vi.get(iCompi, 0) * mat_Pij.get(iCompi, iCompj);
                        dResTime = (dCLij / iDRij) * dVij; // Communication time * Expected visits
                        dRet += dResTime;
                    }
                }
            }
        }
        return dRet;
    }

    /**
     * Returns the execution time of a component in an ECU
     * @param iCompID
     * @param iHostID
     * @param oDep
     * @return
     */
    protected double getExecutionTime(int iCompID, int iHostID, DeploymentSolution oDep) {
        int iMIPS = m_Problem.getECU(iHostID).getIntProperty(Host.SPEED_MIPS);
        if (Conf.instance().getSchedulingStratergy() == SchedulingStratergy.FIXED) {
            double dMI = 0.0;
            List<Component> lstOtherComps = oDep.getComponentListForHost(iHostID, m_Problem);
            Iterator<Component> ite = lstOtherComps.iterator();
            while (ite.hasNext()) {
                Component oComp = ite.next();
                if (b_SeviceRespTime) {
                    //Summing up for all services
                    List<Double> lstMIs = oComp.getDoubleListProperty(Component.COMPUTATIONS_MI);
                    Iterator<Double> ite2 = lstMIs.iterator();
                    while (ite2.hasNext()) {
                        dMI += ite2.next();
                    }
                } else {
                    dMI += oComp.getDoubleProperty(Component.COMPUTATIONS_MI);
                }
            }
            return (dMI / (double) iMIPS);
        } else {
            double dMI = b_SeviceRespTime ? m_Problem.getComponent(iCompID).getDoubleFromListProperty(Component.COMPUTATIONS_MI, i_ServiceID) : m_Problem.getComponent(iCompID).getDoubleProperty(Component.COMPUTATIONS_MI);
            return (dMI / (double) iMIPS);
        }

    }

    protected void ComputeExpectedVisitsMatrix() {
        if (mat_Q0 == null) {
            mat_Q0 = new Matrix(m_Problem.getComponentCount(), 1);
            mat_Pij = new Matrix(m_Problem.getComponentCount(), m_Problem.getComponentCount());
        }

        for (int iCompi = 0; iCompi < m_Problem.getComponentCount(); iCompi++) {
            double dQ0 = b_SeviceRespTime ? m_Problem.getComponent(iCompi).getDoubleFromListProperty(Component.INITILIZE_PROB, i_ServiceID) : m_Problem.getComponent(iCompi).getDoubleProperty(Component.INITILIZE_PROB);
            mat_Q0.set(iCompi, 0, dQ0);
            for (int iCompj = 0; iCompj < m_Problem.getComponentCount(); iCompj++) {
                if (m_Problem.areInteracting(iCompi, iCompj)) {
                    double dPij = b_SeviceRespTime ? m_Problem.getInteractionProperties(iCompi, iCompj).getDoubleFromList(InteractionProperties.TRANSITION_PROB, i_ServiceID) : m_Problem.getInteractionProperties(iCompi, iCompj).getDouble(InteractionProperties.TRANSITION_PROB);
                    mat_Pij.set(iCompi, iCompj, dPij);
                }
            }
        }

        //System.out.println ( "Pij for Service "+ i_ServiceID + " =" +  mat_Pij.toString() );
        //System.out.println ( "Q0 for Service "+ i_ServiceID + " =" + mat_Q0.toString() );

        //System.out.println ( "S=" + i_ServiceIndex + "EX[i]" + PrintMatrix(E) );
        //Expected Number of Visits for Components
        Matrix T = (Matrix.identity(mat_Pij.getColumnDimension(), mat_Pij.getRowDimension())).minus(mat_Pij.transpose());
        mat_Vi = T.inverse().times(mat_Q0);
        //System.out.println ( "S=" + i_ServiceIndex + "EX[i]" + PrintMatrix(E) );
        System.out.println("Expected Visits in Service" + i_ServiceID + " =" + mat_Vi.toString());
    }
}
