/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package archeopterix.experiments;

import archeopterix.util.statistics.ProbabilityDistribution;
import archeopterix.util.matrix.Matrix;
import archeopterix.util.probability.Distribution;
import archeopterix.util.probability.MonteCarlo;
import java.io.Serializable;
import java.util.List;
import java.util.Random;
import java.util.Vector;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 *
 * @author indika
 */
public class DTMCKubat implements Serializable{

    private static final Log log = LogFactory.getLog(DTMCKubat.class);
    private ProbabilityDistribution[] arr_PDFs;
    private List<Ref>[] arr_Subscribers;
    private int i_nParams;
    private int i_MatrixSize;
    private Matrix m_P;
    private Matrix m_Q;
    private double d_InteractionProbability = 0.8; //Probability of Existance of transistion
    
    public DTMCKubat(int iMatrixSize, ProbabilityDistribution[] arrPDFs){
        arr_PDFs = arrPDFs;
        i_nParams = arr_PDFs.length;
        i_MatrixSize = iMatrixSize;
        m_P = new Matrix (i_MatrixSize, i_MatrixSize,0.0);
        m_Q = new Matrix (i_MatrixSize,1,0.0);
        parameterizeMatrix();
        resetToMeans();

    }



    /**
     * Inner class to hold a reference to matrix elements
     */
    private class Ref implements Serializable{
        public int iX, iY;
        public Ref(int X, int Y){
            iX = X;
            iY = Y;
        }

        @Override
        public String toString(){
            return "(" + iX + "," + iY + ")";
        }
    }

    /** subscribe matrix elements to pdfs. Gurantees absoarbing DTMC
     * 
     */
    private void parameterizeMatrix(){
        arr_Subscribers = new Vector[i_nParams];
        for (int iE =0; iE<i_nParams; iE++)
            arr_Subscribers[iE] = new Vector<Ref>();
        
        Random rnd = new Random();
        int iIgoreRow = 1+ rnd.nextInt(i_MatrixSize - 1 ); //To avoid the fist one being absorbing
        for (int i=0; i<i_MatrixSize; i++){
            if (i== iIgoreRow)
                continue; //ignore one row, i.e. DTMC is absoarbing
            
            for(int j=0; j<i_MatrixSize; j++){
                double dVal = rnd.nextDouble();
                if ((dVal >= d_InteractionProbability) || (i==j)){
                //if (dVal >= d_InteractionProbability){ //Self loops allowed
                        continue;
                }
                int iParam = rnd.nextInt(i_nParams);
                //Assiging i, j to parameter ipdf
                Ref oRef = new Ref(i,j);
                arr_Subscribers[iParam].add(oRef);
            }
        }
        if (log.isTraceEnabled()){
            for (int iE =0; iE<i_nParams; iE++)
                log.debug("Subscribers of Param["+ iE +"] =" + getSubscribersList(iE));
        }

        //Initialize Q
        double dSum = 0.0;
        for (int i=0;i<i_MatrixSize; i++){
            m_Q.set(i, 0, rnd.nextDouble());
            dSum += m_Q.get(i, 0);
        }

        //Normalizing
        if (dSum > 0){
            for (int i=0;i<i_MatrixSize; i++)
                m_Q.set(i,0, (m_Q.get(i, 0)/dSum));
        }
    }

    public String getSubscribersList(int iParam){
        return arr_Subscribers[iParam].toString();
    }

    /** update relavant values in the matrix corresponding to that parameter
     * 
     * @param iParam, ID of the parameter.
     * @param dValue , this value will be set at all subscribed pints in the matrix
     */
    private void updateSubscribers( int iParam, double dValue){
        for(Ref oRef : arr_Subscribers[iParam])
            m_P.set(oRef.iX, oRef.iY, dValue);
    }


    /** Update the DTMC with the mean values of the PDFs
     *
     * @return the Matrix consturcted from means, gurantees, absoarbing, and DTMC properties.
     */
    public Matrix resetToMeans(){
        for(int iParam=0; iParam < i_nParams; iParam++){
            ProbabilityDistribution oPDF = arr_PDFs[iParam];
            double dMean = Distribution.getAnalyticalMean(oPDF);
            //System.out.println("PDF["+ iParam + "]= " +
             //       oPDF.toString() + " MEAN=" + dMean );
            updateSubscribers(iParam, dMean);
        }
        return m_P;
    }

    /**
     * Get random sample from each distribution and return the updated Transistion matrix
     * @return Transistion matrix with random samples
     */
    public Matrix getRandom() {
        MonteCarlo oMC = new MonteCarlo();
        for (int iParam = 0; iParam < i_nParams; iParam++) {
            ProbabilityDistribution oPDF = arr_PDFs[iParam];
            double dSample = oMC.Sample(oPDF);
            //System.out.println("PDF["+ iParam + "]= " +
            //       oPDF.toString() + " Sample=" + dSample );
            updateSubscribers(iParam, dSample);
        }
        return m_P;
    }

    /**
     * Change only one parameter from the current state of the matrix. The given parameter will be sampled from its distribution
     * @param iParamID
     * @return
     */
    public double getDelta(int iParamID){
        MonteCarlo oMC = new MonteCarlo();
        ProbabilityDistribution oPDF = arr_PDFs[iParamID];
        double dSample = oMC.Sample(oPDF);
        //double dSample =Distribution.getAnalyticalMean(oPDF);
        updateSubscribers(iParamID, dSample);
        getNormalizedCopy();
        //log.debug("After Updating parameter["+ iParamID+"]="+ dSample+" Matrix" + m_P.toString());
        return dSample;

    }

    
    /**
     * Ensure the sum of each row is 1.0
     */
    private Matrix getNormalizedCopy(){
        Matrix P = m_P.copy();
        for (int i = 0; i < i_MatrixSize; i++) {
            double dRowSum = 0.0;
            for (int j = 0; j < i_MatrixSize; j++) {
                double dVal = P.get(i, j);
                dRowSum += dVal;
            }
            //Ensure that the Rowsum =1
            if (dRowSum > 0) {
                for (int j = 0; j < i_MatrixSize; j++) {
                    double dVal = P.get(i, j);
                    P.set(i, j, dVal / dRowSum);
                }
            }
        }
        return P;
    }

    
    /**
     * Compute the reliability from kubats model
     * @param adRel : array of reliabilities
     * @return PROD ( ri^vi)
     * Evalate V= (I-P)^-1 * Q
     *
     */
    public double Compute(double[] adRel){
        //m_Watch.Reset();
        //long lstart= System.nanoTime();

        if ( adRel.length != i_MatrixSize){
            log.error("Reliability array length(" + adRel.length + ") != matrix size(" + i_MatrixSize + ")");
            return 0.0;
        }
        Matrix mV = new Matrix(i_MatrixSize,1,0.0);
        Matrix mP = getNormalizedCopy();
        Matrix mT=  ((Matrix.identity( i_MatrixSize, i_MatrixSize)).minus(mP));
        if ( mT.det() == 0.0 ){
            log.error("Det = 0.0");
            return 0.0;//FIXME
        }
        mV = mT.inverse().times(m_Q);
        double dRel = 1.0;
        for (int i=0; i<i_MatrixSize; i++){
            //System.out.println("i="+ i + " REL=" + adRel[i] + " V="+ mV.get(i, 0) + " SREL=" + dRel );
            dRel *= (Math.pow(adRel[i],mV.get(i, 0)));
        }
        return dRel;
        //return   mT.inverse().get(0,i_MatrixSize-1);
        //return (System.nanoTime() - lstart);
        //Logger.Logln("OUT::T="+ lElapsed +";Rs="+ d_Rs);
	}


}

