/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package archeopterix.problem.deployment.algorithm.heuristic;

import archeopterix.algorithm.Algorithm;
import archeopterix.algorithm.PopulationGenerator;
import archeopterix.algorithm.Stats;
import archeopterix.algorithm.ea.EAOperators;
import archeopterix.algorithm.ea.EAOperatorsFactory;
import archeopterix.conf.Conf;
import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.problem.deployment.solution.DeploymentSolution;
import archeopterix.problem.deployment.solution.OriginalDeploymentSolution;
import archeopterix.reporting.HypervolumeDevelopment;
import archeopterix.solution.Architecture;
import archeopterix.solution.ConstraintHandling;
import archeopterix.solution.OutOfOptionsException;
import archeopterix.solution.Repairable;
import archeopterix.util.PropertyMap;
import archeopterix.util.matrix.Matrix;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/** Bayesian Heuristic for Component Deployment Optimisation
 *
 * @author indika
 */
public class BHCDO extends Algorithm {

    public static final String PAR_POP_SIZE ="popSize";
    public static final String PAR_EXPLORATION_RATE ="explorationRate";

    protected int populationSize = 100;
    protected double exploraitionRate = 0.1;
    protected Random rnd = Conf.instance().random;

    /** H-C combinations Occurances in Elite Set **/
    private int[][] occuranceInEliteSet;
    /** H-C combination Occurances in Total **/
    private int[][] occuranceInTotal;
    /** Conditional probabilities of H-C **/
    private Matrix mat_CondProb;

    
    protected OriginalDeploymentSolution currentAllocation;
	protected EAOperators operators; /** For exploration purpose **/
    
    private static final Log log = LogFactory.getLog(BHCDO.class);

    
    @Override
    public void initialise(EmbeddedSystemProblem problem, PropertyMap newParams, Collection<Architecture> partialInitPop,
            PopulationGenerator oPopGenerator) {
        super.initialise(problem, newParams, partialInitPop, oPopGenerator);
        populationSize = params.getInt(PAR_POP_SIZE);
        exploraitionRate = params.getDouble(PAR_EXPLORATION_RATE);
        if (partialInitPop != null)
			population.addAll( partialInitPop );
		operators = EAOperatorsFactory.instance()
							.getSuitableOperators( Conf.instance().getSolutionRepresentation() );
        population = popGenerator.generatePopulation(population, populationSize, problem, ConstraintHandling.REPAIR);
		currentAllocation = new OriginalDeploymentSolution(i_ECUs, i_Comps);
        initStructures();
		evaluatePopulation();
		printCurrentNondominatedSolutions("Intial");
    }

	protected void initStructures(){
		occuranceInEliteSet = new int[i_Comps][i_ECUs];
        occuranceInTotal = new int[i_Comps][i_ECUs];
        mat_CondProb = new Matrix(i_Comps, i_ECUs, 0.0);
		
	}
	
    @Override
    public void oneIteration() {
        populateBayesianStructures();
        population.clear();
        while (population.size() < populationSize){
            try {
                DeploymentSolution deployment = createSolution();
                processNewSolution(deployment);
                doExplorationStep(deployment);
                //Do a exploration move
            } catch (OutOfOptionsException ex) {
                if (log.isDebugEnabled())
                    log.debug(m_Watch.GetElapsed()+"No Allocation could be found " + ex.toString());
            }
        }
        double hv = computeBalancedHypervolume(nondominatedSolutions);
        HypervolumeDevelopment.getInstance().addHypervolumeForIteration(hv, getNumberEvaluations());
        if (log.isDebugEnabled() && ((Stats.instance().getEvaluationCounter()%1000) == 0 )) { //&& ((Stats.instance().l_CntEvalus%1000) == 0 )
                log.debug(m_Watch.GetElapsed() + ": Itr: " + (currIteration + 1) + " Evals=" + Stats.instance().getEvaluationCounter());
                printCurrentNondominatedSolutions("Iteration :" + (currIteration + 1));
        }
        incrementIteration();
		
    }

        protected void doExplorationStep( DeploymentSolution newSolution){
            Architecture parent1 = newSolution;
            Architecture parent2 = nondominatedSolutions.get(rnd.nextInt(nondominatedSolutions.size()));
            try {
                Architecture[] children = operators.singlePointCrossover(parent1, parent2, problem);
                for (int i = 0; i < children.length; i++) {
                    children[i].setFeasible(problem.validate(children[i]));
                    if (!children[i].isFeasible()) {
                        if (constraintHandling == ConstraintHandling.REPAIR) {
                            //the bad thing is if this fails for one child, both are discarded
                            ((Repairable) children[i]).attemptRepair(problem);
                            children[i].setFeasible(true);
                        } else {
                            children[i].setConstraintViolation(problem);
                        }
                    }
                }
                processNewSolution(children[0]);
                processNewSolution(children[1]);
            } catch (OutOfOptionsException ex) {
                    if (log.isDebugEnabled())
                            log.debug(m_Watch.GetElapsed()+"Exploration step failed. " + ex.toString());
            }
        }

        protected void processNewSolution(Architecture newArch){
            problem.evaluateAttributes(newArch);
            population.add(newArch);
            checkDominance(newArch);
        }
    protected void evaluatePopulation() {
        for (Architecture arch : population) {
            problem.evaluateAttributes(arch);
            checkDominance(arch);
        }
    }

    protected DeploymentSolution createSolution()throws OutOfOptionsException{
            List<Integer> lstCompsToAllocate = new LinkedList<Integer>();
        for (int i=0; i< i_Comps; i++){
            lstCompsToAllocate.add(i);
            currentAllocation.set(i,-1); // Intialising
        }
        while ( lstCompsToAllocate.size() > 0){
            int iCIdx = rnd.nextInt(lstCompsToAllocate.size());
            int iC = lstCompsToAllocate.remove(iCIdx);
            int iAttempts = 0;
            currentAllocation.set(iC,selectHostBayesian(iC));
            while(! problem.validate(currentAllocation)){
                currentAllocation.set(iC,selectHostBayesian(iC));
                if (iAttempts > 100){
                    throw new OutOfOptionsException();
                }
            }
        }
        return new OriginalDeploymentSolution(currentAllocation);
    }

    /**
     * Use bayesian heuristic to get host allocation for a given component
     * @param iC component index
     * @return HostIndex
     */
    private int selectHostBayesian( int iC){
        //Compute recursive prob. multiplication previous allocations
		//Commented following: as it has no effect on the decision
        /*double dPreProb = 1.0;
        for (int iCx=0; iCx< i_Comps; iCx++){
            int iHx = currentAllocation.getHostIndex(iCx);
            if (iHx > 0){ //Only allocated ones
                dPreProb *= mat_CondProb.get(iCx, iHx);
            }
        }*/
        double[] adWeighs = new double[i_ECUs];
        for (int i=0; i<i_ECUs; i++)
            adWeighs[i] = mat_CondProb.get(iC, i); //* dPreProb;
        return doRoulette(adWeighs);
    }

    /** Compute the occurances of H-C combinations and update
     * conditional probaility tables
     */
    protected void populateBayesianStructures() {
        resetCounts();
        for (Architecture arch : population) {
            for (int iC = 0; iC < i_Comps; iC++) {
                int iH = ((DeploymentSolution) arch).getHostIndex(iC);
                occuranceInTotal[iC][iH] += 1;
            }
        }

        for (Architecture arch : nondominatedSolutions) {
            for (int iC = 0; iC < i_Comps; iC++) {
                int iH = ((DeploymentSolution) arch).getHostIndex(iC);
                occuranceInEliteSet[iC][iH] += 1;
            }
        }

        for (int iC = 0; iC < i_Comps; iC++) {
            double dSumCP = 0.0;
            int inEmptyCells = 0;
            for (int iH = 0; iH < i_ECUs; iH++) {
                double dE = occuranceInEliteSet[iC][iH];
                double dT = occuranceInTotal[iC][iH];
                double dCP = 0.0;
                if (dT > 0) {
                    dCP = dE / dT;
                }
                dSumCP += dCP;
                mat_CondProb.set(iC, iH, dCP);
                if (occuranceInEliteSet[iC][iH] <= 0) {
                    inEmptyCells++;
                }
            }
            //Normalising
            double dRowSum = dSumCP / (1 - exploraitionRate);
            double dCompensation = 0.0;
            if (inEmptyCells > 0) {
                dCompensation = exploraitionRate / (double) inEmptyCells;
            }
            for (int iH = 0; iH < i_ECUs; iH++) {
                double dCP = mat_CondProb.get(iC, iH);
                if (occuranceInEliteSet[iC][iH] <= 0) {
                    dCP = dCompensation;
                } else {
                    dCP = dCP / dRowSum;
                }
                mat_CondProb.set(iC, iH, dCP);
            }
        }
        //System.out.println( mat_CondProb.toString());
    }

    private void resetCounts() {
        for (int iC = 0; iC < i_Comps; iC++) {
            for (int iH = 0; iH < i_ECUs; iH++) {
                occuranceInEliteSet[iC][iH] = 0;
                occuranceInTotal[iC][iH] = 0;
            }
        }
        mat_CondProb.reInit(0.0);
    }

    /** execute a rouletted with the given weights and return an index
     *
     * @param arrValues
     * @return
     */
    protected int doRoulette(double[] arrWeights){
        
        double dSum = 0.0;
        for ( int i=0; i< arrWeights.length; i++)
            dSum += arrWeights[i];
        double dTh = rnd.nextDouble()* dSum;
        int iIndex=0;
        double dCum = arrWeights[iIndex];
        //Selecting the host based on probability
        while( dCum < dTh){
            iIndex++;
            dCum += arrWeights[iIndex];
        }
        return iIndex;
    }
}
