package archeopterix.problem.deployment.solution;

import java.util.LinkedList;
import java.util.Random;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;

import archeopterix.conf.Conf;
import archeopterix.problem.Component;
import archeopterix.problem.Host;
import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.problem.deployment.Deployment;
import archeopterix.solution.Architecture;
import archeopterix.solution.OutOfOptionsException;
import archeopterix.solution.Repairable;
import archeopterix.util.Pair;

/**
 *
 * @author alda
 */
public class ECUBasedDeployment extends DeploymentSolution implements Repairable{
	private static final long serialVersionUID = 1L;
	private ArrayList<List<Component>> lst_Alloc;
	private final int[] allMinusOnes;
	private int[] memoryAvailabilities;

	public ECUBasedDeployment(int iHosts, int iComps){
            i_Hosts = iHosts;
            i_Components = iComps;
            lst_Alloc = new ArrayList<List<Component>>();
            allMinusOnes = new int[iHosts];
            memoryAvailabilities = new int[iHosts];
            for (int i = 0; i < i_Hosts; i++){
                    allMinusOnes[i] = -1;
                    lst_Alloc.add(new LinkedList<Component>());
            }
	}

	public ECUBasedDeployment(EmbeddedSystemProblem problem){
            i_Hosts = problem.getECUCount();
            i_Components = problem.getComponentCount();
            lst_Alloc = new ArrayList<List<Component>>();
            allMinusOnes = new int[i_Hosts];
            memoryAvailabilities = new int[i_Hosts];
            for (int i = 0; i < i_Hosts; i++){
                allMinusOnes[i] = -1;
                lst_Alloc.add(new LinkedList<Component>());
            }
            randomlyAllocate(problem.getCompsSortedByPossibleAssignments());
	}

	public ECUBasedDeployment(ECUBasedDeployment other){
            this(other.i_Hosts, other.i_Components);
            for (int ecuID = 0; ecuID < i_Hosts; ecuID++){
                lst_Alloc.get(ecuID).addAll(other.lst_Alloc.get(ecuID));
            }
	}
	//only there so we can use Class.forName();

/*	public ECUBasedDeployment()
	{
		lst_Alloc = new ArrayList<List<Component>>();
		m_UserData = null;
	}*/

	@Override
	public void initializeFromString(String sDepString, EmbeddedSystemProblem oProblem)
	{
		StringTokenizer st = new StringTokenizer(sDepString, System.getProperty("line.separator"), false);
		String line = st.nextToken();
		String[] aSt = line.substring(0, sDepString.indexOf(":")).split(",");
		lst_Alloc = new ArrayList<List<Component>>();
		i_Hosts = Integer.parseInt(aSt[0]);
		i_Components = Integer.parseInt(aSt[1]);
		lst_Alloc = new ArrayList<List<Component>>();
		memoryAvailabilities = new int[i_Hosts];

		for (int i = 0; i < i_Hosts; i++)
		{
			line = st.nextToken(); //this will throw an exception if there aren't enough hosts.
			line = line.trim();
			if (Integer.parseInt(line.substring(0, line.indexOf("{"))) != i)
			{
				System.out.println("Wrong host at " + i);
			}
			LinkedList<Component> lst = new LinkedList<Component>();
			aSt = line.substring(line.indexOf("{") + 1, line.indexOf("}")).split(",");
			List<Component> comps = oProblem.getComponents();
			for (int j = 0; j < aSt.length; j++)
			{
				if (aSt[j].length() > 0)
				{
					lst.add(comps.get(Integer.parseInt(aSt[j])));
				}
			}

			lst_Alloc.add(lst);
		}

	}

        @Override
	public void reset(){
            for (int i = 0; i < i_Hosts; i++){
                lst_Alloc.get(i).clear();

            }
	}

	@Override
	public Architecture clone(){
            return new ECUBasedDeployment(this);
	}

        @Override
	public final void randomlyAllocate(List<Component> lst_sortedComponents)
	{
		Random random = Conf.instance().random;
		BitSet[] assignedToHosts = new BitSet[i_Hosts];
		for (int i = 0; i < i_Hosts; i++)
		{
			assignedToHosts[i] = new BitSet(i_Components);
		}
		BitSet assigned = new BitSet(i_Components);
		Iterator<Component> comp_iterator = lst_sortedComponents.iterator();

		// assign ECUs with few options first
		while (comp_iterator.hasNext())
		{
			Component comp = comp_iterator.next();

			if (assigned.get(comp.index))
			{
				continue;
			}

			boolean okToAssign;
			Host ecu;
			do{
                            int opt = random.nextInt(comp.getCanUseHostListLength());
                            ecu = comp.getHostOption(opt);

                            okToAssign = true;
                            //TODO Wrap up checking functionality in a method
                            //check if components in the same group have colocate restrictions with other comp.
                            Iterator<Component> mustNotColocateWith = comp.getMustNotColocateWith().iterator();
                            while (mustNotColocateWith.hasNext() && okToAssign){
                                okToAssign = !assignedToHosts[ecu.index].get(mustNotColocateWith.next().index);
                            }
                            Iterator<Component> buddies = comp.getMustColocateWith().iterator();
                            while (buddies.hasNext() && okToAssign){
                                mustNotColocateWith = buddies.next().getMustNotColocateWith().iterator();
                                while (mustNotColocateWith.hasNext() && okToAssign){
                                    okToAssign = !assignedToHosts[ecu.index].get(mustNotColocateWith.next().index);
                                }
                            }

                            //FIXME May wish to log this information 
//				if (! okToAssign)
//					System.out.println("Disallowed component " + comp.index + " from ecu " + ecu.index + " because it is assigned: " + assignedToHosts[ecu.index]);
			} while (!okToAssign);
			List<Component> ecuList = lst_Alloc.get(ecu.index);
			ecuList.add(comp);
			assigned.set(comp.index);
			assignedToHosts[ecu.index].set(comp.index);

			//assign all components that need to co-locate with this

			for (Component otherComp : comp.getMustColocateWith()){
                            ecuList.add(otherComp);
                            assigned.set(otherComp.index);
                            assignedToHosts[ecu.index].set(otherComp.index);
			}

		}

		//FIXME May wish to log this
//		System.out.println("Generated deployment: " + getDepString());
	}

        @Override
	public int getHostIndex(int iCompIndex){
            int ecuID = 0;
            while (ecuID < i_Hosts){
                List<Component> ecuComponents = lst_Alloc.get(ecuID);
                for (Component comp : ecuComponents){
                    if (comp.index == iCompIndex){
                        return ecuID;
                    }
                }
                ecuID++;
            }
            return -1;
	}

	//context is an unnecessary parameter but it makes it override the method in Architecture
	@Override
	public List<Component> getComponentListForHost(int iHostIndex, EmbeddedSystemProblem oProblem){
            return lst_Alloc.get(iHostIndex);
	}

        @Override
	public String getSerializedString(){
            String sRet = new String();
            sRet = i_Hosts + "," + i_Components + ": " + System.getProperty("line.separator");
            for (int i = 0; i < lst_Alloc.size(); i++){
                sRet = sRet + "\t" + i + "{";
                for (Component component : lst_Alloc.get(i)){
                    sRet += component.index + ",";
                }
                sRet += "}" + System.getProperty("line.separator");
            }
            return sRet;
	}

	@Override
	public boolean isSameAs(Architecture other){
            //lst_Alloc.size() is number of hosts in this case
            for (int i = 0; i < lst_Alloc.size(); i++){
                List<Component> thisCompList = lst_Alloc.get(i);
                List<Component> otherCompList = ((ECUBasedDeployment) other).lst_Alloc.get(i);
                if (thisCompList.size() != otherCompList.size()){
                    return false;
                }

                for (Component component : otherCompList){
                    if (!thisCompList.contains(component)){
                        return false;
                    }
                }
            }
            return true;
	}

	//component-to-host (i.e. comp is int[0] and host is int[1]
	public Pair<Integer> getNextAssignedPair(int number){
            int i = 0;
            int sum = 0;
            while (i < lst_Alloc.size()){
                if (lst_Alloc.get(i).isEmpty()){
                    i++;
                    continue;
                }
                if (sum + lst_Alloc.get(i).size() > number){
                    break;
                }
                sum += lst_Alloc.get(i).size();
                i++;
            }

            return new Pair<Integer>(lst_Alloc.get(i).get(number - sum).index, i);
	}

        @Override
	public void attemptRepair(EmbeddedSystemProblem problem) throws OutOfOptionsException{
            repairMemViolation((Deployment) problem);
	}

	public void repairMemViolation(Deployment problem) throws OutOfOptionsException{
		System.arraycopy(allMinusOnes, 0, memoryAvailabilities, 0, i_Hosts);		
		List<Component> colocations = null;
		boolean violationSolved = false;
		for (int i = 0; i < lst_Alloc.size(); i++)
		{
                    violationSolved = false;

                    if (memoryAvailabilities[i] == -1)
                            memoryAvailabilities[i] = problem.getMemoryAvailabilityForHost(i, this);

                    if (memoryAvailabilities[i] >= 0)
                    {
                            continue;
                    }

                    //groups are more likely found at the beginning of the lists
                    for (int candComp = 0; candComp < lst_Alloc.get(i).size(); candComp++)
                    {

                    Component cand = lst_Alloc.get(i).get(candComp);

                    int memReq = cand.getIntProperty(Component.MEMORY);

                    if (cand.mustColocate())
                    {
                            colocations = cand.getMustColocateWith();
                            for (Component colcomp : colocations)
                            {
                                    memReq += colcomp.getIntProperty(Component.MEMORY);
                            }
                    }

                    int j = (i+1)%cand.getCanUseHostListLength();
                    do
                    {
                            Host candHost = cand.getHostOption(j);
                            j = (j+1)%cand.getCanUseHostListLength();
                            //if (candHost.index == j)// i think this should be i
                            if (candHost.index == i)
                            {
                                    continue;
                            }
                            if (memoryAvailabilities[candHost.index] == -1)
                                    memoryAvailabilities[candHost.index] = problem.getMemoryAvailabilityForHost(candHost.index, this);
                            if (memoryAvailabilities[candHost.index] < memReq){
                                    continue;
                            }

                            if (cand.conflicts(lst_Alloc.get(candHost.index)))
                            {
                                    continue;
                            }
                            lst_Alloc.get(candHost.index).add(cand);
                            memoryAvailabilities[candHost.index] -= cand.getIntProperty(Component.MEMORY);
                            lst_Alloc.get(i).remove(cand);
                            memoryAvailabilities[i] += cand.getIntProperty(Component.MEMORY);
                            //no option but to start from the beginning
                            candComp = 0;
                    /*	if (cand.mustColocate())
                            {
                                    for (Component colcomp : colocations)
                                    {
                                            lst_Alloc.get(candHost.index).add(colcomp);
                                            lst_Alloc.get(i).remove(colcomp);

                                    }
                            }*/
                            //have found allocation for this component, find next comp
                            break;

                        }while (j != i);


                        if (memoryAvailabilities[i] >= 0)
                        {
                                violationSolved = true;
                                break;
                        }

                    }

                    if (!violationSolved){
                        throw new OutOfOptionsException("Can't fix memory constraint violation for host" + i);
                    }
		}
	}

	public void assignComponent(Component comp, int ECUindex){
            this.lst_Alloc.get(ECUindex).add(comp);
	}

	public void assignComponentList(int whichHost, List components){
            lst_Alloc.set(whichHost, components);
	}

}


