package archeopterix.problem.deployment.solution;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;
import java.util.Vector;

import archeopterix.conf.Conf;
import archeopterix.evaluation.Attribute;
import archeopterix.problem.Component;
import archeopterix.problem.Host;
import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.problem.deployment.Deployment;
import archeopterix.solution.Architecture;
import archeopterix.solution.OutOfOptionsException;
import archeopterix.solution.Repairable;

public final class GeneralDeploymentSolution extends DeploymentSolution implements Repairable
{	
	private static final long serialVersionUID = 1L;
	private List<Host> compAlloc;
	private ArrayList<LinkedList<Component>> ecuAlloc;
	
        public List<Host> getAllocation(){
            return compAlloc;
        }
	public GeneralDeploymentSolution(int ecuCount, int compCount){
            i_Hosts		= ecuCount;
            i_Components= compCount;
            compAlloc = new Vector<Host>();
            ecuAlloc = new ArrayList<LinkedList<Component>>();
            for (int i = 0; i < i_Hosts; i++) ecuAlloc.add(new LinkedList<Component>());
            for (int i = 0; i < i_Components; i++) compAlloc.add(null);
	}
	
	protected GeneralDeploymentSolution(int ecuCount, int compCount, EmbeddedSystemProblem problem, boolean doRandomlyAllocate) {
            this(ecuCount,compCount);
            if (doRandomlyAllocate) randomlyAllocate(problem.getCompsSortedByPossibleAssignments());
	}
	
	public GeneralDeploymentSolution(EmbeddedSystemProblem problem){
            this(problem.getECUCount(),problem.getComponentCount());
	}
	
	public GeneralDeploymentSolution(EmbeddedSystemProblem problem, boolean doRandomlyAllocate){
            this(problem.getECUCount(), problem.getComponentCount());
            if (doRandomlyAllocate) randomlyAllocate(problem.getCompsSortedByPossibleAssignments());
	}

	public GeneralDeploymentSolution (GeneralDeploymentSolution solution){
            super(solution);
            i_Hosts = solution.i_Hosts;
            i_Components = solution.i_Components;
            d_Fitness = solution.d_Fitness;
		
            compAlloc = new Vector<Host>();
            ecuAlloc = new ArrayList<LinkedList<Component>>();
	
            compAlloc.addAll(solution.compAlloc);
            ecuAlloc.addAll(solution.ecuAlloc);		
            lst_Attributes = new LinkedList<Attribute>();
            for(Attribute att:solution.getAttributeList()){
                lst_Attributes.add(new Attribute(att));
            }
	}
	
	public GeneralDeploymentSolution(String solString, EmbeddedSystemProblem problem){
            initializeFromString(solString, problem);
	}
	
	
	@Override
	public void initializeFromString(String solString, EmbeddedSystemProblem problem){
            String[] arrString = solString.split(",");
            i_Hosts = Integer.parseInt(arrString[0]);
            i_Components = Integer.parseInt(arrString[1]);

            if ( arrString.length != i_Components+2 )
                System.out.println("String Sizes not mach! " + arrString.length + "," + i_Components+2);

            d_Fitness	= 0.0;
            lst_Attributes = null;

            compAlloc = new Vector<Host>();
            ecuAlloc = new ArrayList<LinkedList<Component>>();
            for (int i = 0; i < i_Hosts; i++) ecuAlloc.add(new LinkedList<Component>());
            for (int i = 0; i < i_Components; i++) compAlloc.add(null);

            for (int i = 2; i< arrString.length; i++){
                int index = Integer.parseInt(arrString[i]);
                set(problem.getComponent(i-2),index == -1? null: problem.getECU(index));
            }
	}
	
	
        @Override
	public String getSerializedString(){
            String info = i_Hosts+ "," + i_Components + ",";
            for (Host e: compAlloc)
                if(e == null)info = info + "-1" + ",";
                else info = info + e.index + ",";
            return info;
	}
	
	public String getSerializedStringEcus(){
            String info = i_Hosts + "," + i_Components + ": "+System.getProperty("line.separator") ;
            for (int i=0; i< ecuAlloc.size(); i++){
                info = info + "\t"+i+"{";
                for (Component c: ecuAlloc.get(i)) info += c.index + ",";
                    info += "}"+System.getProperty("line.separator");
            }
            return info;
	}
	
	@Override
	public GeneralDeploymentSolution clone(){
            return new GeneralDeploymentSolution(this);
	}
	
	
	public void set( int compIndex, int ecuIndex, EmbeddedSystemProblem problem){
            if(ecuIndex == -1) set(problem.getComponent(compIndex), null);
            else set(problem.getComponent(compIndex),problem.getECU(ecuIndex));
	}

	
	public void set(Component comp, Host ecu){	
            Host oldEcu = compAlloc.get(comp.index);
            if(oldEcu != null) ecuAlloc.get(oldEcu.index).remove(comp);
            if(ecu != null) ecuAlloc.get(ecu.index).add(comp);
            compAlloc.set(comp.index, ecu);
	}
	
	@Override
	public void reset(){
            for (int i = 0; i < i_Components; i++) compAlloc.set(i, null);
            for (int i=0; i< i_Hosts ; i++) ecuAlloc.get(i).clear();
	}
	
	
	@Override
	public boolean isSameAs(Architecture other){
            if(compAlloc.size() != ((GeneralDeploymentSolution)other).compAlloc.size()) return false;
		
            for (int i = 0; i < compAlloc.size(); i++)
            if(compAlloc.get(i).index != ((GeneralDeploymentSolution)other).compAlloc.get(i).index)
                return false;
	
            return true;
	}
	
	@Override
	public int getHostIndex(int compIndex){
            Host ecu = compAlloc.get(compIndex);
            return ecu == null ? -1 : ecu.index;
	}
	
	
	public Host getHost(int compIndex){
            return compAlloc.get(compIndex);
	}
	
	public List<Integer> getComponentsIndexList(int hostIndex){
            List<Integer> compIndexes = new Vector<Integer>(); 
            for(Component c:ecuAlloc.get(hostIndex)) compIndexes.add(c.index);
		return compIndexes;
	}
	
	
	@Override
	public List<Component> getComponentListForHost(int hostIndex, EmbeddedSystemProblem problem){
            return ecuAlloc.get(hostIndex);
	}
	

	public void randomlyAllocate(EmbeddedSystemProblem problem){
            Random rnd = Conf.instance().random;
            for(int i=0; i <i_Components; i++)
                set(problem.getComponent(i), problem.getECU(rnd.nextInt(i_Hosts)));
	}
	
	
	@Override
	public void randomlyAllocate(List<Component> lstSortedComponents) {
		//Adapted from same method in ECUBasedDeployment
		Random random = Conf.instance().random;
		BitSet[] assignedToHosts = new BitSet[i_Hosts];
		for (int i = 0; i < i_Hosts; i++)
			assignedToHosts[i] = new BitSet(i_Components);
		BitSet assigned = new BitSet(i_Components);
		Iterator<Component> comp_iterator = lstSortedComponents.iterator();

		// assign ECUs with few options first
		while (comp_iterator.hasNext())
		{
			Component comp = comp_iterator.next();

			if (assigned.get(comp.index))
				continue;

			boolean okToAssign;
			Host ecu;
			do
			{
				int opt = random.nextInt(comp.getCanUseHostListLength());
				ecu = comp.getHostOption(opt);

				okToAssign = true;
				//TODO Wrap up checking functionality in a method
				//check if components in the same group have colocate restrictions with other comp.
				Iterator<Component> mustNotColocateWith = comp.getMustNotColocateWith().iterator();
				while (mustNotColocateWith.hasNext() && okToAssign)
					okToAssign = ! assignedToHosts[ecu.index].get(mustNotColocateWith.next().index);
				Iterator<Component> buddies = comp.getMustColocateWith().iterator();
				while (buddies.hasNext() && okToAssign)
				{
					mustNotColocateWith = buddies.next().getMustNotColocateWith().iterator();
					while (mustNotColocateWith.hasNext() && okToAssign)
						okToAssign = ! assignedToHosts[ecu.index].get(mustNotColocateWith.next().index);
				}

				//FIXME May wish to log this information 
				//                    if (! okToAssign)
				//                        System.out.println("Disallowed component " + comp.index + " from ecu " + ecu.index + " because it is assigned: " + assignedToHosts[ecu.index]);

			} while (! okToAssign);

			set(comp, ecu);
			assigned.set(comp.index);
			assignedToHosts[ecu.index].set(comp.index);

			//assign all components that need to co-locate with this

			for (Component otherComp : comp.getMustColocateWith())
			{
				set(otherComp, ecu);
				assigned.set(otherComp.index);
				assignedToHosts[ecu.index].set(otherComp.index);
			}

		}

	}
	

        @Override
	public void attemptRepair(EmbeddedSystemProblem problem) throws OutOfOptionsException
	{
		Deployment dep = (Deployment)problem;
		boolean[] hostDone = new boolean[i_Hosts];
		int memViol = 0;
		List<Component> colocations = null;
		List<Integer> componentsForHost;
		boolean violationSolved = false;
		for (int i = 0; i < compAlloc.size(); i++)
		{
			//traverse by hosts; some have already been dealt with
			if (hostDone[compAlloc.get(i).index])
				continue;
			violationSolved = false;

			memViol = dep.getMemoryAvailabilityForHost(compAlloc.get(i).index, this);
			if (memViol >= 0)
				continue;

			componentsForHost = getComponentsIndexList(compAlloc.get(i).index);

			//groups are more likely found at the beginning of the lists
			for (int candComp = 0; candComp < componentsForHost.size(); candComp++)
			{

				Component cand = problem.getComponent(componentsForHost.get(candComp));

				int memReq = cand.getIntProperty(Component.MEMORY) ;

				if (cand.mustColocate())
				{
					colocations = cand.getMustColocateWith();
					for (Component colcomp : colocations)
					{
						memReq += colcomp.getIntProperty(Component.MEMORY);
					}
				}


				for (int j = 0; j < cand.getCanUseHostListLength(); j++)
				{
					Host candHost = cand.getHostOption(j);
					if (candHost.index == compAlloc.get(i).index)
						continue;
					if (dep.getMemoryAvailabilityForHost(candHost.index, this) < memReq)
						continue;
					if (cand.conflicts(getComponentListForHost(candHost.index, problem)))
						continue;
					set(cand, candHost);


					if (cand.mustColocate())
					{
						for (Component colcomp : colocations)
						{
							set(colcomp, candHost);
						}
					}
					//have found allocation for this component, find next comp
					break;
				}


				memViol = dep.getMemoryAvailabilityForHost(compAlloc.get(i).index, this);

				if (memViol >= 0)
				{
					violationSolved = true;
					hostDone[compAlloc.get(i).index] = true;
					break;
				}

			}

			if (!violationSolved){
                            //  System.out.println(getDepString());
                            throw new OutOfOptionsException("Can't fix memory constraint violation for host"+i);
			}
		}
	}
}
