package archeopterix.problem.deployment.solution;

import java.util.BitSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;

import archeopterix.conf.Conf;
import archeopterix.problem.Component;
import archeopterix.problem.Host;
import archeopterix.problem.InteractionProperties;
import archeopterix.problem.NetworkProperties;
import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.problem.deployment.Deployment;
import archeopterix.solution.Architecture;
import archeopterix.solution.OutOfOptionsException;
import archeopterix.solution.Repairable;
import java.util.Vector;

public final class OriginalDeploymentSolution extends DeploymentSolution implements Repairable {
    private static final long serialVersionUID = 1L;
    private List<Integer> lst_Alloc;
    public OriginalDeploymentSolution(int iHosts, int iComps) {
        this(iHosts, iComps, true);
    }

    public OriginalDeploymentSolution(EmbeddedSystemProblem problem) {
        this(problem.getECUCount(), problem.getComponentCount(), false);
        randomlyAllocate( problem.getCompsSortedByPossibleAssignments() );
    }

    protected OriginalDeploymentSolution(int iHosts, int iComps, boolean doRandomlyAllocate) {
        i_Hosts = iHosts;
        i_Components = iComps;
        lst_Alloc = new Vector<Integer>();
        for (int i = 0; i < iComps; i++)
            lst_Alloc.add(-1);
            if (doRandomlyAllocate)
                randomlyAllocate();
    }

    @Override
    public void reset() {
        for (int i = 0; i < i_Components; i++)
            set(i, -1);
    }

    public OriginalDeploymentSolution(String sDepString){
        initializeFromString(sDepString, null);
    }

    public OriginalDeploymentSolution (OriginalDeploymentSolution oDep)
    {
        super(oDep);
        i_Hosts = oDep.i_Hosts;
        i_Components = oDep.i_Components;
        d_Fitness = oDep.d_Fitness;
        lst_Alloc = new LinkedList<Integer>();
        for (int i=0; i < oDep.lst_Alloc.size(); i++){
            int iH = oDep.lst_Alloc.get(i).intValue();
            lst_Alloc.add(iH);
        }
        lst_Attributes = null;
    }

    @Override
    public String getSerializedString(){
        String sRet = "";
        sRet = i_Hosts+ "," + i_Components + ",";
        for (int i=0; i< lst_Alloc.size(); i++){
            sRet = sRet + lst_Alloc.get(i).toString()+ ",";
        }
        return sRet;
    }

    public OriginalDeploymentSolution crossover(Architecture oArch, int iIndex){
        if (! (oArch instanceof OriginalDeploymentSolution))
            return null;
        OriginalDeploymentSolution oDep = (OriginalDeploymentSolution)oArch;
        OriginalDeploymentSolution oOffspring = new OriginalDeploymentSolution(i_Hosts, i_Components);
        for (int i=0; i < iIndex; i++){
            int iH = (oDep.lst_Alloc.get(i)).intValue();
            oOffspring.set(i, iH);
        }
        for (int i=iIndex; i<i_Components;i++){
            int iH = (lst_Alloc.get(i)).intValue();
            oOffspring.set(i, iH);
        }
        return oOffspring;	
    }

    @Override
    public int getNoOfMutations(){
        return 2;
    }

    public void randomlyAllocate(){
        Random rnd = Conf.instance().random;
        for(int i=0; i <i_Components; i++){
            set(i, rnd.nextInt(i_Hosts));
        }
    }

    public void set( int iCompIndex, int iHostIndex ){
        lst_Alloc.set(iCompIndex, iHostIndex);
    }

    public void set( Component oComp, Host oHost ){
        lst_Alloc.set(oComp.index, oHost.index);
    }

    public boolean validateBWConstraint(EmbeddedSystemProblem oContext){
        for(int obj=0; obj< Conf.instance().getNObj();obj++){
            for(int iHi=0; iHi<i_Hosts; iHi++){
                for(int iHj=0; iHj<i_Hosts; iHj++){
                    if( ! checkBWCapacity(iHi, iHj, oContext))
                        return false;
                }
            }
        }
        return true;
    }

    private boolean checkBWCapacity(int iHi, int iHj, EmbeddedSystemProblem oContext){
        LinkedList<Integer> lstCompsi = getComponentsIndexList(iHi);
        LinkedList<Integer> lstCompsj = getComponentsIndexList(iHj);
        int iDRij 	= oContext.getNetworkProperties(iHi, iHj).getInt( NetworkProperties.DATA_RATE_KBPS );

        for(int i=0; i< lstCompsi.size(); i++){
            int iCi = lstCompsi.get(i);
            for (int j=0; j<lstCompsj.size(); j++){
                int iCj = lstCompsj.get(j);
                if (oContext.areInteracting(iCi, iCj)){
                    int iCLij 	= oContext.getInteractionProperties(iCi, iCj).getIntFromList( InteractionProperties.COMM_LOAD_KB, 0 ); //SERVICE 0
                    double dFRij 	= oContext.getInteractionProperties(iCi, iCj).getDouble( InteractionProperties.FREQUENCY  );
                    iDRij -= (iCLij* dFRij);
                }
                if (iDRij <= 0){
                    return false;
                }
            }
        }
        return true;
    }

    public LinkedList<Integer> getComponentsIndexList( int iHostIndex ){
        LinkedList<Integer> lstComps = new LinkedList<Integer>();
        for ( int iComp=0; iComp<i_Components; iComp++){
            Integer  iVal = lst_Alloc.get(iComp);
            if ( iVal.intValue() == iHostIndex) {
                lstComps.add(iComp);
            }
        }		
        return lstComps;
    }

    public int getComponentCount( int iHostIndex ){
        int iUsage = 0;
        for ( int iComp=0; iComp<i_Components; iComp++){
            Integer  iVal = lst_Alloc.get(iComp);
            if ( iVal.intValue() == iHostIndex) {
                iUsage++;
            }
        }		
        return iUsage;
    }

    @Override
    public OriginalDeploymentSolution clone() {
        OriginalDeploymentSolution oDep = new OriginalDeploymentSolution(this);
        return oDep;
    }

    @Override
    public void randomlyAllocate(List<Component> lstSortedComponents) {
        //Adapted from same method in ECUBasedDeployment
        Random random = Conf.instance().random;
        BitSet[] assignedToHosts = new BitSet[i_Hosts];
        for (int i = 0; i < i_Hosts; i++)
            assignedToHosts[i] = new BitSet(i_Components);
            BitSet assigned = new BitSet(i_Components);
            Iterator<Component> comp_iterator = lstSortedComponents.iterator();

		// assign ECUs with few options first
		while (comp_iterator.hasNext())
		{
			Component comp = comp_iterator.next();

			if (assigned.get(comp.index))
				continue;

			boolean okToAssign;
			Host ecu;
			do
			{
				int opt = random.nextInt(comp.getCanUseHostListLength());
				ecu = comp.getHostOption(opt);

				okToAssign = true;
				//TODO Wrap up checking functionality in a method
				//check if components in the same group have colocate restrictions with other comp.
				Iterator<Component> mustNotColocateWith = comp.getMustNotColocateWith().iterator();
				while (mustNotColocateWith.hasNext() && okToAssign)
					okToAssign = ! assignedToHosts[ecu.index].get(mustNotColocateWith.next().index);
				Iterator<Component> buddies = comp.getMustColocateWith().iterator();
				while (buddies.hasNext() && okToAssign)
				{
					mustNotColocateWith = buddies.next().getMustNotColocateWith().iterator();
					while (mustNotColocateWith.hasNext() && okToAssign)
						okToAssign = ! assignedToHosts[ecu.index].get(mustNotColocateWith.next().index);
				}

				//FIXME May wish to log this information 
				//                    if (! okToAssign)
				//                        System.out.println("Disallowed component " + comp.index + " from ecu " + ecu.index + " because it is assigned: " + assignedToHosts[ecu.index]);

			} while (! okToAssign);

			lst_Alloc.set(comp.index, ecu.index);
			assigned.set(comp.index);
			assignedToHosts[ecu.index].set(comp.index);

			//assign all components that need to co-locate with this

			for (Component otherComp : comp.getMustColocateWith())
			{
				lst_Alloc.set(otherComp.index, ecu.index);
				assigned.set(otherComp.index);
				assignedToHosts[ecu.index].set(otherComp.index);
			}

		}

	}

	@Override
	public List<Component> getComponentListForHost(int iHostIndex, EmbeddedSystemProblem oProblem) {
		List<Component> lstComps = new Vector<Component>();
		for (int iComp = 0; iComp < i_Components; iComp++) {
			if (lst_Alloc.get(iComp) == iHostIndex)
				lstComps.add(oProblem.getComponent(iComp));
		}
		return lstComps;
	}

	@Override
	public int getHostIndex(int iCompIndex) {
		return lst_Alloc.get(iCompIndex).intValue();
	}

	@Override
	public void initializeFromString(String sDepString, EmbeddedSystemProblem context) {
            String[] aSt = sDepString.split(",");
            i_Hosts = Integer.parseInt(aSt[0]);
            i_Components = Integer.parseInt(aSt[1]);
            if ( aSt.length != i_Components+2 ){
                System.out.println("String Sizes not mach! " + aSt.length + "," + i_Components+2 );
            }
            d_Fitness	= 0.0;
            lst_Alloc = new LinkedList<Integer>();
            lst_Attributes = null;
            for ( int i=2; i< aSt.length ; i++){
                lst_Alloc.add(Integer.parseInt(aSt[i]));
            }
	}

	@Override
	public boolean isSameAs(Architecture other){
            //lst_Alloc.size() is number of hosts in this case
            for (int i = 0; i < lst_Alloc.size(); i++){
                Integer thisCompList = lst_Alloc.get(i);
                Integer otherCompList = ((OriginalDeploymentSolution)other).lst_Alloc.get(i);
                if (thisCompList != otherCompList)
                    return false;
            }
            return true;
	}



        @Override
	public void attemptRepair(EmbeddedSystemProblem problem) throws OutOfOptionsException
	{
		Deployment dep = (Deployment)problem;
		boolean[] hostDone = new boolean[i_Hosts];
		int memViol = 0;
		List<Component> colocations = null;
		List<Integer> componentsForHost = null;
		boolean violationSolved = false;
		for (int i = 0; i < lst_Alloc.size(); i++)
		{
			//traverse by hosts; some have already been dealt with
			if (hostDone[lst_Alloc.get(i)])
				continue;
			violationSolved = false;

			memViol = dep.getMemoryAvailabilityForHost(lst_Alloc.get(i), this);
			if (memViol >= 0)
				continue;

			componentsForHost = getComponentsIndexList(lst_Alloc.get(i));

			//groups are more likely found at the beginning of the lists
			for (int candComp = 0; candComp < componentsForHost.size(); candComp++)
			{

				Component cand = problem.getComponent(componentsForHost.get(candComp));

				int memReq = cand.getIntProperty(Component.MEMORY) ;

				if (cand.mustColocate())
				{
					colocations = cand.getMustColocateWith();
					for (Component colcomp : colocations)
					{
						memReq += colcomp.getIntProperty(Component.MEMORY);
					}
				}


				for (int j = 0; j < cand.getCanUseHostListLength(); j++)
				{
					Host candHost = cand.getHostOption(j);
					if (candHost.index == lst_Alloc.get(i))
						continue;
					if (dep.getMemoryAvailabilityForHost(candHost.index, this) < memReq)
						continue;
					if (cand.conflicts(getComponentListForHost(candHost.index, problem)))
						continue;
					set(cand.index, candHost.index);


					if (cand.mustColocate())
					{
						for (Component colcomp : colocations)
						{
							set(colcomp.index, candHost.index);
						}
					}
					//have found allocation for this component, find next comp
					break;
				}


				memViol = dep.getMemoryAvailabilityForHost(lst_Alloc.get(i), this);

				if (memViol >= 0)
				{
					violationSolved = true;
					hostDone[lst_Alloc.get(i)] = true;
					break;
				}

			}

			if (!violationSolved)
			{
				//  System.out.println(getDepString());
				throw new OutOfOptionsException("Can't fix memory constraint violation for host"+i);
			}
		}
	}

    /** Initialize the allocation from double array
     *
     * @param allocation Comp->H allocation
     */
    public void intializeAllocation(int[] allocation) {
        for (int iC=0; iC < allocation.length; iC++)
            lst_Alloc.set(iC, allocation[iC] );
    }

}
