package archeopterix.solution;



import archeopterix.algorithm.learning.utils.Parameter;
import archeopterix.conf.Conf;
import archeopterix.constraints.ConstraintHandler;
import java.util.List;
import archeopterix.evaluation.Attribute;
import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.solution.comparator.MOSolutionRanking;
import archeopterix.util.CommonFormats;
import archeopterix.util.PropertyMap;
import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;

public abstract class Architecture implements Serializable {
    private static final long serialVersionUID = 1L;

    public double d_Fitness;

    /** The front to which this solution belongs, 1.. .*/
    protected int rank = 1;
    /** This solution's crowding distance measure within its front. */
    protected double crowdingDistance = 0;
    public Object 	m_UserData = null;
    private PropertyMap m_PropertyMap = null;
    protected List<Attribute> lst_Attributes = null;
    private boolean isFeasible = false;
    protected double violations = 0;
    private HashMap<Object, Parameter> parameterMap;
    
    private MOSolutionRanking ranking;
    private int[] MSOPSRanks;
    
    public int getMSOPSRank(int index){return MSOPSRanks[index];}
    public void setMSOPSRank(int fitness, int index, int targets){
        if(MSOPSRanks==null)
            MSOPSRanks = new int[targets];
        MSOPSRanks[index]= fitness;
    }
    public MOSolutionRanking getRanking(){return ranking;}
    public Architecture() {
        m_PropertyMap = new PropertyMap();
    }
    public Object getUserData(){
        return m_UserData;
    }
    public HashMap<Object, Parameter> getParameterMap(){return parameterMap;}
    public final void setParameterMap(HashMap<Object, Parameter> params){
        parameterMap = new HashMap<Object, Parameter>();
        for(Map.Entry pEntry : params.entrySet()){
            Parameter newParam = ((Parameter)pEntry.getValue()).clone();
            parameterMap.put(pEntry.getKey(), newParam);
        }
    }
    public void setProperty(String sKey, Object oVal){
        m_PropertyMap.put(sKey, oVal);
    }

    public Object getProperty(String sKey){
        return m_PropertyMap.get(sKey);
    }

    //FIXME Wouldn't making this class Cloneable be more effective?
    protected Architecture (Architecture oArch) {
        m_UserData = oArch.m_UserData;
        d_Fitness = oArch.d_Fitness;
        lst_Attributes = null;
        if(Conf.instance().getAlgorithmClass().getName().contains("Adaptive"))
            setParameterMap(oArch.getParameterMap());
    }
    public boolean isFeasible() { return isFeasible; }

    public void setFeasible(boolean isFeasible) { this.isFeasible = isFeasible; }

    public int getRank() { return rank; }

    public void setRank(int rank) { this.rank = rank; }

    public double getCrowdingDistance() { return crowdingDistance; }

    public void setCrowdingDistance(double crowdingDistance) { this.crowdingDistance = crowdingDistance; }

    public void incrementCrowdingDistance(double delta) { this.crowdingDistance += delta; }

    public void setConstraintViolation (EmbeddedSystemProblem problem) {
        violations=0;
        for (ConstraintHandler constraint : problem.getListConstraints()) {
            violations+=constraint.quantifySatisfaction(this);
        }   
    }

    public double getConstraintViolation (){return this.violations;}

    //FIXME But this doesn't describe the solution
    @Override
    public String toString() {
        String sRet = getSerializedString();//"";
        if (lst_Attributes == null)
            sRet = "Non-evaluated " + getClass().getSimpleName();
        else {
            sRet+=":ObjectiveValues:";
            for (Attribute attrib : lst_Attributes)
                sRet += CommonFormats.dFormatter.format( attrib.getValue() ) +  ":";
        }
        return sRet;
    }

    public void setAttributeList(List<Attribute> lstAttributes) {
        lst_Attributes = lstAttributes;
    }

    public List<Attribute> getAttributeList() {
        return lst_Attributes;
    }
    public double getAttributeValue(int iAttributeIndex) {
        return lst_Attributes.get(iAttributeIndex).getNormalisedOptimiserValue();
    }

    public double getNormalAttributeValue(int iAttributeIndex) {
        return lst_Attributes.get(iAttributeIndex).getValue();
    }

    public int getNoOfMutations() {
        return 1;
    }

    //TODO: needs to go into separate constraint handling class
    public double getNormalisedMemoryPenalty(EmbeddedSystemProblem context){
        return violations/context.getMaxMemoryConstraint();
    }

    /** Returns whether this Architecture is dominated by {@code other}. */
    public boolean isDominated(Architecture other) {
        List<Attribute> oAttListDep2 = other.getAttributeList();
        for (int i = 0, c = lst_Attributes.size(); i < c; i++) {
            if (lst_Attributes.get(i).getOptimizerValue() > oAttListDep2.get(i).getOptimizerValue() )
                return false;
        }
        return true;
    }

    public boolean isConstrainDominated(Architecture other) {
        if (this.isFeasible()) {
            if (!other.isFeasible())
                return false;
            return this.isDominated(other);
        } else {
        if (other.isFeasible())
            return true;
            //violation is negative
        if (this.violations < other.violations)
            return true;
        }
        return false;
    }

    //uses Wolfesenbet & Yen "Constraint handling in mo eo"
    public boolean isPenaltyDominated(EmbeddedSystemProblem context, Architecture other, double feasibilityRatio) {
            List<Attribute> oAttListDep2 = other.getAttributeList();
            //large values are good; cannot have negatives because of the squaring
            double thisPenaltyValue = Math.abs(this.getNormalisedMemoryPenalty(context));
            double otherPenaltyValue = Math.abs(other.getNormalisedMemoryPenalty(context));
            double d_i_ofThis ;
            double d_i_ofOther ;
            double p_i_ofThis;
            double p_i_ofOther;

            if (feasibilityRatio == 0) { //there are no feasible individuals
                //equation 6
                d_i_ofThis = thisPenaltyValue;
                d_i_ofOther = otherPenaltyValue;
                //equation 8 has no effect when there are no feasible individuals

                if (d_i_ofThis < d_i_ofOther )
                        return false;
            } else { //there are feasible individuals
                for (int i = 0, c = lst_Attributes.size(); i < c; i++) {
                    //equation 6
                    double thisObjectiveValue = lst_Attributes.get(i).getNormalisedOptimiserValue(context, this);
                    double otherObjectiveValue = oAttListDep2.get(i).getNormalisedOptimiserValue(context, other);

                    //adjustment for neg objective values because of imminent squaring:
                    // adjust them so we can minimise

                    if (thisObjectiveValue < 0) {
                        thisObjectiveValue = Math.abs(thisObjectiveValue);
                        otherObjectiveValue = Math.abs(otherObjectiveValue);
                    } else {
                        thisObjectiveValue = 1- thisObjectiveValue;
                        otherObjectiveValue = 1- otherObjectiveValue;
                    }
                    d_i_ofThis = Math.sqrt(Math.pow(thisObjectiveValue, 2)+ Math.pow(thisPenaltyValue, 2));
                    d_i_ofOther = Math.sqrt(Math.pow(otherObjectiveValue, 2)+ Math.pow(otherPenaltyValue, 2));
                    p_i_ofThis = (1 - feasibilityRatio) * thisPenaltyValue;
                    p_i_ofOther = (1 - feasibilityRatio) * otherPenaltyValue;
                    if (!this.isFeasible())
                        p_i_ofThis += feasibilityRatio * thisObjectiveValue;
                    if (!other.isFeasible())
                        p_i_ofOther += feasibilityRatio * otherObjectiveValue;
                    if ((d_i_ofThis + p_i_ofThis) < (d_i_ofOther + p_i_ofOther) )
                            return false;
                }
            }
            return true;
    }

    //FIXME Add javadoc comments to all of these abstract methods to explain their expected behaviour
    @Override
    public abstract Architecture clone();
    public abstract boolean isSameAs(Architecture other);
    public abstract void reset();
    public abstract String getSerializedString();
    public abstract void initializeFromString(String sStr,EmbeddedSystemProblem oProblem);
}