package archeopterix.util.io;

import java.io.File;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import archeopterix.conf.Conf;
import archeopterix.conf.RandomSeedSource;
import archeopterix.evaluation.ArchitectureEvaluator;
import archeopterix.problem.SchedulingStratergy;
import archeopterix.util.PropertyMap;

/**
 * Parses XML representations of the {@link Conf} class, setting its fields
 * using the data therein.
 */
public final class ConfXMLParser extends XMLParser {
    //Possibly temporarily disabled, since trying to allow configuration to specify log file name
//	private static final Log log = LogFactory.getLog(ConfXMLParser.class);

    /**
     * Assumed package for evaluators
     */
    public static final String DEFAULT_EVALUATOR_PACKAGE = "archeopterix.evaluation.";
    public static final String CONFIGURATION = "configuration";
    public static final String PROBLEM = "problem";
    public static final String PROBLEM_TYPE = "type";
    public static final String PROBLEM_DEFINITION = "definition";
    public static final String ATTRIB_DEF_VARIES = "varies";
    public static final String OBJECTIVES = "objectives";
    public static final String OBJECTIVE = "objective";
    public static final String ATTRIB_ACTIVE = "active";
    public static final String ATTRIB_ID = "id";
    public static final String ATTRIB_MAX_VALUE = "maxvalue";
    public static final String ATTRIB_MIN_VALUE = "minvalue";
    public static final String SOLUTION_REP = "solutionRep";
    public static final String SCHEDULING_STRATERGY = "schedulingStratergy";
    public static final String RUN = "run";
    public static final String ATTRIB_REPEATS = "repeats";
    public static final String ALGORITHM = "algorithm";
    public static final String RANDOM_SEEDS = "randomSeedSource";
    public static final String RANDOM_SEED_START = "firstRandomSeed";
    public static final String LOG_NAME_PREFIX = "logFileNamePrefix";
    public static final String POPULATION_FILE_NAME = "populationFileName";
    
    public static final String PARAMETER_CONTROL = "parameterControl";
    public static final String ALGO_PARAMETER = "algoParameter";
    public static final String ALGO_PARAMETER_NAME ="name";
    public static final String ALGO_PARAMETER_MIN_VALUE ="minvalue";
    public static final String ALGO_PARAMETER_MAX_VALUE = "maxvalue";
    public static final String ALGO_PARAMETER_ACTIVE = "active";
    public static final String ALGO_PARAMETER_INTERVALS = "intervals"; 
    public static final String ALGO_PARAMETER_TYPE = "type"; 
    
    
    public static final String PARAMETERS = "parameters";
    public static final String ATTRIB_FOR = "for";
    public static final String PARAMETER = "param";

    private ConfXMLParser() {
    }

    public static void loadConfig(String sFileName, Conf conf) {

        Document dom = parseXMLFile(sFileName);
        if (dom == null) {
            throw new Error("Unable to load configuration from file; is it a bit extreme to throw an unrecoverable error?");
        }

        Element docEle = dom.getDocumentElement();

        loadAllParameters(docEle, conf);
        loadProblemDetails(docEle, conf);
        loadRunDetails(docEle, conf);
        
    }

    private static void loadProblemDetails(Element rootEl, Conf conf) {
        NodeList problemNodes = rootEl.getElementsByTagName(PROBLEM);
        if (problemNodes.getLength() != 1) {
            throw new RuntimeException("Expected exactly one problem to be defined, found " + problemNodes.getLength());
        }

        Element problemEl = (Element) problemNodes.item(0);

        //Note: Does not check for problems in XML, such as missing or too many elements of each type.

        try {
            String typeString = getFirstMatchingElementTextValue(problemEl, PROBLEM_TYPE);
            if (typeString != null) {
                conf.setProblemClass(Class.forName(typeString));
            }
        } catch (ClassNotFoundException cnfe) {
//			if (log.isErrorEnabled())
//				log.error("Could not load class for problem", cnfe);
            throw new RuntimeException(cnfe);
        }

        Element defEl = (Element) problemEl.getElementsByTagName(PROBLEM_DEFINITION).item(0);
        String definition = combineAllTextChildren(defEl);
        if (definition != null) {
            conf.setProblemFile(new File(definition));
        }
        String variesStr = defEl.getAttribute(ATTRIB_DEF_VARIES);
        conf.setProblemVaries(variesStr.isEmpty() ? false : Boolean.parseBoolean(variesStr));

        NodeList objectiveNodes = problemEl.getElementsByTagName(OBJECTIVE);
        for (int i = 0, c = objectiveNodes.getLength(); i < c; i++) {
            Element objective = (Element) objectiveNodes.item(i);
            String name = objective.getAttribute(ATTRIB_NAME);
            String idStr = objective.getAttribute(ATTRIB_ID);
            if (Boolean.parseBoolean(objective.getAttribute(ATTRIB_ACTIVE))) {
                String className = name;
                if (!name.contains(".")) {
                    className = DEFAULT_EVALUATOR_PACKAGE + name;
                }
                try {
                    ArchitectureEvaluator evaluator = conf.addEvaluator(Class.forName(className), idStr.isEmpty() ? null : new Integer(idStr));
                    String smaxValue = objective.getAttribute(ATTRIB_MAX_VALUE);
                    if (!smaxValue.isEmpty()) {
                        evaluator.setMaxValue(Double.parseDouble(smaxValue));
                    }
                    String minValue = objective.getAttribute(ATTRIB_MIN_VALUE);
                    if (!minValue.isEmpty()) {
                        evaluator.setMinValue(Double.parseDouble(minValue));
                    }

                } catch (ClassNotFoundException cnfe) {
                    throw new RuntimeException("Couldn't load evaluator named \"" + name + "\""
                            + (name.equals(className) ? "" : "; assumed fully-qualified name was " + className), cnfe);
                }
            }
        }

        String solutionRep = getFirstMatchingElementTextValue(problemEl, SOLUTION_REP);
        if (solutionRep != null) {
            try {
                conf.setSolutionRepresentation(Class.forName(solutionRep));
            } catch (ClassNotFoundException cnfe) {
                throw new RuntimeException(cnfe);
            }
        }
        String sSchedulingStratergy = getFirstMatchingElementTextValue(problemEl, SCHEDULING_STRATERGY);
        if (sSchedulingStratergy != null) {
            try {
                SchedulingStratergy sR = SchedulingStratergy.valueOf(sSchedulingStratergy);
                conf.setSchedulingStratergy(sR);
            } catch (Exception e) {
                throw new RuntimeException("Unable to obtain instance of stated enumerated type", e);
            }
        }
    }

    private static void loadRunDetails(Element rootEl, Conf conf) {
        NodeList runNodes = rootEl.getElementsByTagName(RUN);
        if (runNodes.getLength() != 1) {
            throw new RuntimeException("Expected exactly one run to be specified, found " + runNodes.getLength());
        }
        Element runEl = (Element) runNodes.item(0);

        NodeList algNodes = runEl.getElementsByTagName(ALGORITHM);
        if (algNodes.getLength() != 1) {
            throw new RuntimeException("Expected exactly one algorithm to be specified, found " + algNodes.getLength());
        }
        String algClassString = combineAllTextChildren((Element) algNodes.item(0));

        String repeatsString = runEl.getAttribute(ATTRIB_REPEATS);
        if (repeatsString.isEmpty()) {
            repeatsString = null;
        }
       

        String rsString = getFirstMatchingElementTextValue(runEl, RANDOM_SEEDS);
        String rsStartString = getFirstMatchingElementTextValue(runEl, RANDOM_SEED_START);
        String logFileNamePrefix = getFirstMatchingElementTextValue(runEl, LOG_NAME_PREFIX);
        String populationFileName = getFirstMatchingElementTextValue(runEl, POPULATION_FILE_NAME);

        try {
            conf.setAlgorithmClass(Class.forName(algClassString));
            if (repeatsString != null) {
                conf.setRunRepeats(Integer.parseInt(repeatsString));
            }
            if (rsString != null) {
                conf.setRandomSeedSource(RandomSeedSource.valueOf(rsString));
            }
            if (rsStartString != null) {
                conf.setInitialRandomSeed(Long.parseLong(rsStartString));
            }
            if (logFileNamePrefix != null) {
                conf.setLogFileNamePrefix(logFileNamePrefix);
            }
            if (populationFileName != null){
                    conf.setPopulationFileName(populationFileName);
            }

        } catch (ClassNotFoundException cnfe) {
//			if (log.isErrorEnabled())
//				log.error("Could not load class for main algorithm", cnfe);
            throw new RuntimeException(cnfe);
        }
    }

    private static void loadAllParameters(Element rootEl, Conf conf) {
        NodeList paramsNodes = rootEl.getElementsByTagName(PARAMETERS);
        for (int i = 0, c = paramsNodes.getLength(); i < c; i++) {
            loadParameters((Element) paramsNodes.item(i), conf);
        }
    }

    private static void loadParameters(Element paramsEl, Conf conf) {
        PropertyMap params = getPropertyMap(paramsEl, PARAMETER);

        try {
            conf.addParameters(Class.forName(paramsEl.getAttribute(ATTRIB_FOR)), params);
        } catch (ClassNotFoundException cnfe) {
//			if (log.isErrorEnabled())
//				log.error("Could not load class for algorithm", cnfe);
            throw new RuntimeException(cnfe);
        }
    }

}
