/**
 * 
 */
package archeopterix.util.io;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Performs textual conversion of old problem file format (based on an example
 * I believe is from early 2010). Expects file to contain details about ECUs
 * before components, before buses, before interactions. ID tags are converted
 * to string-valued properties with a name (Component for components and ECU
 * for ECUs) followed by the integer ID. It should identify and correct
 * properties with different names. Lists wrapped by parentheses have those
 * removed. Doesn't adjust the indentation to be correct. Feedback welcome.
 * 
 * @author JMontgomery
 *
 */
public class OldToNewProblemFileConverter {

	private static enum Processing { START, ECUS, COMPONENTS, BUSES, INTERACTIONS, END };
	
	private static Processing processing;
	
	public static void main(String[] args) {
		if (args.length != 1) {
			System.err.println("Usage: " + OldToNewProblemFileConverter.class.getSimpleName() + " <file_to_convert.xml>");
			System.exit(1);
		}
		
		File inFile = new File(args[0]);
		
		if (!inFile.exists()) {
			System.err.println("Given file doesn't exist: " + inFile);
			System.exit(2);
		}
		
		String fullInPath = inFile.getAbsolutePath();
		File outFile = new File( fullInPath.replaceAll("\\.xml$", "_new.xml") );
		
		//TODO Now do the conversion
		//Quick and dirty conversion (still faster than co-opting old XMLParser and changing it to put details in new Problem structure
		try {
			processing = Processing.START;
			BufferedReader in = new BufferedReader( new FileReader(inFile) );
			PrintWriter out = new PrintWriter( outFile );
			String aLine = in.readLine();
			while (aLine != null) {
				String outLine = convertLine(aLine);
				if (outLine != null)
					out.print( outLine + '\n');
				aLine = in.readLine();
			}
				
			in.close();
			out.close();
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
	}
	
	private static String convertLine(String line) {
		if (line.isEmpty()) {
			return null;
		} else {
			//Very inefficient, but will get the job done
			line = line.replace("System>", "system>");
			line = line.replace("ECU>", "ecu>");
			line = line.replace("Component>", "component>");
			line = line.replace("DataBus>", "bus>");
			line = line.replace("Interaction>", "interaction>");
			
			line = replaceProperties(line);
			
			//Then determine if have changed sections
			if (processing == Processing.START && line.indexOf("<ecu>") >= 0) {
				processing = Processing.ECUS;
				line = "<processors>\n" + line;
			} else if (processing == Processing.ECUS && line.indexOf("<component>") >= 0) {
				processing = Processing.COMPONENTS;
				line = "</processors>\n" + "<components>\n" + line;
			} else if (processing == Processing.COMPONENTS && line.indexOf("<bus>") >= 0) {
				processing = Processing.BUSES;
				line = "</components>\n" + "<buses>\n" + line;
			} else  if (processing == Processing.BUSES && line.indexOf("<interaction>") >= 0) {
				processing = Processing.INTERACTIONS;
				line = "</buses>\n" + "<interactions>\n" + line;
			} else  if (processing == Processing.INTERACTIONS && line.indexOf("</system>") >= 0) {
				processing = Processing.END;
				line = "</interactions>\n" + line;
			}
		}
		return line;
	}

	private static String replaceProperties(String line) {
		Matcher matcher = Pattern.compile("<([A-Z][^\\s]*)").matcher(line);
		if (matcher.find()) {
			line = line.replace("<" + matcher.group(1), "<property name=\"" + matcher.group(1) + "\"");
			line = line.replaceAll("</[A-Z].*>$", "</property>");
			
			//Special case for IDs
			if (line.matches(".*name=\"ID\".*")) {
				line = line.replace("type=\"integer\"", "type=\"string\"");
				String name;
				switch (processing) {
					case COMPONENTS : name = "Component"; break;
					case ECUS : name = "ECU"; break;
					default : name = "UnknownEntity"; break;
				}
				Matcher idMatcher = Pattern.compile("type=\"string\">([0-9]+)<").matcher(line);
				idMatcher.find(); //assume will find something
				line = line.replace(">" + idMatcher.group(1) + "<", ">" + name + idMatcher.group(1) + "<");
			}
			
			line = replaceAlteredNames(line);
			line = removeParenthesesAroundLists(line);
		}
		return line;
	}
	
	private static String replaceAlteredNames(String line) {
		line = line.replace("FailureIntensity", "FailureRate");
		line = line.replace("Relability", "Reliability");
		line = line.replace("Probatility", "Probability");
		line = line.replace("type=\"intlist\"", "type=\"integerlist\"");
		return line;
	}

	private static String removeParenthesesAroundLists(String line) {
		Matcher matcher = Pattern.compile("list\">\\(([^\\)]*)").matcher(line);
		if (matcher.find())
			line = line.replace("(" + matcher.group(1) + ")", matcher.group(1));
		return line;
	}

}
