package archeopterix.util.io;

import java.io.File;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import archeopterix.problem.Component;
import archeopterix.problem.NetworkProperties;
import archeopterix.problem.Host;
import archeopterix.problem.InteractionProperties;
import archeopterix.problem.EmbeddedSystemProblem;
import archeopterix.util.PropertyMap;
import java.lang.reflect.Constructor;
import java.util.LinkedList;

/**
 * An XML parser for Problems; BUT how should different problems be dealt with, or do they all have the same details?
 */
public final class ProblemXMLParser extends XMLParser {
	private static final Log log = LogFactory.getLog(ProblemXMLParser.class);
	private static final DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
	private static final TransformerFactory transformerFactory = TransformerFactory.newInstance();	
	public static final String SYSTEM = "system";
	public static final String COMPONENTS = "components";
        public static final String COMPONENT = Component.LABEL.toLowerCase();
	public static final String PROCESSORS = "processors";
        public static final String PROCESSOR = Host.LABEL.toLowerCase();
	public static final String BUSES = "buses";
        public static final String BUS = NetworkProperties.LABEL.toLowerCase() ;
	public static final String INTERACTIONS = "interactions";
        public static final String INTERACTION = InteractionProperties.LABEL.toLowerCase();
		
	private ProblemXMLParser() { }
	
	public static EmbeddedSystemProblem loadProblem(String sFileName, Class<?> problemClass) {
            return loadProblem( new File(sFileName) , problemClass );
	}
	
    public static EmbeddedSystemProblem loadProblem(File file, Class<?> problemClass){
        if (!file.exists())
            throw new Error("File not found: "+file);

        Document dom = parseXMLFile(file);
        if (dom == null)
            throw new Error("Unable to load configuration from file is it a bit extreme to throw an unrecoverable error?");

        List<Component> lstComps = new LinkedList<Component>();
        List<Host> lstECUs = new LinkedList<Host>();
        Element docEle = dom.getDocumentElement();

        NodeList nl = docEle.getElementsByTagName(PROCESSOR);
        if(nl != null && nl.getLength() > 0) {
            for(int i = 0, c = nl.getLength(); i < c; i++) {
                lstECUs.add( new Host(Host.LABEL+i, i, getPropertyMap(nl.item(i))) );
            }
        }

        final int nECUs = lstECUs.size();
        List<List<PropertyMap>> oNetworkMatrix = new LinkedList<List<PropertyMap>>();

        //First creating the Network Matrix
        for (int i = 0; i < nECUs; i++) {
                List<PropertyMap> oNetworkList = new LinkedList<PropertyMap>();
                for (int j = 0; j < nECUs; j++) //bit of a nasty hack to account for XML possibly not containing them in order
                        oNetworkList.add(null);
                oNetworkMatrix.add(oNetworkList);
        }

        //Filling up the network matrix from xml file inputs
        nl = docEle.getElementsByTagName(BUS);
        if (nl != null && nl.getLength() > 0) {
            for (int i = 0 ; i < nl.getLength(); i++) {
                PropertyMap propMap = getPropertyMap(nl.item(i));
                int iSrc = propMap.getInt(NetworkProperties.SRC_NODE);
                int iDest = propMap.getInt(NetworkProperties.DEST_NODE);
                List<PropertyMap> oNetworkList = oNetworkMatrix.get(iSrc);
                oNetworkList.set(iDest, propMap);
                oNetworkMatrix.set(iSrc, oNetworkList); // for safety :)
            }
        }

        //get Components
        nl = docEle.getElementsByTagName(COMPONENT);
        if(nl != null && nl.getLength() > 0) {
                for(int i = 0 ; i < nl.getLength();i++)
                        lstComps.add( new Component(COMPONENT+i, i, getPropertyMap(nl.item(i)) ) );
        }

        final int nComponents = lstComps.size();
        List<List<PropertyMap>> oInteractMatrix = new LinkedList<List<PropertyMap>>();

        //First creating the Interaction Matrix
        for (int i = 0; i < nComponents; i++) {
            List<PropertyMap> oInteractList = new LinkedList<PropertyMap>();
            for (int j = 0; j < nComponents; j++) //bit of a nasty hack to account for XML possibly not containing them in order
                oInteractList.add(null);
            oInteractMatrix.add(oInteractList);
        }
        
        nl = docEle.getElementsByTagName(INTERACTION);
        if (nl != null && nl.getLength() > 0) {
            for (int i = 0; i < nl.getLength(); i++) {
                PropertyMap propMap = getPropertyMap(nl.item(i));
                int iSrc = propMap.getInt(InteractionProperties.SRC_NODE);
                int iDest = propMap.getInt(InteractionProperties.DEST_NODE);
                try{
                    oInteractMatrix.get(iSrc).set(iDest, propMap); // for safety :)
                }catch(Exception ex){
                    System.out.println(ex.toString()+" IScr"+iSrc+" Ides"+iDest);
                }
            }
        }
        try {
            Constructor<?> cons = problemClass.getConstructor(List.class, List.class, List.class, List.class);
            return (EmbeddedSystemProblem) cons.newInstance(lstECUs, lstComps, oNetworkMatrix, oInteractMatrix);
//            return new EmbeddedSystemProblem(lstECUs, lstComps, oNetworkMatrix, oInteractMatrix);
        } catch (Exception e) {
            throw new RuntimeException("Unabled to instantiate problem "
					+ problemClass.getCanonicalName(), e);
        }
		//return new EmbeddedSystemProblem(lstECUs, lstComps, oNetworkMatrix, oInteractMatrix);
    }
	
    public static boolean saveProblem(EmbeddedSystemProblem problem, String sFileName) {
        try {
            Document doc = generateXMLDocumentForProblem(problem);	
            DOMSource domSource = new DOMSource(doc);
            Transformer transformer = transformerFactory.newTransformer();
            transformer.setOutputProperty(OutputKeys.METHOD, "xml");
            transformer.setOutputProperty(OutputKeys.ENCODING,"ISO-8859-1");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");

            StreamResult sr = new StreamResult( new File(sFileName));
            transformer.transform(domSource, sr);
            return true;
        } catch (Exception e) {
            if (log.isErrorEnabled())
                log.error("Error writing problem to file", e);
        }
        return false;
    }
	
    private static Document generateXMLDocumentForProblem(EmbeddedSystemProblem problem) {
        DocumentBuilder builder;
        try {
            builder = docBuilderFactory.newDocumentBuilder();
            DOMImplementation impl = builder.getDOMImplementation();
            Document doc = impl.createDocument(null, null, null);
            Element rootElement = doc.createElement(SYSTEM);
            doc.appendChild(rootElement);
            attachObjectsWithProperties(problem.getComponents(), COMPONENTS, COMPONENT, doc, rootElement);
            attachObjectsWithProperties(problem.getECUs(), PROCESSORS, PROCESSOR, doc, rootElement);
            attachMatrixElementPropertyMaps(problem.getNetworkMatrix(), BUSES, BUS, doc, rootElement);          
            attachMatrixElementPropertyMaps(problem.getInteractionsMatrix(), INTERACTIONS, INTERACTION, doc, rootElement);       
            return doc;
        } catch (ParserConfigurationException pce) {
            if (log.isErrorEnabled())
                log.error(pce);
            throw new RuntimeException(pce);
        }
    }

}
