package archeopterix.util.io;

import archeopterix.util.statistics.BetaDistribution;
import archeopterix.util.statistics.ExponentialDistribution;
import archeopterix.util.statistics.GammaDistribution;
import archeopterix.util.statistics.NormalDistribution;
import archeopterix.util.statistics.ProbabilityDistribution;
import java.io.File;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.LinkedList;
import javax.xml.parsers.DocumentBuilderFactory;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;

import archeopterix.problem.ObjectWithProperties;
import archeopterix.problem.PairwiseProperties;
import archeopterix.util.PropertyMap;
import archeopterix.util.probability.DiscreteDistribution;
import archeopterix.util.probability.ShiftedBetaDistribution;
import archeopterix.util.probability.TriangularDistribution;


public abstract class XMLParser implements ArcheOpterixXMLConstants {
        public static String ARRAY_SEPERATOR = "|";
        public static String ARRAY_SEPERATOR_REGEX = "\\|";
        public static String ELEMENT_SEPERATOR_REGEX = ",[\\s]*";       
	private static final Log log = LogFactory.getLog(XMLParser.class);
	protected static final Map<Class<?>, String> classToTypeSpecifierMap = new HashMap<Class<?>, String>();	
	static {
            classToTypeSpecifierMap.put(Double.class, DT_DOUBLE);
            classToTypeSpecifierMap.put(Float.class, DT_FLOAT);
            classToTypeSpecifierMap.put(Integer.class, DT_INTEGER);
            classToTypeSpecifierMap.put(Boolean.class, DT_BOOLEAN);
            classToTypeSpecifierMap.put(String.class, DT_STRING);
            classToTypeSpecifierMap.put(ProbabilityDistribution.class, DT_PROB_DIST);
            classToTypeSpecifierMap.put(Class.class, DT_CLASS);
	}

	public static Document parseXMLFile(String sFileName) {
            return parseXMLFile( new File(sFileName) );
	}
	
	public static Document parseXMLFile(File file) {
            Document dom = null;
            try {
                dom = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(file);
            } catch (Exception e) {
                if (log.isErrorEnabled())
                    log.error("Problem parsing XML file", e);
            }
            return dom;
	}
	public static PropertyMap getPropertyMap(Node node) {
            return getPropertyMap(node, PROPERTY);
	}
	
	public static PropertyMap getPropertyMap(Node node, final String propertyElementAlias) {
            PropertyMap propMap = new PropertyMap();
            for (Node childNode = node.getFirstChild(); childNode != null; childNode = childNode.getNextSibling()) {

                if (childNode.getNodeName().equals(propertyElementAlias) ) {
                    String name = ((Element)childNode).getAttribute(ATTRIB_NAME);
                    Element childEl = (Element)childNode;
                    propMap.put(childEl.getAttribute(ATTRIB_NAME), parseValue( combineAllTextChildren(childEl), childEl.getAttribute(ATTRIB_TYPE) ));
                    //System.out.println("Node" + sKey + "=" + sVal.toString());
                }
            }
            return propMap;
	}

	public static Object parseValue(String stringValue, String typeSpecifier) {
		if (typeSpecifier.isEmpty())
                    throw new RuntimeException("Type of property or parameter cannot be empty (given value was \"" + stringValue + "\"");
		if (typeSpecifier.equals(DT_STRING))
                    return stringValue;
		else if (typeSpecifier.equals(DT_BOOLEAN))
                    return Boolean.parseBoolean(stringValue);
		else if (typeSpecifier.equals(DT_INTEGER))
                    return Integer.parseInt(stringValue);
		else if (typeSpecifier.equals(DT_FLOAT))
                    return Float.parseFloat(stringValue);
		else if (typeSpecifier.equals(DT_DOUBLE))
                    return Double.parseDouble(stringValue);
		else if (typeSpecifier.equals(DT_PROB_DIST))
                    return parseProbabilityDistribution(stringValue);
		else if (typeSpecifier.equals(DT_INTLIST)) {
                    String[] cells = stringValue.split(ELEMENT_SEPERATOR_REGEX);
                    List<Integer> lst = new LinkedList<Integer>();
                    for (String cell : cells)
                        lst.add( Integer.parseInt(cell) );
                    return lst;
		} else if (typeSpecifier.equals(DT_DOUBLELIST)) {
                    String[] cells = stringValue.split(ELEMENT_SEPERATOR_REGEX);
                    List<Double> lst = new LinkedList<Double>();
                    for (String cell : cells)
                        lst.add( Double.parseDouble(cell) );
                    return lst;
		} else if (typeSpecifier.equals(DT_PROB_DIST_LIST)) {
                    String[] cells = stringValue.split(ARRAY_SEPERATOR_REGEX);
                    List<ProbabilityDistribution> lst = new LinkedList<ProbabilityDistribution>();
                    for (String cell : cells)
                        lst.add( parseProbabilityDistribution(cell) );
                    return lst;
		} else if (typeSpecifier.equals(DT_CLASS)) {
                    Class<?> classValue = loadClassFromName(stringValue);
                    if (classValue == null) //FIXME Too extreme? More appropriate exception type?
                        throw new RuntimeException("Unable to load class using name " + stringValue);
                    return classValue;
		} else if (typeSpecifier.equals(DT_CLASSLIST)) {
                    String[] cells = stringValue.split(ARRAY_SEPERATOR_REGEX);
                    List<Class<?>> lst = new LinkedList<Class<?>>();
                    for (String cell : cells) {
                        Class<?> classValue = loadClassFromName(cell);
                        if (classValue == null) //FIXME Too extreme? More appropriate exception type?
                                throw new RuntimeException("Unable to load class using name " + stringValue);
                        lst.add( classValue );
                    }
                    return lst;
		} else { //Attempt to parse type as fully-qualified name of an enum and then interpret value as value of that enum
                    try {
                        Class<?> type = Class.forName(typeSpecifier);
                        if (! type.isEnum())
                            throw new ClassNotFoundException(typeSpecifier);
                        //Messy, but Enum.valueOf requires that we already know the exact type of the Enum, in which case we would call *its* valueof method instead, so why do they bother to have it!?
                        try {
                            return type.getMethod("valueOf", String.class).invoke(null, stringValue);
                        } catch (Exception e) {
                                throw new ClassNotFoundException("Unable to obtain instance of stated enumerated type", e);
                        }
                    } catch (ClassNotFoundException cnfe) {
                        throw new IllegalArgumentException("Specified type not recognised and not the name of an enumerated type: "+ cnfe.toString() + typeSpecifier);
                    }
		}
	}

	private static Class<?> loadClassFromName(String className) {
            try {
                return Class.forName(className);
            } catch (ClassNotFoundException cnfe) {
                if (log.isWarnEnabled())
                    log.warn("Failed to load class using name", cnfe);
                return null;
            }
	}

	public static void addPropertiesToElement(PropertyMap propertyMap, Document context, Node anchorPoint) {
            for (Map.Entry<String,Object> property : propertyMap.entrySet()){
                anchorPoint.appendChild( createPropertyWithValue(context, property.getKey(), property.getValue()) );
            }
	}

	protected static Element createPropertyWithValue(Document context, String propertyName, Object value) {
            Element el = context.createElement(PROPERTY);
            el.setAttribute(ATTRIB_NAME, propertyName);
            el.setAttribute(ATTRIB_TYPE, determineTypeSpecifier(value) );
            //FIXME Need a cleaner, simpler system for mapping types to type specifiers and to their String representations
            String stringValue;
            if (value instanceof List<?>) {
                List<?> lstTemp = (LinkedList<?>) value;
                if (! lstTemp.isEmpty() && ((List<?>)value).get(0) instanceof ProbabilityDistribution) {
                    stringValue = "";
                    for (int i = 0; i < lstTemp.size(); i++) {
                        ProbabilityDistribution pd = (ProbabilityDistribution) lstTemp.get(i);
                        stringValue += pd.getBufString() + ARRAY_SEPERATOR;
                    }
                    stringValue = stringValue.substring(0, stringValue.length() - 1);
                } else {
                    stringValue = value.toString();
                    stringValue = stringValue.substring(1, stringValue.length() - 1);
                }
            } else {
                stringValue = value.toString();
            }
            el.appendChild( context.createTextNode(stringValue) );
            return el;
	}

	protected static String determineTypeSpecifier(Object value) {
            String simpleType = classToTypeSpecifierMap.get(value.getClass());
            if (simpleType != null)
                    return simpleType;
            else if (value instanceof Enum<?>)
                    return value.getClass().getCanonicalName();
            else if (value instanceof ProbabilityDistribution)
                    return DT_PROB_DIST;
            //Warning, all subsequent collection types must contain at least one element to determine their type (stupid Generics)
            else if (value instanceof List<?> && !((List<?>)value).isEmpty()) {
                if (((List<?>)value).get(0) instanceof Integer)
                    return DT_INTLIST;
                else if (((List<?>)value).get(0) instanceof Double)
                    return DT_DOUBLELIST;
                else if (((List<?>)value).get(0) instanceof ProbabilityDistribution)
                    return DT_PROB_DIST_LIST;
                else if (((List<?>)value).get(0) instanceof Class<?>)
                    return DT_CLASSLIST;
            }
            return null;
	}

	public static String getFirstMatchingElementTextValue(Element root, String tagName) {
		String textVal = null;
		NodeList nl = root.getElementsByTagName(tagName);
		if (nl != null && nl.getLength() > 0)
			textVal = combineAllTextChildren( (Element) nl.item(0) );
		return textVal;
	}

	/**
	 * Naive approach to dealing with contents of a node being on the next
	 * line; if {@code el} has structural elements as children, the result of
	 * this method will not be what was desired.
	 */
	public static String combineAllTextChildren(Element el) {
		StringBuilder builder = new StringBuilder();
		NodeList children = el.getChildNodes();
		for (int i = 0, c = children.getLength(); i < c; i++)
			builder.append(children.item(i).getNodeValue());
		return builder.toString().trim();
	}

	public static void attachObjectsWithProperties(Collection<? extends ObjectWithProperties> obs,
			final String collectionTag, final String itemTag,
			Document context, Element anchorPoint){
            Element collectionEl = context.createElement(collectionTag);
            anchorPoint.appendChild(collectionEl);
            for (ObjectWithProperties o : obs) {
                    Element oEl = context.createElement(itemTag);
                    addPropertiesToElement(o.getPropertyMap(), context, oEl);
                    collectionEl.appendChild(oEl);
            }
	}

	public static void attachMatrixElementPropertyMaps(List<List<PropertyMap>> matrix,
			final String matrixTag, final String elementTag,
			Document context, Element rootElement){
            Element matrixEl = context.createElement(matrixTag);
            rootElement.appendChild(matrixEl);
            int iSrc = 0;
            for (List<PropertyMap> srcs : matrix) {
                int iDest = 0;
                for (PropertyMap propertyMap : srcs) {
                    if (propertyMap != null ) {
                        Element elementEl = context.createElement(elementTag);
                        addPropertiesToElement(propertyMap, context, elementEl);
                        elementEl.appendChild( createPropertyWithValue(context, PairwiseProperties.SRC_NODE, iSrc) );
                        elementEl.appendChild( createPropertyWithValue(context, PairwiseProperties.DEST_NODE, iDest) );
                        matrixEl.appendChild(elementEl);
                    }
                    iDest++;
                }
                iSrc++;
            }
         
	}
    public static void attachMatrixElementPropertyMaps(LinkedList<List<List<PropertyMap>>> matrix,
			final String matrixTag, final String elementTag,
			Document context, Element rootElement)
	{
	Element matrixEl = context.createElement(matrixTag);
	rootElement.appendChild(matrixEl);
	int iSrc = 0;
	for (List<List<PropertyMap>> mx : matrix) {
            for (List<PropertyMap> srcs : mx) {
                int iDest = 0;
                for (PropertyMap propertyMap : srcs) {
                    if (propertyMap != null ) {
                        Element elementEl = context.createElement(elementTag);
                        addPropertiesToElement(propertyMap, context, elementEl);
                        elementEl.appendChild( createPropertyWithValue(context, PairwiseProperties.SRC_NODE, iSrc) );
                        elementEl.appendChild( createPropertyWithValue(context, PairwiseProperties.DEST_NODE, iDest) );
                        matrixEl.appendChild(elementEl);
                    }
                    iDest++;
                }
                iSrc++;
            }
	}
}
	public static ProbabilityDistribution parseProbabilityDistribution(String sParams){
		String[] sArr = sParams.split(ELEMENT_SEPERATOR_REGEX);
		if (sArr.length < 1) {
                    System.err.println("Invalid probability distribution specification:"+sParams);
                    return null;
		}

		if (sArr[0].trim().equalsIgnoreCase(ProbabilityDistribution.PD_NORMAL)) {
                    if (sArr.length < 3) {
                        System.err.print("Invalid Arguments for Distribution :"+ sParams);
                        return null;
                    }
                    return new NormalDistribution( Double.parseDouble(sArr[1].trim()), Double.parseDouble(sArr[2].trim()) );
		}
		else if (sArr[0].trim().equalsIgnoreCase(ProbabilityDistribution.PD_BETA)) {
                    if (sArr.length < 3) {
                        System.err.print("Invalid Arguments for Distribution :"+ sParams);
                        return null;
                    }
                    return new BetaDistribution( Double.parseDouble(sArr[1].trim()), Double.parseDouble(sArr[2].trim()) );
		}
		else if (sArr[0].trim().equalsIgnoreCase(ProbabilityDistribution.PD_BETA_SHD)) {
                    if (sArr.length < 5) {
                        System.err.print("Invalid Arguments for Distribution :"+ sParams);
                        return null;
                    }
                    return new ShiftedBetaDistribution( Double.parseDouble(sArr[1].trim()), Double.parseDouble(sArr[2].trim()),
                            Double.parseDouble(sArr[3].trim()), Double.parseDouble(sArr[4].trim()));
		}
		else if (sArr[0].trim().equalsIgnoreCase(ProbabilityDistribution.PD_GAMMA)){
			if (sArr.length < 2) {
				System.err.print("Invalid Arguments for Distribution :"+ sParams);
				return null;
			}
			return new GammaDistribution( Double.parseDouble(sArr[1].trim()) );
		}
		else if (sArr[0].trim().equalsIgnoreCase(ProbabilityDistribution.PD_EXP)){
			if (sArr.length < 2) {
				System.err.print("Invalid Arguments for Distribution :"+ sParams);
				return null;
			}
			return new ExponentialDistribution( Double.parseDouble(sArr[1].trim()) );
		}
		else if (sArr[0].trim().equalsIgnoreCase(ProbabilityDistribution.PD_TRIANGULAR)){
			if (sArr.length < 4) {
				System.err.print("Invalid Arguments for Distribution :"+ sParams);
				return null;
			}
			return new TriangularDistribution( Double.parseDouble(sArr[1].trim()), Double.parseDouble(sArr[2].trim()),Double.parseDouble(sArr[3].trim()) );
		}
		else if (sArr[0].trim().equalsIgnoreCase(ProbabilityDistribution.PD_DISCRETE)){
			if (sArr.length < 2) {
				System.err.print("Invalid Arguments for Distribution :"+ sParams);
				return null;
			}
			double[] discreteParams = new double[ sArr.length - 1 ];
			for (int i = 0, j = 1; j < sArr.length; i++, j++)
				discreteParams[i] = Double.parseDouble( sArr[j] );
			return new DiscreteDistribution(discreteParams);
		}
                else{ //Check for Single values, and present them as discrete distibutions
                    try{
                        Double dVal = Double.parseDouble(sArr[0]);
                        double[] discreteParams = {dVal, 1.0};
                        return new DiscreteDistribution(discreteParams);
                    }
                    catch(NumberFormatException ex){
                        System.err.println("Unsupported probability distribution:"+ sParams);
                        return null;
                    }
                }
	}
}
