package archeopterix.util.probability;

import archeopterix.util.statistics.ProbabilityDistribution;
import archeopterix.util.statistics.OutOfRangeException;

/**
 *
 * @author indika
 */
public class TriangularDistribution extends ProbabilityDistribution {

    private double a, b, c;
    private double h, m1, m2;

    /**
     * Constructs a beta distribution.
     * @param dgrP degrees of freedom p.
     * @param dgrQ degrees of freedom q.
     */
    public TriangularDistribution(double dMin, double dMid, double dMax) {
        s_Name = PD_TRIANGULAR;
        if (dMid < dMin || dMax < dMid) {
            throw new OutOfRangeException("Min= " + dMin + " Mid=" + dMid + " Max=" + dMax);
        }
        a = dMin;
        b = dMid;
        c = dMax;
        h = 2 / (c - a);
        m1 = (h / (b - a));
        m2 = (h / (c - b)); //made positive
        }

    public double getMin() {
        return a;
    }

    public double getMid() {
        return b;
    }

    public double getMax() {
        return c;
    }

    public double probability(double X) {
        //checkRange(X);
        if (X < a) {
            return 0.0;
        } else if (X <= b) {
            return (X - a) * m1;
        } else if (X <= c) {
            return (c - X) * m2;
        } else {
            return 0.0;
        }
    }

    /**
     * Cumulative Triangular distribution.
     * @return the probability that a stochastic variable x is less then X, i.e. P(x&lt;X).
     */
    public double cumulative(double X) {
        checkRange(X);
        if (X < a) {
            return 0.0;
        } else if (X <= b) {
            return 0.5 * (X - a) * (X - a) * m1;
        } else if (X <= c) {
            return 1 - (0.5 * (c - X) * (c - X) * m2);
        } else {
            return 1;
        }
    }

    /**
     * Inverse of the cumulative beta distribution function.
     * @return the value X for which P(x&lt;X).
     */
    public double inverse(double probability) {
        checkRange(probability);
        if (probability == 0.0) {
            return 0.0;
        }
        if (probability == 1.0) {
            return 1.0;
        }
        return findRoot(probability, 0.5, a, c);
    }

    public double getMean() {
        return ((getMin() + getMid() + getMax()) / 3); //Wiki page
        }

    @Override
    public String getBufString() {
        return PD_TRIANGULAR + "," + a + "," + b + "," + c;
    }

    @Override
    public String toString(){
        return getName() + "(" + df.format(a) + ","+ df.format(b) + ","+ df.format(c) +")";
    }

}
